{-|
  This module defines the command line options for the main
  @ipc@ executable, but can also be used by most of the example
  programs in the example directory.
-}
module Ipc.Cli
   ( getCliArgs
   , Cli                  ( .. )
   , CliOpt               ( .. )
   , CliVersion
   , CliArg
   , toCli

   -- Exported for @CliDoc.hs@
   , ipcManual
   , printManual

   -- Exported for @Smc/Cli.h@
   , Manual
   , ManualSection
   , ManualDesc
   , ManualEntry
   , optDescsOfManual


   , baseCliOptions

   , getStaticAnalysis
   , getStaunch

   , getProbeSpecs
   , getMasterBool
   , getStartActions
   , getStopActions

   , getRateOptions
   , getProcessRenames
   , getRateRenames
   , getPrioritised

   , containsHideNonCoop
   , getProcessNumber
   , getModFileName
   , getPrismFileName
   , getDizzyFileName
   , getOutputFileName
   , getSolutionControl
   , getMeasurementSpecs

   , PostHydra            ( .. )
   , getPostHydra

   , shouldRunHydra
   , getHydraCommands

   , getStandardOut

   , getNonStandard
   )
where

{- Standard imported libraries -}
import Data.Char
   ( isSpace )
import Data.List
   ( find )
import Data.Maybe
   ( mapMaybe )
import System.Environment
   ( getArgs )
import System.Console.GetOpt
   ( getOpt
   , usageInfo
   , ArgOrder  ( .. )
   , OptDescr  ( .. )
   , ArgDescr  ( .. )
   )
import Text.ParserCombinators.Parsec
   ( ParseError )
{- External Library modules imported -}
{- Local imported libraries -}
import Language.Pepa.Syntax
   ( QualifiedName     ( .. )
   , RateIdentifier
   , ActionIdentifier

   , ParsedPriority
   , defaultIncreasedPriority
   )
import Language.Pepa.Utils
   ( mapFst
   , switchSuffix
   , switchBaseName
   )
import Language.Pepa.Parser
   ( parseCommaSeparatedLowers
   , parseRateOption
   )
import Language.Pepa.Transform.Replace
   ( ProcessRename
   , ProcessRenameMap
   , RateRename
   , RateRenameMap
   )
import Ipc.CliParsers
   ( joinParseErrorList
   , parseProcessRename
   , parseRateRename
   , parseCommaActionsMaybeAssigns
   )

{- Hydra stuff required for measurement specifications -}
import Language.Hydra.Syntax
   ( DNAMperformMeasure    ( .. )
   , DNAMpassageMeasure    ( .. )
   , DNAMtransientMeasure  ( .. )
   , DNAMsteadyMeasure     ( .. )
   , DNAMcountMeasure      ( .. )
   , DNAMsteadyEstimator   ( .. )

   , DNAMsolutionControl   ( .. )
   , parseSolutionMethod
   , defaultSolutionMethod

   , DNAMcondition         ( .. )
   )

import Language.Pepa.Utils
  ( mapSnd
  , mkSepList
  )
{- End of module imports -}


{-
  These definitions allow us to generate latex from the option descriptions.
-}
optDescOfManualDesc :: ManualDesc a -> OptDescr (CliOpt a)
optDescOfManualDesc = snd

optDescsOfManualSection :: ManualSection a -> [ OptDescr (CliOpt a) ]
optDescsOfManualSection = (map optDescOfManualDesc) . snd

optDescsOfManual :: Manual a -> [ OptDescr (CliOpt a) ]
optDescsOfManual = concat . (map optDescsOfManualSection)

-- This top version is very strange it's an uncallable option. Shouldn't occur.
manualPartOfManualDesc :: ManualDesc a -> String
manualPartOfManualDesc ( "", Option "" [] _ _)                = ""
manualPartOfManualDesc ("", optDesc@(Option _ _ _ usage))     =
  unlines [ "\\begin{description}"
          , "\\item[" ++ headerBitOfOptDesc optDesc ++ "]"
          , "No manual entry but the usage information states:"
          , usage
          , "\\end{description}"
          ]
manualPartOfManualDesc (s, optDesc)                           =
  unlines [ "\\begin{description}"
          , "\\item[" ++ headerBitOfOptDesc optDesc ++ "]"
          , s
          , "\\end{description}"
          ]

-- From the actual opt desc we can print out a useful header bit
headerBitOfOptDesc :: OptDescr (CliOpt a) -> String
headerBitOfOptDesc (Option shortNames longNames args _) =
  ( mkSepList ", " $ concat [ map mkShort shortNames
                            , map mkLong longNames
                            ]
  ) ++ " " ++ argument
  where
  mkShort :: Char -> String
  mkShort c = [ '-', c ]
  mkLong :: String -> String
  mkLong s = "\\ipcflag{" ++ s ++ "}"

  argument :: String
  argument = case args of
               NoArg  _   -> ""
               ReqArg _ s -> s
               OptArg _ s -> "<" ++ s ++ ">"
               

printManualSection :: ManualSection a -> String
printManualSection (header, descs) =
  unlines $ header : (map manualPartOfManualDesc descs)

printManual :: Manual a -> String
printManual = unlines . (map printManualSection)

type Manual a        = [ ManualSection a ]
type ManualSection a = (ManualEntry, [ ManualDesc a ] )
type ManualDesc a    = (ManualEntry, OptDescr (CliOpt a))
type ManualEntry     = String



{-|
   A simple wrapper for 'System.Environment.getArgs' which interprets any
   @--argfile file@ options.
-}
getCliArgs :: IO [ String ]
getCliArgs =
   do args <- getArgs
      interpretArgFiles [] args
   where
   -- The first argument is the list of argument files we have
   -- already seen, we do not wish to enter a loop.
   interpretArgFiles :: [ FilePath ] -> [ String ] -> IO [ String ]
   interpretArgFiles seen ("--argfile" : argfileName : rest)
      | elem argfileName seen         = 
         do putStrLn "Warning: loop in '--argfiles' files"
            interpretArgFiles seen rest
      | otherwise                     =
         do contents <- readFile argfileName
            interpretArgFiles (argfileName : seen) $ 
                  (argsOfFile contents) ++ rest
   interpretArgFiles _seen []         = return []
   interpretArgFiles seen (h : rest)  = 
      do remainder <- interpretArgFiles seen rest
         return $ h : remainder
   
   argsOfFile :: String -> [ String ]
   argsOfFile = (filter (not . (all isSpace))) . parseWords

   -- This is a small parser for argument files.
   parseWords :: String -> [ String ]
   parseWords "" = []
   parseWords ( '#' : rest) =
      parseWords $ ignoreComment rest
   parseWords ( '"' : rest) =
      w : (parseWords remainder)
      where (w, remainder) = parseStringLiteral rest
   parseWords s             =
      w : (parseWords rest)
      where (w, rest) = parseWord $ dropWhile isSpace s

   parseWord :: String -> (String, String)
   parseWord = break endsWord

   endsWord :: Char -> Bool
   endsWord '#' = True
   endsWord '"' = True
   endsWord c   = isSpace c

   ignoreComment :: String -> String
   ignoreComment = dropWhile (not . (== '\n'))

   parseStringLiteral :: String -> (String, String)
   parseStringLiteral ""                   =
      error "string literal not finished"
   parseStringLiteral ( '\\' : '"' : rest) =
      mapFst ( '"' :) $ parseStringLiteral rest
   parseStringLiteral ( '"' : rest)        =
      ("" , rest)
   parseStringLiteral (h : rest)           =
      mapFst ( h :) $ parseStringLiteral rest

data CliOpt a = 
   CliVerbose 
 | CliVersion 
 | CliHelp 

 {- Options controlling static analysis -}
 | CliStaunch
 | CliNoStaticAnalysis

 {- Options for controlling performance measurements -} 
 | CliProbeSpec String
 | CliNoMasterProbe
 | CliStartActions [ String ]
 | CliStopActions  [ String ]

 | CliPassage
 | CliSteady
 | CliTransient
 | CliCount        [ String ]
 | CliNoMeasurement

 | CliSteadyMean
 | CliSteadyVariance
 | CliSteadyStdDev
 | CliSteadyDistrib

 | CliStartTime Double
 | CliStopTime  Double
 | CliTimeStep  Double
 
 | CliSolver String

 {- Options for pre-modifying the model -}
 | CliRate (RateIdentifier, Double)

 {- Options for replacing names within the model -}
 | CliReplaceProcess String
 | CliReplaceRate    String

 {- This may well be a temporary command-line option, we definitely
    need to allow this in the syntax anyway and maybe as an option
    it doesn't work so well, but this is more or less just a proof
    of concept idea.
    Actually I now believe that we are as well to have a command-line
    option as well, I think it works okay.
 -}
 | CliPrioritise [ (ActionIdentifier, ParsedPriority) ]


 {- Options controlling the model transformations -}
 | CliNoReduceRateExps
 | CliHideNonCoop

 | CliProcessNum Int

 | CliShowSimplified


 {- Options for running hydra -}
 | CliRunHydra
 | CliHydra String

   {- compilations other than hydra -}
 | CliPrism
 | CliFsp
 | CliDizzy

   {- options for controlling the output locations -}
 | CliModFileName   FilePath
 | CliPrismFileName FilePath
 | CliDizzyFileName FilePath
 | CliOutput        FilePath
 | CliStandardOut

 | CliNonStandard a
 deriving (Show, Eq)


{-| The type of the version used by the command-line -}
type CliVersion = String


{-|
   The main argument given to 'System.Console.GetOpt.getOpt',
   it describes the flags or options that the program may be given.
   Each element basically describes how the user gives the argument on the
   command line and a function from a possible option argument to the
   above data type. 
   Given this the 'System.Console.GetOpt.getOpt can return as a part of the
   triple list of the above 'CliOpt' data type.

   These are the baseCliOptions which are common to several or all of
   the ipc-related programs (well okay there may be some specific to 
   just ipc itself). The 'toCli' function below takes a list of this
   kind as input rather than simply using this list, this is such that
   each individual program can add to this list before command-line
   parsing.

These then are the options for the ipc compiler
-}
baseCliOptions :: [ OptDescr (CliOpt a) ]
baseCliOptions = 
  optDescsOfManual ipcManual

ipcManual :: Manual a
ipcManual =
  [ trivialOptionSection
  , outputOptionsSection
  , staticAnalysisSection
  , probeSpecificationSection
  , runningHydraSection
  , prismModelSection
  , dizzyModelSection
  , miscelleaneousSection
  ]
   

{-
  The options for the trivial invocations of ipc.
-}


trivialCliOptions :: [ OptDescr (CliOpt a) ]
trivialCliOptions = optDescsOfManualSection trivialOptionSection

trivialOptionSection :: ManualSection a
trivialOptionSection =
  ( header, entries)
  where
  header = unlines [ "\\subsection{Trivial Invocation Options}"
                   , "These options only apply to the trivial"
                   , "invocations of \\commandNameIpcSmc."
                   ]

  entries =
   [
     ( unlines [ "\\indexipcflag{ version }%"
               , "The command-line:\n"
               , "\\showcommandline{\\ipcflag{version}}"
               , "will print the version of the ipc compiler and exit."
               ]
     , Option "v"     [ "version" ]      ( NoArg CliVersion )        
        "show version info"
     )
   , ( unlines [ "\\indexipcflag{ help }%"
               , "The command-line\n"
               , "\\showcommandline{\\ipcflag{ help}}"
               , ""
               , "will print an option and usage summary and exit."
               ]
     , Option "h"     [ "help" ]         ( NoArg CliHelp )
        "show options and documentation"
     )
   ]


{-
  These options deal with re-directing the output
-}

outputOptionsSection :: ManualSection a
outputOptionsSection =
  (header, entries)
  where
  header = unlines [ "\\subsection{Output Re-directing Options}"
                   , "By default the \\commandNameIpcSmc\\ command"
                   , "will generate output in the same directory and"
                   , "with a file name based on the input source name"
                   , "The options in this section allow the user to"
                   , "redirect the output from the \\ipc\\ compiler"
                   ]
  entries =
    [
     ( unlines [ "Sets the output \\texttt{.mod} file which will be"
               , "the input to the run of \\hydra"
               ]
     , Option ""      [ "mod-file" ]       ( ReqArg CliModFileName "FILE" )
       "specify the file to output the mod file to"
     )

   , ( unlines [ "A generic output flag, this works regardless of the kind"
               , "output that \\ipc\\ is set to produce. For example"
               , "instead of the default \\hydra\\ the compiler may have"
               , "been set to produce a prism model file."
               , "The \\ipcflag{output} can be used to set the"
               , "output file name"
               ]
     , Option ""      [ "output" ]         ( ReqArg CliOutput "FILE" )
       "generic flag for specifying the output file"
     )
   , ( unlines [ "When debugging the compiler it can often be useful"
               , "for the output to be redirected to the terminal for"
               , "immediate inspection by the programmer."
               , "The \\ipcflag{stdout} sets the output file to be"
               , "the standard out. This may also prove useful for"
               , "piping the output into further processing tools."
               ]
     , Option ""      [ "stdout" ]         ( NoArg CliStandardOut )
       "instead of an output file write to standard out, useful for debugging"
     )
   ]


{-
  These are the options effecting static analysis.
-}


staticAnalysisCliOptions :: [ OptDescr (CliOpt a) ]
staticAnalysisCliOptions = optDescsOfManualSection staticAnalysisSection

staticAnalysisSection :: ManualSection a
staticAnalysisSection =
  (header, entries)
  where
  header = unlines [ "\\subsection{Static Analysis Options}"
                   , "These options control the way that"
                   , "static analysis is performed."
                   ]
  entries =
   [ 
     ( unlines [ "The flag \\ipcflag{ staunch } allows the compiler"
               , "to ignore warnings."
               , "By default this flag is off and when \\ipc\\ performs"
               , "any static-analys is over the input \\pepa\\ model"
               , "if there are any warnings it will cause the compiler"
               , "to cancel compilation. This behaviour can be suppressed with"
               , "the \\ipcflag{ staunch } flag. This will cause the warnings"
               , "to still be emitted but compilation will proceed anyway."
               ]
     , Option ""      [ "staunch" ]      ( NoArg CliStaunch )
       "perform the static analysis but compile even in the face of warnings"
     )
   , ( unlines [ "The flag \\ipcflag{ no-static-analysis } causes"
               , "\\ipc\\ to avoid performing the static analysis over"
               , "the \\pepa\\ model. It will therefore produce no"
               , "warnings or errors. Because of this compilation may fail"
               , "mysteriously and hence the user is advised only to use this"
               , "flag if they know exactly what they are doing and expect"
               , "their model to fail static-analysis for some reason but"
               , "wish to proceed to compilation anyway."
               ]
     , Option ""      [ "no-static-analysis" ] ( NoArg CliNoStaticAnalysis )
       "do not perform static analysis over the model"
     )
   ]


{-
  The options for probe specification.
-}

probeSpecificationOptions :: [ OptDescr (CliOpt a) ]
probeSpecificationOptions = optDescsOfManualSection probeSpecificationSection

probeSpecificationSection :: ManualSection a
probeSpecificationSection =
  (header, entries)
  where
  header = unlines [ "\\subsection{Probe Specification Options}"
                   , "This section describes the options which control"
                   , "the performance measure specification probes which"
                   , "are added to the model."
                   , "% For more information on this please see section ..."
                   ]
  entries =
   [
     ( unlines [ "The flag \\ipcflag{ probe } with the shortened version"
               , "\\texttt{p} is used to give a full probe specification."
               , "% In the probe specification language given in section ??"
               ]
     , Option "p"     [ "probe" ]        ( ReqArg CliProbeSpec "PROBESPEC" )
       "specify a performance evaluation probe"
     )
   , ( unlines [ "The flag \\ipcflag{ no-master } is used to specify that"
               , "no master probe % as described in section ..."
               , "should be automatically added to the model."
               ]
     , Option ""      [ "no-master" ]    ( NoArg CliNoMasterProbe )
       "do not automatically generate and add a master start stop probe"
     )
   , ( unlines [ "The flags \\ipcflag{ source } and \\ipcflag{ target }"
               , "with the short versions \\texttt{s} and \\texttt{t}"
               , "respectively, set the state switching actions used"
               , "in the master probe. If other probes are added using the"
               , "\\ipcflag{ probe } option then they may perform immediate"
               , "communication actions which are specified in the source and"
               , "target action list."
               , "Both the \\ipcflag{ source } and \\ipcflag{ target } flags" 
               , "accept as argument a comma separated list of action names."
               ]
     , Option "s"     [ "source" ]       ( ReqArg startActions "ACTIONS" )
       "specify a list of actions to start the master probe (default 'start')"
     )
   , ( "see \\texttt{--source}"
     , Option "t"     [ "target" ]       ( ReqArg stopActions "ACTIONS" )
       "specify a list of actions to stop the master probe (default 'stop')"
     )
   ]
   where
   startActions = CliStartActions . (parseCommaSeparatedLowers startError)
   startError   = "error whilst parsing a list of specified start actions"
   stopActions  = CliStopActions  . (parseCommaSeparatedLowers stopError)
   stopError    = "error whilst parsing a list of specified stop actions"


{-
  The command-line options for running hydra automatically.
-}

runningHydraOptions :: [ OptDescr (CliOpt a) ]
runningHydraOptions = optDescsOfManualSection runningHydraSection

runningHydraSection :: ManualSection a
runningHydraSection =
  (header, entries)
  where
  header = unlines [ "\\subsection{Running Hydra}"
                   , "This section details options used for running"
                   , "the hydra tool after processing of the model by \\ipc."
                   ]
  entries =
   [ ( unlines [ "The flag \\ipcflag{run-hydra} causes \\ipc\\ to"
               , "automatically run the hydra tool on the produced"
               , "\\texttt{.mod} file."
               ]
     , Option ""      [ "run-hydra" ]      (NoArg CliRunHydra )
       "After compilation to a mod file, run hydra on the mod file"
     )
   , ( unlines [ "If the hydra tool is not installed in a standard"
               , "location the path to the \\hydra\\ executable can"
               , "be given as an argument to the \\ipcflag{hydra} option."
               ]
     , Option ""       [ "hydra" ]          (ReqArg CliHydra "PATH")
       "Provide the path to hydra if it is not in the system path"
     )
   ]

{- 
  Some options for generating prism models.
-}
prismModelSection :: ManualSection a
prismModelSection =
  (header, entries)
  where
  header = unlines [ "\\subsection{Prism Model Files}"
                   , "At some point we wish to be able to generate prism"
                   , "model files from Pepa model descriptions."
                   , "All of the options described in this section should be"
                   , "considered experimental and not for general use."
                   ]
  entries =
   [ ( unlines [ "\\ipcflag{prism} this is an experimental option to produce"
               , "a prism model. This should not be considered working"
               ]
     , Option ""      [ "prism" ]          ( NoArg CliPrism )
       "Generate a prism model"
     )
   , ( unlines [ "The \\ipcflag{prism-file} option only has an effect if the"
               , "\\texttt{--prism} flag is set. It redirects the output"
               , "prism model to the specified file."
               ]
     , Option ""      [ "prism-file" ]     ( ReqArg CliPrismFileName "FILE" )
       "specify the file to output the prism model to"
     )    
   ]


{-
  Some options for generating dizzy models
-}
dizzyModelSection :: ManualSection a
dizzyModelSection =
  (header, entries)
  where
  header = unlines [ "\\subsection{Dizzy Model Files}"
                   , "At some point we wish to be able to generate dizzy"
                   , "model files from Pepa model descriptions."
                   , "The options in this section describe the control of"
                   , "the \\pepa\\ $\\rightarrow$ dizzy translation."
                   ]
  entries =
   [ ( unlines [ "\\ipcflag{dizzy} this is an experimental option to produce"
               , "a dizzy model. This should not be considered working"
               ]
     , Option ""      [ "dizzy" ]          ( NoArg CliDizzy )
       "Generate a dizzy model"
     )
   , ( unlines [ "The \\ipcflag{dizzy-file} option only has an effect if the"
               , "\\texttt{--dizzy} flag is set. It redirects the output"
               , "dizzy model to the specified file."
               ]
     , Option ""      [ "dizzy-file" ]     ( ReqArg CliDizzyFileName "FILE" )
       "specify the file to output the dizzy model to"
     )    
   ]

{-
  A bunch of miscelleanous entries which don't fit anywhere else.
  Or more rather which have yet to be fitted in elsewhere.
-}
miscelleaneousSection :: ManualSection a
miscelleaneousSection =
  (header, entries)
  where
  header = unlines [ "\\subsection{Miscelleanous Options}"
                   , "This section describes options which don't fit"
                   , "under any of the previous sub-sections."
                   ]
  entries =
   [
     ( ""
     , Option "V"     [ "verbose" ]      ( NoArg CliVerbose )        
       "logfile output to STDOUT"
     )

   , ( ""
     , Option ""      [ "passage" ]       ( NoArg CliPassage )
       "perform a passage-time performance measurement (default)"
     )

   , ( ""
     , Option ""      [ "steady"  ]       ( NoArg CliSteady )
       "perform a steady-state measurement"
     )

   , ( ""
     , Option ""      [ "transient" ]     ( NoArg CliTransient )
       "perform a transient analysis measurement"
     )

   , ( ""
     , Option ""      [ "count-measure" ] ( ReqArg countActions "ACTIONS" )
       "perform a count measure over the given actions"
     )

   , ( ""
     , Option ""      [ "no-measurement" ]  ( NoArg CliNoMeasurement )
       "do not output a performance measurement specification"
     )

   , ( ""
     , Option ""      [ "steady-mean" ]     ( NoArg CliSteadyMean )
       "use the 'mean' estimator in a steady-state measure"
     )
   
   , ( "" 
     , Option ""      [ "steady-variance" ] ( NoArg CliSteadyVariance )
       "use the 'variance' estimator in a steady-state measure"
     )

   , ( ""
     , Option ""      [ "steady-stddev" ]   ( NoArg CliSteadyStdDev )
       "use the 'stddev' estimator in a steady-state measure"
     )

   , ( ""
     , Option ""      [ "steady-distrib" ]  ( NoArg CliSteadyDistrib )
       "use the 'distribution' estimator in a steady-state measure"
     )

   , ( ""
     , Option ""      [ "start-time" ]   ( ReqArg (CliStartTime . read) "TIME" )
       "specify a time at which to start a performance measure eg passage-time"
     )

   , ( ""
     , Option ""      [ "stop-time" ]   ( ReqArg (CliStopTime . read) "TIME" )
       "specify a time at which to stop a performance measure eg passage-time"
     )

   , ( ""
     , Option ""      [ "time-step" ]   ( ReqArg (CliTimeStep . read) "TIME" )
       "specify a the time steps for a performance measurement"
     )

   , ( ""
     , Option ""      [ "solver" ]      ( ReqArg CliSolver "SOLVER" )
       "specify which solution method to use/specify to hydra"
     )

   , ( ""
     , Option ""      [ "rate" ]        ( ReqArg parseRateOpt "DOUBLE" )
       "Override/specify a rate value on the command-line"
     )

   , ( ""
     , Option ""      [ "rename-proc" ]  (ReqArg CliReplaceProcess "P=s" )
       "cause a renaming on the given process within the model"
     )
   
   , ( ""
     , Option ""      [ "rename-rate" ]  (ReqArg CliReplaceRate "r=s" )
       "cause a renaming on the given rate within the model"
     )

   , ( ""
     , Option ""      [ "prioritise" ] ( ReqArg priorityActions "ACTIONS" )
       "increase the priority of the given actions"
     )

   , ( ""
     , Option ""      [ "no-reduce-rate-exps" ] ( NoArg CliNoReduceRateExps )
       "do not reduce the rate expressions"
     )
   
   , ( ""
     , Option ""      [ "hide-non-coop" ]       ( NoArg CliHideNonCoop )
       ("hide any activities which a component performs " ++
         "but does not cooperate over")
     )

   , ( ""
     , Option ""      [ "process-num" ]   (ReqArg (CliProcessNum . read) "NUM" )
       "provide a process number which is used to select rates and processes"
     )

   , ( ""
     , Option ""      [ "show-simplified"] (NoArg CliShowSimplified)
       "Show the simplified model in the comments of the compiled file"
     )
   
   , ( unlines [ "\\ipcflag{fsp} this is an experimental option to produce"
               , "an LTSA model. This should not be considered working"
               ]
     , Option ""      [ "fsp" ]            ( NoArg CliFsp )
       "Generate a finite state processes (FSP) model"
     )
   ]
   where
   countActions = CliCount . (parseCommaSeparatedLowers countError)
   countError   = "error whilst parsing a list of specified count actions"
   
   parseRateOpt :: String -> CliOpt a
   parseRateOpt = CliRate . parseRateOption

   priorityActions :: String -> CliOpt a
   priorityActions = CliPrioritise . (map defaultise) .
                      (parseCommaActionsMaybeAssigns prioritiseErr)
   prioritiseErr   = "error whilst parsing a priority option"
   defaultise :: (ActionIdentifier, Maybe ParsedPriority)
              -> (ActionIdentifier, ParsedPriority)
   defaultise = mapSnd $ maybe defaultIncreasedPriority id



-- In here I should place such things as information that the user should
-- definitely be aware of.
extraInfo :: String
extraInfo = 
    unlines [   ]



{-| 
   The 'System.Console.GetOpt' library returns a triple giving the flags,
   arguments and errors. 
   This data type translates the triple into a more explicit
   representation of the result of getting the program arguments.
-}
data Cli a =
      CliValid [ CliOpt a ] [ CliArg ] 
    | CliInter [ CliOpt a ] [ CliArg ]
    | CliInfo  [ CliOpt a ] [ CliArg ] String 
    | CliError [ CliOpt a ] [ CliArg ] String

{-| The type of arguments to the command-line,
    these will generally be pepa model files 
-}
type CliArg = String

{-|
  Turn the list of command line options into a 'Cli' description.
  Note that the version string may be different for different
  ipc-related programs.
-}
toCli :: Eq a => Bool            --  Whether or not we are in batch mode
      -> CliVersion              --  The version string
      -> String                  --  The program name
      -> [ OptDescr (CliOpt a) ] -- The command-line options descriptions
      -> [ CliArg ]              --  The command line arguments
      -> Cli a
toCli batchMode version progName cliOptions cliArgs = 
   case getOpt order cliOptions cliArgs of
    (options, args, [])
      | elem CliVersion options -> CliInfo  options args versionOutput
      | elem CliHelp    options -> CliInfo  options args helpOutput
      | batchMode && null args  -> CliError options []   inputFileMissing
      | not batchMode           -> CliInter options args
      | otherwise               -> CliValid options args
    (options, args, errs)       -> 
      CliError options args errMsg
      where 
      errMsg = addProgName $ unlines [ unwords errs
                                     , usage
                                     ]
   where
   shortUsg = unlines [ header, info ]
   order    = Permute
   header   = concat [ "Usage: "
                     , progName
                     , " [OPTION...] [PEPAFILE]"
                     ]
   info     = concat [ "Try `"
                     , progName
                     , " --help' for more information."
                     ]
   usage    = unlines [ usageInfo header cliOptions
                      , extraInfo
                      ]

   addProgName :: String -> String
   addProgName s = concat [ progName, ": ", s ]
   inputFileMissing  = addProgName $ "Input file missing \n\n" ++ shortUsg
   versionOutput     = addProgName version
   helpOutput        = addProgName usage


{-| Returns whether the /staunch/ probe is set or not -}
getStaunch :: [ CliOpt a ] -> Bool
getStaunch options =
   not $ null [ True | CliStaunch <- options ]

{-| Returns whether or not static analysis is to be performed on the model -}
getStaticAnalysis :: [ CliOpt a ] -> Bool
getStaticAnalysis options =
   null [ True | CliNoStaticAnalysis <- options ]

{-| Return the probe specifications from a list of @CliOpt@ s -}
getProbeSpecs :: [ CliOpt a ] -> [ String ]
getProbeSpecs = 
   mapMaybe isProbe 
   where
   isProbe :: CliOpt a -> Maybe String
   isProbe ( CliProbeSpec s) = Just s
   isProbe _                 = Nothing

{-| Returns whether or not a master start stop probe should be generated -}
getMasterBool :: [ CliOpt a ] -> Bool
getMasterBool = 
   not . (any isCliNoMaster)


isCliNoMaster :: CliOpt a -> Bool
isCliNoMaster CliNoMasterProbe = True
isCliNoMaster _                = False

{-| Returns whether or not *ipc* should add any measurement
    probes\/specifications.
   Basically it's a bit poor that I have to have this here,
   it sort of underhandedly predicts whether or not ipc will
   add a Master Probe, if not then there is no way to know
   what to measure for.
-}
shouldAddMeasurements :: [ CliOpt a ] -> Bool
shouldAddMeasurements options =
   (getMasterBool options) &&
   any isMeasurementRelated options

isMeasurementRelated :: CliOpt a -> Bool
isMeasurementRelated ( CliProbeSpec _ )    = True
isMeasurementRelated ( CliStartActions _ ) = True
isMeasurementRelated ( CliStopActions _ )  = True
isMeasurementRelated ( CliCount _ )        = True
isMeasurementRelated _                     = False

{-| Returns the list of specified start actions -}
getStartActions :: [ CliOpt a ] -> [ String ]
getStartActions options =
   concat [ a | CliStartActions a <- options ]

{-| Returns the list of specified stop actions -}
getStopActions :: [ CliOpt a ] -> [ String ]
getStopActions options =
   concat [ a | CliStopActions a <- options ]


{-| Returns the list of rate overrides -}
getRateOptions :: [ CliOpt a ] -> [ (RateIdentifier, Double) ]
getRateOptions options =
   [ spec | CliRate spec <- options ]



{-| Returns all the process replacements -}
getProcessRenames :: [ CliOpt a ] -> Either ParseError ProcessRenameMap
getProcessRenames =
   joinParseErrorList . (mapMaybe makeRename)
   where
   makeRename :: CliOpt a -> Maybe (Either ParseError ProcessRename)
   makeRename (CliReplaceProcess s) = 
      Just $ parseProcessRename "command-line" s
      
   makeRename _                     = Nothing

{-| Returns all the rate name replacements -}
getRateRenames :: [ CliOpt a ] -> Either ParseError RateRenameMap
getRateRenames =
   joinParseErrorList . (mapMaybe makeRename)
   where
   makeRename :: CliOpt a -> Maybe (Either ParseError RateRename)
   makeRename (CliReplaceRate s) =
      Just $ parseRateRename "command-line" s
   makeRename _                  = Nothing


{-| Return the list of actions which should be prioritised in the model -}
getPrioritised :: [ CliOpt a ] -> [ (ActionIdentifier, ParsedPriority) ]
getPrioritised options =
   concat [ a | CliPrioritise a <- options ]

{-| Returns whether or not the option to hide all non cooperating
    activities is set
-}
containsHideNonCoop :: [ CliOpt a ] -> Bool
containsHideNonCoop =
   any isHideNonCoop
   where
   isHideNonCoop :: CliOpt a -> Bool
   isHideNonCoop CliHideNonCoop = True
   isHideNonCoop _              = False

{-| Returns the process number option closest to the front of the list -}
getProcessNumber :: [ CliOpt a ] -> Maybe Int
getProcessNumber []                     = Nothing
getProcessNumber (CliProcessNum i : _ ) = Just i
getProcessNumber (_ : rest)             = getProcessNumber rest

{-| Returns the output mod file, based on the command line options and
    the input pepa file. That is if the command line options specify a
    mod-file to output to then use that, otherwise replace the .pepa
    suffix with the .mod suffix
-}
getModFileName :: FilePath -> [ CliOpt a ] -> FilePath
getModFileName pepaFileName options =
   case [ f | CliModFileName f <- options ] ++
        [ f | CliOutput f <- options ] 
   of
      []      -> switchSuffix ".pepa" ".mod" pepaFileName
      [ one ] -> one
      _       -> error "More than one .mod file specified on the command line"

{-| Returns the output prism-file name based on the command line options
   and on the input pepa-file. Exactly equivalent to 
   'getModFileName'
-}
getPrismFileName :: FilePath -> [ CliOpt a ] -> FilePath
getPrismFileName pepaFileName options =
   case [ f | CliOutput f <- options ] ++
        [ f | CliPrismFileName f <- options ] 
   of
      []      -> switchSuffix ".pepa" ".sm" pepaFileName
      [ one ] -> one
      _       -> error "more than one output prism file specified"


{-| Returns the output dizzy-file name based on the command line options
   and on the input pepa-file. Exactly equivalent to 
   'getModFileName'
-}
getDizzyFileName :: FilePath -> [ CliOpt a ] -> FilePath
getDizzyFileName pepaFileName options =
   case [ f | CliOutput f <- options ] ++
        [ f | CliDizzyFileName f <- options ] 
   of
      []      -> switchSuffix ".pepa" ".dizzy" pepaFileName
      [ one ] -> one
      _       -> error "more than one output dizzy file specified"


{-| Gets the generic @--output@ flag's argument if there is one -}
getOutputFileName :: [ CliOpt a ] -> Maybe FilePath
getOutputFileName options =
   case [ name | CliOutput name <- options ] of
      [ one ] -> Just one
      []      -> Nothing
      _       -> 
         error "more than one output file specified on the command-line"

{-|
   Returns a solution control based on the command line options.
-}
getSolutionControl :: [ CliOpt a ] -> DNAMsolutionControl
getSolutionControl options = 
   DNAMsolution { solutionMethod = method }
   where
   method = case [ s | CliSolver s <- options ] of
               []    -> defaultSolutionMethod
               [ s ] -> parseSolutionMethod s
               _     -> error "More than one solution method given"

{-
isCliSolutionControl :: CliOpt a -> Bool
isCliSolutionControl (CliSolver _) = True
isCliSolutionControl _             = False
-}

{-| 
   Returns the performance measurement specifications by examining
   the command-line arguments. The type of this will likely change
   when we make it possible to specify measurements other than a
   single passage-time measurement, for now though this will do.
-}
getMeasurementSpecs :: [ CliOpt a ] -> [ DNAMperformMeasure ]
getMeasurementSpecs options
   | any isCliNoMeasure          options = []
   | not $ shouldAddMeasurements options = []
   | any isCliTransient          options =
      [ DNAMtransient $
        DNAMtransientMeasure { transientSourceCond = DNAMprocpresent runningId
                             , transientTargetCond = DNAMprocpresent stoppedId
                             , transientStartTime  = startTime
                             , transientStopTime   = stopTime
                             , transientStepTime   = timeStep
                             }
      ]
   | any isCliSteady options       =
      [ DNAMsteady $
        DNAMsteadyMeasure { steadyMeasureName = Unqualified "steady_measure"
                          , steadyEstimator   = estimator
                          , steadyCondition   = DNAMprocpresent runningId
                          }
      ]
   | not $ null countActions       =
      [ DNAMcount $
        DNAMcountMeasure { countMeasureName = Unqualified "count_measure"
                         , countTransitions = countActions }
      ]
   | otherwise                     =
      -- The default is a passage time measurement
      [ DNAMpassage $
        DNAMpassageMeasure { passageSourceCond = DNAMprocpresent runningId
                           , passageTargetCond = DNAMprocpresent stoppedId
                           , passageStartTime  = startTime
                           , passageStopTime   = stopTime
                           , passageStepTime   = timeStep
                           }
      ]
   where
   -- Careful, these must match up with those defined in
   -- Language.Pepa.Probes.AddProbes, in fact that should
   -- really export them, the point is though I think that
   -- they should be 'specifiable' on the command-line, so
   -- they shouldn't really be defined there.
   stoppedId       = Unqualified "ProbeStopped"
   runningId       = Unqualified "ProbeRunning"

   -- If none of the time options are specified we should have defaults of
   -- start-time = 1.0
   -- stop-time  = 10.0
   -- time-step  = 0.5
   startTime       = case find isCliStartTime options of
                        Just (CliStartTime d)  -> d
                        Just _                 -> error weirdError
                        Nothing                -> 1.0
   stopTime        = case find isCliStopTime options of
                        Just (CliStopTime d)   -> d
                        Just _                 -> error weirdError
                        Nothing                -> 10.0 * startTime
   timeStep        = case find isCliTimeStep options of
                        Just (CliTimeStep d)   -> d
                        Just _                 -> error weirdError
                        Nothing                -> defaultTimeStep
   defaultTimeStep = (stopTime - startTime) / 18.0

   countActions    = map Unqualified $ concat [ s | CliCount s <- options ]

   estimator
      | any isCliSteadyMean     options = EstimatorMean
      | any isCliSteadyVariance options = EstimatorVariance
      | any isCliSteadyStdDev   options = EstimatorStdDeviation
      | any isCliSteadyDistrib  options = EstimatorDistribution
      | otherwise                       = EstimatorMean

   weirdError      = unwords [ "Cli.getMeasurementSpecs:"
                             , "strange error we have searched for"
                             , "a given kind of options but the one"
                             , "found was not of the correct kind"
                             , "I am flabbergasted if this error comes up"
                             , "please do mail me."
                             ]
                             


isCliStartTime :: CliOpt a -> Bool
isCliStartTime (CliStartTime _ ) = True
isCliStartTime _                 = False

isCliStopTime :: CliOpt a -> Bool
isCliStopTime (CliStopTime _ )   = True
isCliStopTime _                  = False

isCliTimeStep :: CliOpt a -> Bool
isCliTimeStep (CliTimeStep _ )   = True
isCliTimeStep _                  = False


isCliSteadyMean :: CliOpt a -> Bool
isCliSteadyMean CliSteadyMean = True
isCliSteadyMean _             = False

isCliSteadyVariance :: CliOpt a -> Bool
isCliSteadyVariance CliSteadyVariance = True
isCliSteadyVariance _                 = False

isCliSteadyStdDev :: CliOpt a -> Bool
isCliSteadyStdDev CliSteadyStdDev = True
isCliSteadyStdDev _               = False

isCliSteadyDistrib :: CliOpt a -> Bool
isCliSteadyDistrib CliSteadyDistrib = True
isCliSteadyDistrib _                = False

{-|
   The data type returned to indicate which kind of performance
   measurements are to be performed on the model which also indicates
   what should be run after hydra.
-}
data PostHydra =
   PostHydraPassage
 | PostHydraTransient
 | PostHydraSteady
 | PostHydraCount
 | PostHydraNone


{-|
   Returns the kind of analysis that the command-line options specify
   to be performed on the model.
-}
getPostHydra :: [ CliOpt a ] -> PostHydra
getPostHydra options
   | any isCliTransient options = PostHydraTransient
   | any isCliSteady    options = PostHydraSteady
   | any isCliCount     options = PostHydraCount
   | any isCliNoMeasure options = PostHydraNone
   | otherwise                  = PostHydraPassage


{-| Returns whether or not the command-line options specify that we
    should run hydra after producing the .mod file.
-}
shouldRunHydra :: [ CliOpt a ] -> Bool
shouldRunHydra =
   any suggestsRunHydra
   where
   suggestsRunHydra :: CliOpt a -> Bool
   suggestsRunHydra (CliHydra _)  = True
   suggestsRunHydra (CliRunHydra) = True
   suggestsRunHydra _             = False

{-| Returns the two hydra commands which may have had their paths
   specified on the command-line
-}
getHydraCommands :: [ CliOpt a ] -> (FilePath, FilePath)
getHydraCommands options =
   case [ file | CliHydra file <- options ] of
      [ ]     -> ("hydra-s", "hydra-uniform")
      [ one ] -> (one, switchBaseName one "hydra-uniform")
      _       -> error "--hydra specified more than once"


isCliTransient :: CliOpt a -> Bool
isCliTransient CliTransient = True
isCliTransient _            = False

isCliSteady :: CliOpt a -> Bool
isCliSteady CliSteady = True
isCliSteady _         = False

isCliCount :: CliOpt a -> Bool
isCliCount (CliCount _) = True
isCliCount _            = False

isCliNoMeasure :: CliOpt a -> Bool
isCliNoMeasure CliNoMeasurement = True
isCliNoMeasure _                = False


{-| Returns true if the output is to go to standard out -}
getStandardOut :: [ CliOpt a ] -> Bool
getStandardOut options = not $ null [ True | CliStandardOut <- options ]


{-| Returns all the non-standard arguments, these are those which must have
    been added by the individual program.
-}
getNonStandard :: [ CliOpt a ] -> [ a ]
getNonStandard []                          = []
getNonStandard ( CliNonStandard a : rest ) = a : (getNonStandard rest)
getNonStandard ( _ : rest )                = getNonStandard rest


