{-|
   This is the main module of the @ipc@ compiler, the flagship
   application of the @ipclib@ library.
-}

module Main
   ( main )
where

{- Standard library modules imported -}
import System.Cmd
   ( system )
import System.Console.GetOpt
   ( OptDescr  ( .. ) )
import System.Exit
   ( ExitCode       ( .. )
   , exitFailure
   )
import System.IO
   ( stderr
   , hPutStrLn
   )
{- External Library modules imported -}
{- Local modules imported -}
import Ipc.Cli
   ( getCliArgs
   , Cli                  ( .. )
   , CliOpt               ( .. )
   , toCli
   , baseCliOptions

   , getStaticAnalysis
   , getStaunch

   , getProbeSpecs
   , getMasterBool
   , getStartActions
   , getStopActions

   , getRateOptions
   , getPrioritised

   , getModFileName
   , getPrismFileName
   , getDizzyFileName
   , getMeasurementSpecs
   , getSolutionControl
   , PostHydra            ( .. )
   , getPostHydra

   , shouldRunHydra
   , getHydraCommands

   , getStandardOut
   )
import Language.Pepa.Syntax
   ( ParsedModel
   , RateIdentifier
   , RateSpec
   , ParsedRateExp        ( .. )
   )
import Language.Pepa.Parser
    ( parseFile
    , pepaFile
    )
import Language.Pepa.Print
    ( hPrintPepaModel
    , HumanPrint      ( .. )
    )
import Language.Pepa.Transform.Simplify
   ( simplify )
import Language.Pepa.Transform.Replace
   ( overrideRateDefinitionList )
import Language.Pepa.Analysis.Analysis
   ( analyseModel
   , splitMessages
   , messagesOfReport
   )

import Language.Pepa.Utils
   ( removeSuffix
   , switchSuffix
   , addLeadingDotSlash
   , mapSnd
   )

{- Performance specification probes -}
import Language.Pepa.Probes.ProbeParser
    ( parseProbe )
import Language.Pepa.Probes.AddProbes
   ( addMasterProbe
   , addProbeDefs
   , ProbeTranslateFlags     ( .. )
   )

{- Hydra Imports -}
import Language.Pepa.Compile.Hydra
   ( translateTransition
   , dnamFromPepa
   , addMeasurementSpecs
   )
import Language.Hydra.Transform
   ( translateRates
   , prioritiseTrans
   )
import Language.Hydra.Print
   ( printDnamModelFile )
{- Prism imports -}
import Language.Pepa.Compile.Prism
   ( prismFromPepa
   , makePrismFile
   )
import Language.Prism.Print
   ( printPrismModelFile )
{- Fsp Imports -}
import Language.Pepa.Compile.Fsp
   ( pepaModelToFspModel )
import Language.Fsp.Print
   ( printFspModel )

import Language.Pepa.IpcControl
   ( MainControl
   , closeOrError
   , eitherToMainControl
   , resultWarning
   , resultError
   )

-- import Language.Prism.Print
--    ( printPrismModel )
{- End of module imports -}


main :: IO ()
main = do putStrLn ipcBanner
          args         <- getCliArgs
          processArgs args

-- Processing of the command-line arguments.
processArgs :: [ String ] -> IO () 
processArgs cliArgs =
   case toCli True ipcVersion "ipc" ipcOptions cliArgs of
     CliValid options [ inputFile ] ->
        compileFile options inputFile
     CliValid _options []           ->
         putStrLn "sorry but you must specify some input file"
     CliValid _options _            ->
         putStrLn "sorry but currently ipc only accepts one input file"
     CliInter _ _                   ->   
         putStrLn "ipc cannot (yet) be run in interactive mode"
     CliInfo  _ _ infoString        -> 
         putStrLn infoString
     CliError  _ _ errorString      -> 
         putStrLn errorString


ipcVersion :: String
ipcVersion = "0.02"

ipcBanner :: String
ipcBanner = unlines $ zipLines extraBanner ipcBannerInfo

ipcBannerInfo :: [ String ]
ipcBannerInfo =
   [ "** The Imperial PEPA compiler:"
   , "ipc v" ++ ipcVersion
   , "Produced on location in/between Edinburgh, "
     ++ "Leeds, London and Ludlow, 2002-7"
   , "Principal developers: "
     ++ "Jeremy Bradley, Allan Clark & Stephen Gilmore"
   , "Queries or problems: "
     ++ " jb@doc.ic.ac.uk, stg@inf.ed.ac.uk or a.d.clark@ed.ac.uk"
   ]

extraBanner :: [ String ]
extraBanner =
   [ "|\\      /|"
   , "| \\    / |"
   , "|  \\  /  |"
   , "|   \\/   |"
   , "|   <>   |"
   , "|   /\\   |"
   , "|  /  \\  |"
   , "| /    \\ |"
   , "|/      \\|"
   ]


{-
   This is suppose to put to sets of strings side by side.
   It actually doesn't work too well unless the first set
   is longer than the second set, and the first set are all
   the same length. But that's exactly what we have here.
   Might make an interesting function to get completely right
   though.
-}
zipLines :: [ String ] -> [ String ] -> [ String ]
zipLines l  []                     = l -- so zipLines [] [] = []
zipLines [] l                      = l
zipLines (a1 : rest1) (a2 : rest2) = 
   (a1 ++ "  " ++ a2) : (zipLines rest1 rest2)

{-
oldBanner =
   unlines [ "|\\          /|"
           , "| \\        / |"
           , "|  \\      /  |"
           , "|   \\    /   |"
           , "|    \\  /    |"
           , "|     \\/     |"
           , "|     <>     |"
           , "|     /\\     |"
           , "|    /  \\    |"
           , "|   /    \\   |"
           , "|  /      \\  |"
           , "| /        \\ |"
           , "|/          \\|"
           ]
-}

ipcOptions :: [ OptDescr ( CliOpt () ) ]
ipcOptions = 
   baseCliOptions


-- We still need to add the probes !!

-- Shouldn't exit but return an 'ExitCode' and then
-- for multiple input files we can combine these.
-- Obviously also should do the logging.
compileFile :: [ CliOpt a ] -> FilePath -> IO ()
compileFile options file =
   do parseResult <- parseFile pepaFile file
      processModel parseResult
   where
   processModel :: MainControl ParsedModel -> IO ()
   processModel mcModel
      | any isFspOption options   = compileForFsp   options mcModel
      | any isPrismOption options = compileForPrism file options mcModel
      | otherwise                 = compileForHydra file options mcModel
            

{- Compiles the model to a hydra model.
   Notice there appears to be a little support for printing the model
   out as a Prism model, however I now fear that it is not quite such
   a simple compilation mechanism. I think I will now be better off
   doing a separate compilation to Prism models unfortunately, therefore
   we will eventually have a @compileForPrism@ function.
-}
compileForHydra :: FilePath -> [ CliOpt a ] -> MainControl ParsedModel -> IO ()
compileForHydra pepaFileName options mcModel =
   case closeOrError result of
      Right (a, [], _)     -> writeModFile a
      Left err             -> 
         do hPutStrLn stderr errorMessage
            hPutStrLn stderr errorSuggest
            hPutStrLn stderr "--------------------------"
            hPutStrLn stderr err
            exitFailure
      Right (a, warns, _)
         | getStaunch options ->
            do hPutStrLn stderr staunchWarn
               hPutStrLn stderr "--------------------------"
               hPutStrLn stderr $ unlines warns
               writeModFile a
               hPutStrLn stderr staunchWarn
         | otherwise          ->
            do hPutStrLn stderr staunchSuggest
               hPutStrLn stderr "--------------------------"
               hPutStrLn stderr $ unlines warns
            
   where
   staunchWarn    = "There were some warnings - attempting to compile anyway"
   staunchSuggest = "There were some warnings - compilation cancelled" ++
                    " - to compile anyway use --staunch"
   errorMessage   = "There were some error, compilation cancelled"
   errorSuggest   = "To attempt to compile anyway use --no-static-analysis"

   result = createHydraModelString options mcModel

   writeModFile :: String -> IO ()
   writeModFile contents
      | getStandardOut options = putStrLn contents
      | otherwise              = do writeFile modFile contents
                                    runHydra

   modFile = getModFileName pepaFileName options

   -- We should NOT be ignoring the return code (ExitSuccess/Failure)
   runHydra :: IO ()
   runHydra
      | shouldRunHydra options = invokeHydra
      | otherwise              = return ()

   invokeHydra   = do system $ unwords [ hydra, modFile ]
                      postHydra
                      return () 

   (hydra, hydraUniform) = getHydraCommands options

   postHydra     = case getPostHydra options of
                      PostHydraPassage   -> system uniformCommand
                      PostHydraTransient -> system uniformCommand
                      PostHydraSteady    -> system perfCommand
                      PostHydraCount     -> system perfCommand
                      PostHydraNone      -> return ExitSuccess

   -- These command names should be gotten from calls to the Cli
   -- module, as in: getUniformCommand etc because their exact
   -- form may be changed (eg we might wish to allow the user to
   -- NOT specify the @-cdf@ option).
   uniformCommand  = unwords [ hydraUniform
                             , filePrefix
                             , " -cdf"
                             ]
   perfCommand     = unwords [ perfProgram, filePrefix ]
   filePrefix      = removeSuffix ".mod" modFile
   perfProgram     = addLeadingDotSlash $ switchSuffix ".mod" ".perf" modFile

createHydraModelString :: [ CliOpt a ] 
                       -> MainControl ParsedModel 
                       -> MainControl String
createHydraModelString options mcModel =
   -- 'Language.Pepa.Transform.Simplify' should really return
   -- a 'MainControl'
   do parsedModel1    <- mcModel
      let parsedModel2 = overrideRateDefinitionList parsedModel1 rateOverrides
      -- Analyse the model for statically checked errors
      parsedModel3    <- typeCheckModel options parsedModel2
      parsedModel     <- addProbesToModel options parsedModel3
          -- Simplify the model
      let simpModel   =  simplify options parsedModel
          -- Translate to a hydra model containing apparent rates
          appHyModel  =  dnamFromPepa simpModel
          -- Translate the apparent rates to dnam speeds
          hModel1     = translateRates transTrans appHyModel
          hModel      = prioritiseTrans (getPrioritised options) hModel1
          hModelFile  = addMeasurementSpecs solutionControl measures hModel
          header      = unlines [ ipcBanner
                                , "This file was generated by ipc"
                                , "from the following pepa model"
                                , hPrintPepaModel parsedModel
                                , extraHeader
                                ]
          extraHeader = if showSimplified
                           then unlines [ topExtraHeader
                                        , hPrintPepaModel simpModel
                                        ]
                           else ""
          topExtraHeader = "This is the simplified pepa Model"
      return $ printDnamModelFile header hModelFile
   where
   transTrans      = translateTransition reduceRates
   reduceRates     = not $ any isNoReduceRates options
   
   showSimplified  = any isShowSimplified options
   measures        = getMeasurementSpecs  options
   solutionControl = getSolutionControl   options

   rateOverrides   = map makeRateDef $ getRateOptions options
   
   makeRateDef :: (RateIdentifier, Double) -> RateSpec
   makeRateDef = mapSnd RateNum



addProbesToModel :: [ CliOpt a ] 
                 -> ParsedModel
                 -> MainControl ParsedModel
addProbesToModel options model =
   do probes      <- mapM ( eitherToMainControl . parseProbe ) probeStrings
      let mPmodel =  addProbeDefs trFlags probes model
      probedModel <- eitherToMainControl mPmodel
      if addMaster
         then return $ addMasterProbe startActs stopActs probedModel
         else return probedModel
   where
   -- Obviously we do not add a master probe if the '--no-master' option
   -- is set, however we also do not add a master probe if
   -- there are no start/stop actions  and no probe specifications, 
   -- since in that case we assume that the user is
   -- just wanting to translate the unchanged model.
   -- We do not however need to check for probe strings since the
   -- start and stop actions will only be null if probe strings is
   -- as well
   addMaster    = ( getMasterBool options )
                  && ( (not $ null startActs)
                       || ( not $ null stopActs ))

   probeStrings = getProbeSpecs   options

   -- If the user specifies a probe specification they need not
   -- specify start and stop actions since we can assume that the
   -- probe performs the 'start' and 'stop' communication actions.
   -- They user can though of course override the default 'start'
   -- and 'stop' if they wish something else (for example if they
   -- have no probe specification).
   startActs    = case getStartActions options of
                     [] -> if null probeStrings
                           then []
                           else [ "start" ]
                     l  -> l
   stopActs     = case getStopActions  options of
                     [] -> if null probeStrings
                           then []
                           else [ "stop" ]
                     l  -> l
   trFlags      = [ MinSelfLoops ]



compileForPrism :: FilePath -> [ CliOpt a ] -> MainControl ParsedModel -> IO ()
compileForPrism pepaFileName options mcModel =
   case closeOrError result of
      Right (a, [], _)     -> writePrismFile a
      Left err             -> 
         do hPutStrLn stderr err
            exitFailure
      Right (a, _warns, _) ->
         do hPutStrLn stderr "There were some errors"
            writePrismFile a
            hPutStrLn stderr "There were some errors"
   where
   result    = createPrismModelString options mcModel
   prismFile = getPrismFileName pepaFileName options

   writePrismFile :: String -> IO ()
   writePrismFile contents
      | getStandardOut options = putStrLn contents
      | otherwise              = writeFile prismFile contents

   

createPrismModelString :: [ CliOpt a ] 
                       -> MainControl ParsedModel
                       -> MainControl String
createPrismModelString options mcModel =
   do parsedModel1    <- mcModel
      parsedModel     <- addProbesToModel options parsedModel1
          -- Simplify the model
      let simpModel      = simplify options parsedModel
          -- translate to a prism model
          prismModel     = prismFromPepa reduceRates simpModel
          prismModelFile = makePrismFile prismModel
      return $ printPrismModelFile prismModelFile 
   where
   reduceRates     = not $ any isNoReduceRates options

compileForDizzy :: FilePath -> [ CliOpt a ] -> MainControl ParsedModel -> IO ()
compileForDizzy pepaFileName options mcModel =
   case closeOrError result of
      Right (a, [], _)     -> writeDizzyFile a
      Left err             -> 
         do hPutStrLn stderr err
            exitFailure
      Right (a, _warns, _) ->
         do hPutStrLn stderr "There were some errors"
            writeDizzyFile a
            hPutStrLn stderr "There were some errors"
   where
   result    = createDizzyModelString options mcModel
   dizzyFile = getDizzyFileName pepaFileName options

   writeDizzyFile :: String -> IO ()
   writeDizzyFile contents
      | getStandardOut options = putStrLn contents
      | otherwise              = writeFile dizzyFile contents

   

createDizzyModelString :: [ CliOpt a ] 
                       -> MainControl ParsedModel
                       -> MainControl String
createDizzyModelString options mcModel =
   do parsedModel1    <- mcModel
      parsedModel     <- addProbesToModel options parsedModel1
          -- Simplify the model
      let simpModel      = simplify options parsedModel
          -- translate to a dizzy model
          dizzyModel     = undefined
          dizzyModelFile = ""
      return dizzyModelFile
   where
   reduceRates     = not $ any isNoReduceRates options


compileForFsp :: [ CliOpt a ] -> MainControl ParsedModel -> IO ()
compileForFsp options mcModel =
   exitFromMainControl result
   where
   result = do parsedModel <- mcModel
               -- Simplify the model making sure that the option
               -- to hide all non-cooperating activities is set
               -- since this is important for the transformation
               --- to fsp.
               let newOptions  = CliHideNonCoop : options
                   simpModel   = simplify newOptions parsedModel

                   -- translate the model to fsp, pretty simple now that
                   -- we do not have cooperating components performing
                   -- the same activity unless they do cooperate over it.
                   fspModel    = pepaModelToFspModel simpModel
               return $ printFspModel fspModel

typeCheckModel :: [ CliOpt a ] -> ParsedModel -> MainControl ParsedModel
typeCheckModel options pModel
   | getStaticAnalysis options =
      case splitMessages $ messagesOfReport $ analyseModel pModel of
         ([], warns) -> resultWarning pModel $ map hprint warns
         (errs, _)   -> resultError $ unlines (map hprint errs)
   | otherwise                 = return pModel


{- Utility functions over command line options -}
isNoReduceRates :: CliOpt a -> Bool
isNoReduceRates CliNoReduceRateExps = True
isNoReduceRates _                   = False


isShowSimplified :: CliOpt a -> Bool
isShowSimplified CliShowSimplified = True
isShowSimplified _                 = False

isPrismOption :: CliOpt a -> Bool
isPrismOption CliPrism = True
isPrismOption _        = False

isFspOption :: CliOpt a -> Bool
isFspOption CliFsp = True
isFspOption _      = False

exitFromMainControl :: MainControl String -> IO ()
exitFromMainControl result =
   case closeOrError result of
      Right (a, [], _)     -> putStrLn a
      Left err             -> 
         do hPutStrLn stderr err
            exitFailure
      Right (a, _warns, _) ->
         do putStrLn "There were some errors"
            putStrLn a
            putStrLn "There were some errors"


{-
   putStrLn $ printDnamModel dnamModel
   where
   dnamModel = fromPepaComponent 
toGenericParallel 
-}