{-|
   A small module exporting a small number of transformations over hydra
   models.
-}
module Language.Hydra.Transform
   ( translateRates
   , uniquifyTransitions
   , prioritiseTrans
   )
where

{- Imported Standard Libraries -}
import Data.List
  ( find )
{- Imported Local Libraries -}
import Language.Hydra.Syntax
   ( DNAMmodel             ( .. )
   , DNAMtransition        ( .. )
   )
import Language.Pepa.Syntax
   ( prefixQName
   , ActionIdentifier
   , ParsedPriority
   , nameOfAction
   , ShowOrig      ( .. )
   , ParsedAction
   )

{- End of Imports -}

{- Translation of a hydra model translating rates -}

{-| Translates the rates in a model according to the given function
    over transitions
-}
translateRates :: (DNAMtransition a b -> DNAMtransition a c)
               -> DNAMmodel a b
               -> DNAMmodel a c
translateRates translate dModel =
   DNAMmodel { modelHeaders     = modelHeaders     dModel 
             , modelStateVector = modelStateVector dModel
             , modelConstants   = modelConstants   dModel
             , modelInitial     = modelInitial     dModel
             , modelTransitions = transitions
             , modelInvariants  = modelInvariants  dModel
             }
   where
   transitions = map translate $ modelTransitions dModel

{-| Makes each transition name unique
   Essentially I believe we need this in order to
   avoid transition name clashes for example in:
   @ P = (a, r).P + (a, r).P @
-}
uniquifyTransitions :: DNAMmodel a b -> DNAMmodel a b
uniquifyTransitions model = 
   model { modelTransitions = newTrans }
   where
   newTrans = map uniqueTrans $ zip [0..] (modelTransitions model)

   uniqueTrans :: (Int, DNAMtransition a b) -> DNAMtransition a b
   uniqueTrans (i, trans) =
      trans { transName = prefixQName (transName trans) $ show i }


{-| Prioritises the given list of action names in the given model -}
prioritiseTrans :: [ (ActionIdentifier, ParsedPriority) ]
                    -- ^ List priorities to apply
                -> DNAMmodel ParsedAction b -- ^ Input model
                -> DNAMmodel ParsedAction b -- ^ The returned model
prioritiseTrans pActions model =
   model { modelTransitions = transitions }
   where
   transitions = map prioritiseTransition $ modelTransitions model

   -- Notice that we check the original names, this means that action
   -- will be prioritised even if it is hidden within the model.
   -- (hmm, actually not entirely sure that's true)
   prioritiseTransition :: DNAMtransition ParsedAction b 
                        -> DNAMtransition ParsedAction b
   prioritiseTransition trans =
     case find isAppropriate pActions of
       Just (_, p)   -> trans { transPriority = p }
       Nothing       -> trans
      where 
      isAppropriate :: (ActionIdentifier, ParsedPriority) -> Bool
      isAppropriate (name, _) = (showOrig name) == actionName
      actionName    = showOrig $ nameOfAction $ transActionKind trans
   