{-|
  This module represents the implementation of analysis over pepa
  models. These should be static analyses and hence should operate over
  the full parsed syntax, although it may be convenient to do some desugaring
  prior to the actual analysis.
-}
module Language.Pepa.Analysis.Analysis 
   ( AnalysisReport      ( .. )
   , messagesOfReport
   , splitMessages
   , analyseModel
   
   -- * Results of individual analyses
   , AnalysisMessage     ( .. )
   , AnalysisResult      ( .. )
   , AliasMap
   
   -- * Individual analysis functions
   , possibleSuccessors
   , possibleActionsNonAliased
   
   , ConcentrationMap
   , getInitialConcentrations
   )
where

{- Standard Modules Imported -}
import Prelude hiding 
    ( lookup )
import Data.List
    ( nub
    , (\\)
    , intersect
    , union
    , partition
    , find
    )
import Data.Map
   ( Map
   , keys
   )
import Data.Maybe 
    ( fromMaybe 
    , maybeToList
    )

{- Local Modules Imported -}
import Language.Pepa.Utils
    ( mapFst
    , mapSnd
    )
import Language.Pepa.Analysis.FixMap
    ( FixMap
    , empty
    , insertWith
    , insertList
    , lookup
    , toMap
    , fixPoint
    , toList
    )
import Language.Pepa.Syntax
    ( ParsedModel          ( .. )
    , ProcessDef
    , ParsedComponent      ( .. )
    , BeCondition          ( .. )
    , ParsedComponentId
    , RateSpec
    , RateIdentifier
    , ActionIdentifier
    , ParsedTrans          ( .. )
    , ParsedAction         ( .. )
    , ParsedRate           ( .. )
    , ParsedRateExp        ( .. )
    )

import Language.Pepa.PepaUtils
    ( findProcessDefinition )
import Language.Pepa.Analysis.Typing
    ( typeModel )
{- End of Imports -}

{-
  List of analyses I should add:
  One that checks if an action escapes from the main system equation
  with rate Top. This should be quite simple just recurse through the
  system and check if a component does an action at rate Top.
  If it is a cooperation then lose those actions which cooperate and which
  one side of the component does not do at rate Top.

  One that checks for zero rates, I actually think it should incorporate
  the command line rate options for this, in that it should take in a list
  of overrides possible and check if the rates are *ever* zero,
  this would be very good for condor related goodness.
-}


{-| The type of a static analysis warning or error, a list of such
    analysis messages is the result of many of the static analyses, where
    by an empty list returned indicates that the model passed the analysis.
    The idea is that all of the messages are of a single type, but they may
    be wrapped in 'AnalysisError' or 'AnalysisWarning' as per the default.
    Then any caller is free to write a transformation function which, 
    for example, turns all warnings into errors, or only a subset of
    the warnings into errors.
-}
data AnalysisMessage = AnalysisError   AnalysisResult
                     | AnalysisWarning AnalysisResult

{-| The type of information returned from an analysis, 
    which should be packaged up
    in the 'AnalysisMessage' type to indicate whether, by default, 
    it is an error or a warning.
-}
data AnalysisResult = 
    -- | The process cooperates over the first set of actions
    -- but neither side of the cooperation actually performs
    -- the second set of actions (perhaps because they are the same
    -- for example in a proces array)
    ExcessActions ParsedComponent [ ParsedAction ] [ ParsedAction ]
    -- | The process is a parallel cooperation which cooperates 
    -- over the first set of actions but the left hand side never
    -- actually performs the second set
    | ExcessActionsLeft ParsedComponent [ ParsedAction ] [ ParsedAction ]
    -- | The process is a parallel cooperates which cooperates
    -- over the first set of actions but the right hand side never
    -- actually performs the second set of actions.
    | ExcessActionsRight ParsedComponent [ ParsedAction ] [ ParsedAction ]
    -- | The process hides the first set of actions but does
    -- does not actually perform the second set
    | NeedlessActionsHide ParsedComponent [ ParsedAction ] [ ParsedAction ]
    -- | The process identifiers which are used but are not defined
    | UsedButUndefinedProcesses  [ ParsedComponentId ]
    -- | The process identifiers which are defined but not used
    | DefinedButNotUsedProcesses [ ParsedComponentId ]
    -- | The process identifiers which are defined more than once
    | DefinedMultiplyProcesses   [ ParsedComponentId ]
    -- | An alias loop such as @ P = P @
    | AliasLoop [ ParsedComponentId ]
    -- | The main system composition contains a self-loop over the given actions
    | SystemSelfLoop ParsedComponent [ ParsedAction ]
    -- | Rates are defined but not used
    | DefinedButNotUsedRates  [ RateIdentifier ]
    -- | Rates which are used but not defined
    | UsedButUndefinedRates   [ RateIdentifier ]
    -- | Rates which are defined more than once
    | DefinedMultiplyRates    [ RateIdentifier ]
    -- | A generic typing error, should probably change this.
    | TypingError             String

{-| The type of the result of analysing an entire module.
    This is essentially a large record type.
    The individual analyses should not be required to be exported since
    a programmer can just analyse the whole module but then only examine 
    the result of the analyses they are interested in, due to lazy evaluation
    the ignored analyses will not actually be performed.
-}
data AnalysisReport = 
    AnalysisReport
        { -- | Does the parsed model contain process arrays?
          processArraysPresent :: Bool 
          -- | A mapping from process identifiers to the actions they perform.
        , processActionsMap    :: Map ParsedComponentId [ ParsedAction ]
          -- | A list of actions performed (and not hidden) by
          -- the main composition.
        , compositionActions   :: [ ParsedAction ]
          -- | A list of reports on excess actions in cooperations 
          -- involved in the given pepa model
        , excessActionsReport  :: [ AnalysisMessage ]
          -- | A list of the process names defined within the model
        , definedProcessNames  :: [ ParsedComponentId ]
          -- | A list of process names used within the model
        , usedProcessNames     :: [ ParsedComponentId ]
          -- | Results of analysing the defined and used process names
        , definedNamesReport   :: [ AnalysisMessage ]
          -- | An alias database mapping process names to those names
          -- they can become without performing any actions.
        , aliasDataBase        :: AliasMap
          -- | Results of checking for alias loops
        , aliasLoops           :: [ AnalysisMessage ]
          -- | Results of analysing for self-loops
        , selfLoopsReport      :: [ AnalysisMessage ]
          -- | The used action names
        , usedActionNames      :: [ ActionIdentifier ]
          -- | Rate names defined
        , definedRateNames     :: [ RateIdentifier ]
          -- | Used rate names
        , usedRateNames        :: [ RateIdentifier ]
          -- | Results of analysing the rate names defined\/used.
        , rateNamesReport      :: [ AnalysisMessage ]
          -- | Results of the typing phase
        , typingErrors         :: [ AnalysisMessage ]
        }

{-| Returns all the warnings and errors, that is all the
    'AnalysisMessage's from an 'AnalysisReport'.
    This is useful for applications which simply wish to ask
    the question "Does the model pass static analysis?"
    If the list returned from this function is @[]@
-}
messagesOfReport :: AnalysisReport -> [ AnalysisMessage ]
messagesOfReport aReport =
    concat [ excessActionsReport aReport
           , definedNamesReport  aReport
           , aliasLoops          aReport
           , selfLoopsReport     aReport
           , rateNamesReport     aReport
           , typingErrors        aReport
           ]

{-| Splits a list of messages into two lists of 'AnalysisResult's,
    the first of a list of errors the second a list of warnings.
-}
splitMessages :: [ AnalysisMessage ] -> ( [ AnalysisResult ]
                                        , [ AnalysisResult ] )
splitMessages []                         = ( [], [] )
splitMessages (AnalysisError m : rest)   =
    mapFst (m : ) $ splitMessages rest
splitMessages (AnalysisWarning m : rest) =
    mapSnd (m : ) $ splitMessages rest

{-| Performs all of the analyses on the model and returns an
    'AnalysisReport'
-}
analyseModel :: ParsedModel -> AnalysisReport
analyseModel model =
    AnalysisReport 
     { processArraysPresent = containsProcessArrays model 
     , processActionsMap    = processActionsMapping
     , compositionActions   = compActions
     , excessActionsReport  = excessActReports
     , definedProcessNames  = processIdentsDefined
     , usedProcessNames     = processNamesUsed
     , definedNamesReport   = analyseDefinedNames processIdentsDefined 
                                  processNamesUsed
     , aliasDataBase        = aliasMap
     , aliasLoops           = maybeToList $ checkForAliasLoops aliasMap
     , selfLoopsReport      = maybeToList $ selfLoopsModel aliasMap model
     , usedActionNames      = nub $ keys processActionsMapping
     , definedRateNames     = rateNamesDefined
     , usedRateNames        = rateNamesUsed
     , rateNamesReport      = analyseDefinedRates rateNamesDefined 
                                  rateNamesUsed
     , typingErrors         = typingResults
     }
    where
    processActionsMapping = toMap processActionDb
    processActionDb       = processActions model
    compActions           = fst excessActResults
    excessActReports      = snd excessActResults
    excessActResults      = excessActions model processActionDb
    processIdentsDefined  = definedProcesses model
    processNamesUsed      = usedProcesses model
    rateSpecifications    = ratesDefinedModel model
    aliasMap              = buildAliasDb model
    rateNamesUsed         = ratesUsedModel model
    rateNamesDefined      = map fst rateSpecifications
    typingResults         = map (AnalysisError . TypingError) $ typeModel model

{-|
  A very simple function to determine whether or not a 'ParsedModel'
  contains an array of processes construct. This is mostly useful for
  the @ipci@ program to decide whether or not to ask the user if
  we should aggregate or not, if there are no array of processes then
  it doesn't make any difference and we therefore do not need to ask.
  Also it could be used by both @ipc@ and @ipci@ to give a warning
  if @--aggregate@ (or indeed the @--noaggregate @)
  flag is given on the command line but no array of processes are used.
-}
containsProcessArrays :: ParsedModel -> Bool
containsProcessArrays (ParsedModel _rDefs pDefs pComp) =
    containsComp pComp || any containsDef pDefs
    where
    containsComp :: ParsedComponent -> Bool
    containsComp (ProcessArray _ _ _)       = True
    containsComp (StopProcess)              = False
    containsComp (IdProcess _)              = False
    containsComp (Hiding par _)             = containsComp par
    containsComp (Cooperation left _ right) = containsComp left ||
                                              containsComp right
    containsComp (ComponentSum left right)  = containsComp left ||
                                              containsComp right
    containsComp (PrefixComponent _ right)  = containsComp right
    containsComp (CondBehaviour _ right)    = containsComp right
    
    containsDef :: ProcessDef -> Bool
    containsDef = containsComp . snd

{-
  This analysis determines which actions each of the processes within
  a model may perform. Using the results of such an analysis we can,
  for example, determine whether a cooperation synchronises on a spurious
  action, for example
  @
  P = (a, 1.0).P
  Q = (a, 1.0).Q
  ( P <b> Q )
  @
  since neither process can perform the action @b@ the user should be
  warned about this.
-}
type ProcessActionDb = FixMap ParsedComponentId [ ParsedAction ]

{-
  To avoid testing for @Nothing@ we assume that if we do not
  find the component in the database it is merely because we have
  not yet added and we will do so in a future iteration.
  Besides which this analysis does not deal with undefined process names
  that is for another analysis to worry about.
-}
lookupCptActions :: ParsedComponentId -> ProcessActionDb -> [ ParsedAction ]
lookupCptActions name = (fromMaybe []) . (lookup name)


{-
  The strategy is to begin with an identity action database such
  that each process is mapped to the empty set of actions.
  Then we walk through the model adding to the data base entry
  each action that it may do, plus all the actions that may be done
  by any of the processes that the current one may become.
  This gives us a new process action database, we keep doing this
  until the process action database does not change.
-}
processActions :: ParsedModel -> ProcessActionDb
processActions (ParsedModel _rDefs pDefs _mainComp) =
    fixPoint processAllDefs empty
    where
    -- Notice here we always apply it to the same def list
    -- the model does not change, only the data base.
    processAllDefs :: ProcessActionDb -> ProcessActionDb
    processAllDefs db = foldl processDef db pDefs

    processDef :: ProcessActionDb -> ProcessDef -> ProcessActionDb
    processDef db (ident, proc) =
        insertList ident actions db
        where actions = processProc db proc

    processProc :: ProcessActionDb -> ParsedComponent -> [ ParsedAction ]
    processProc _db (StopProcess)                        =
       []
    processProc db (IdProcess name)                      = 
        lookupCptActions name db
    processProc db (PrefixComponent trans nextCpt) =
        nub $ action1 : nextActions
	where 
        nextActions = processProc db nextCpt
        action1     = pepaTransAction trans
    processProc db (CondBehaviour _ nextCpt) =
        processProc db nextCpt
    processProc db (ComponentSum left right)             =
        nub $ leftActions ++ rightActions
        where leftActions  = processProc db left
              rightActions = processProc db right
    processProc db (Cooperation left _ right)            =
        nub $ leftActions ++ rightActions
        where leftActions  = processProc db left
              rightActions = processProc db right
    processProc db (ProcessArray name _ _)               =
        lookupCptActions name db
    processProc db (Hiding p _acts)                      =
        processProc db p


{-
  The analysis done by 'processActions' allowed us to determine for each
  process the set of actions that it may eventually perform.
  This now tests any cooperation for a superflous action within
  the cooperation.

  Since here we are computing the actions done by the main cooperation
  we also return these, since this can be used to determine for example
  if the probe makes sense.
-}
excessActions :: ParsedModel -> ProcessActionDb -> ( [ ParsedAction ]
                                                   , [ AnalysisMessage ] )
excessActions (ParsedModel _rDefs _pDefs pComp) database =
    checkComp pComp
    where
    excessError :: ParsedComponent -> [ ParsedAction ] 
                -> [ ParsedAction ] -> AnalysisMessage
    excessError p acts1 acts2 = 
        AnalysisError $ ExcessActions p acts1 acts2

    needlessHideError :: ParsedComponent -> [ ParsedAction ]
                      -> [ ParsedAction ] -> AnalysisMessage
    needlessHideError p acts1 acts2 = 
        AnalysisError $ NeedlessActionsHide p acts1 acts2

    excessErrorLeft :: ParsedComponent -> [ ParsedAction ] 
                   -> [ ParsedAction ] -> AnalysisMessage
    excessErrorLeft p acts1 acts2 = 
        AnalysisError $ ExcessActionsLeft p acts1 acts2

    excessErrorRight :: ParsedComponent -> [ ParsedAction ]
                     -> [ ParsedAction ] -> AnalysisMessage
    excessErrorRight p acts1 acts2 = 
        AnalysisError $ ExcessActionsRight p acts1 acts2
    {-
      'checkComp' must do two things, it must return errors for
      any cooperation within it which has excess actions (which
      are either cooperated over or needlessly hidden). It must
      also return the actions which the input component can do
      such that the checking of enclosing pars can be done.
    -}
    checkComp :: ParsedComponent -> ( [ ParsedAction ],  [ AnalysisMessage ] )
    checkComp (StopProcess)                        =
      ([], [])
    checkComp (IdProcess ident)                    = 
        ( lookupCptActions ident database, [] )
    checkComp (PrefixComponent trans p)
        | elem act actions = (actions, messages)
        | otherwise        = (act : actions, messages)
        where 
        (actions, messages) = checkComp p
        act                 = pepaTransAction trans
    checkComp (CondBehaviour _ cpt)                =
      checkComp cpt
    checkComp (ComponentSum left right)            = 
        (actions, messages)
        where
        messages = messagesLeft ++ messagesRight
        actions  = nub $ actionsLeft ++ actionsRight
        (actionsLeft, messagesLeft)   = checkComp left
        (actionsRight, messagesRight) = checkComp right
    checkComp (ProcessArray ident _ Nothing)        =
        ( lookupCptActions ident database, [] )
    checkComp par@(ProcessArray ident _ (Just pActions))
        | notDoneBy == [] = ( idActions, [] )
        | otherwise       = ( idActions, [ errorRpt ] )
        where
        idActions = lookupCptActions ident database
        notDoneBy = pActions \\ idActions
        errorRpt  = excessError par pActions notDoneBy
    checkComp par@(Hiding left actions)
        | notDoneBy == [] = (leftActions, leftErrors)
        | otherwise       = (leftActions, errorRpt : leftErrors)
        where
        ( leftActions,
          leftErrors ) = checkComp left
        notDoneBy      = actions \\ leftActions
        errorRpt       = needlessHideError par actions notDoneBy
    checkComp proc@(Cooperation left actions right) = 
        ( leftActions ++ rightActions
        , concat [ leftError
                 , rightError
                 , spuriousError
                 , leftErrors
                 , rightErrors
                 ]
        )
        where
        ( leftActions,
          leftErrors )  = checkComp left
        ( rightActions,
          rightErrors ) = checkComp right

        notDoneByLeft   = actions \\ leftActions
        notDoneByRight  = actions \\ rightActions
        notDoneByEither = intersect notDoneByLeft notDoneByRight
        leftButNotRight = notDoneByRight \\ notDoneByLeft
        rightButNotLeft = notDoneByLeft \\ notDoneByRight
    
        leftError 
            | leftButNotRight == [] = []
            | otherwise             = 
                [ excessErrorRight proc actions leftButNotRight ]
        rightError 
            | rightButNotLeft == [] = []
            | otherwise             = 
                [ excessErrorLeft proc actions rightButNotLeft ]
        spuriousError
            | notDoneByEither == [] = []
            | otherwise             =
                [ excessError proc actions notDoneByEither ]


{-
  Note that this deliberately returns more than one instance of a name
  if it is defined more than once.
-}
definedProcesses :: ParsedModel -> [ ParsedComponentId ]
definedProcesses (ParsedModel _rDefs pDefs _) =
    map fst pDefs

{-
  This function returns the list of processes that can be reached from
  the main composition, whether or not those processes are defined.
  From this information we can easily define
  'unusedProcesses' which returns any processes that are defined but not
  used, and 'undefinedProcesses' which are used but not defined.
-}
type Seen = [ ParsedComponentId ]
usedProcesses :: ParsedModel -> [ ParsedComponentId ]
usedProcesses (ParsedModel _rDefs pDefs pComp) =
    searchComp [] pComp
    where
    {-
      There are two main auxiliary functions for this
      'searchComp' for components and
      'searchId' for looking up an identifier.
      Both of these must take in and maintain a list of "seen" identifiers
      such that we don't end up in an infinite loop.
    -}
    searchComp :: Seen -> ParsedComponent -> [ ParsedComponentId ]
    searchComp seen (IdProcess ident) 
        | elem ident seen = seen
        | otherwise       = searchId (ident : seen) ident pDefs
    searchComp seen (StopProcess)               =
        seen
    searchComp seen (PrefixComponent _ nextCpt) =
        searchComp seen nextCpt
    searchComp seen (ComponentSum left right)   =
        searchComp seenLeft right
        where seenLeft = searchComp seen left
    searchComp seen (CondBehaviour _ next)      =
        searchComp seen next
    searchComp seen (ProcessArray ident _ _)
        | elem ident seen = seen
        | otherwise       = searchId (ident : seen) ident pDefs
    searchComp seen (Hiding left _)            =
        searchComp seen left
    searchComp seen (Cooperation left _ right) =
        searchComp seenLeft right
        where seenLeft = searchComp seen left
              

    {-
      Note here that if we do not find the given identifier we do not
      raise an error, this is for another kind of analysis which can use
      the results of this to find processes that are used but not defined.
    -}
    searchId :: Seen -> ParsedComponentId -> [ ProcessDef ]
             -> [ ParsedComponentId ]
    searchId seen _      []                     = seen
    searchId seen ident ((ident2, comp) : rest)
        | ident == ident2 = searchComp seen comp
        | otherwise       = searchId seen ident rest


{-
  This analyses the 
  The next section checks for an unusued or undefined process definition.
  To do this we write two auxiliary functions, one returns the list
  of process names which are defined and the other the list of process
  definitions which are used. The first is very simple but the second must
  do a deep search, that is it is not enough to see which names are
  mentioned in the main composition but follow the definition of those
  names. These two 
  As a further analysis we check for process names that are defined
  more than once.
-}

analyseDefinedNames :: [ ParsedComponentId ] -> [ ParsedComponentId ] 
                    -> [ AnalysisMessage ]
analyseDefinedNames defNames usedNames =
    concat [ unusedProcesses
           , undefinedProcesses
           , multipleDefinedProcesses
           ]
    where
    unusedProcesses
        | definedButNotUsed == [] = []
        | otherwise               = 
          [ AnalysisWarning $ DefinedButNotUsedProcesses definedButNotUsed ]

    undefinedProcesses 
        | usedButNotDefined == [] = []
        | otherwise               = 
            [ AnalysisError $ UsedButUndefinedProcesses usedButNotDefined ]

    multipleDefinedProcesses
        | definedMultiply == [] = []
        | otherwise             =
            [ AnalysisError $ DefinedMultiplyProcesses definedMultiply ]

    nDefNames         = nub $ defNames
    usedButNotDefined = usedNames \\ nDefNames 
    definedButNotUsed = nDefNames \\ usedNames
    definedMultiply   = defNames  \\ nDefNames

{-
  Very similar to 'analyseDefinedNames' but for rate names defined.
-}
analyseDefinedRates :: [ RateIdentifier ] -> [ RateIdentifier ]
                    -> [ AnalysisMessage ]
analyseDefinedRates defNames usedNames =
    concat [ unusedRates
           , undefinedRates
           , multipleDefinedRates
           ]
    where
    unusedRates
        | definedButNotUsed == [] = []
        | otherwise               = 
          [ AnalysisWarning $ DefinedButNotUsedRates definedButNotUsed ]

    undefinedRates
        | usedButNotDefined == [] = []
        | otherwise               = 
            [ AnalysisError $ UsedButUndefinedRates usedButNotDefined ]

    multipleDefinedRates
        | definedMultiply == [] = []
        | otherwise             =
            [ AnalysisError $ DefinedMultiplyRates definedMultiply ]

    nDefNames         = nub $ defNames
    usedButNotDefined = usedNames \\ nDefNames 
    definedButNotUsed = nDefNames \\ usedNames
    definedMultiply   = defNames  \\ nDefNames


{-|
  The alias database is a bit of a misnomer actually.
  It does not build up aliases that only consist of ones such as
  @ P = S @
  but also ones that consist of
  @
    P = (a, r).S ;
    S = P + ... ;
  @
  Because for the purposes of this analysis it is indeed what
  we want. Basically we want a data base mapping processes to
  those that they can become without doing any actions.
  So in the above, @S can become a P@ without doing any
  actions.
-}
type AliasMap = FixMap ParsedComponentId [ ParsedComponentId ]

{- If the process is not in the alias data base then we assume that it has
   no aliases (at least not yet, we may be currently
   building up the alias database)
-}
findProcessAlias :: AliasMap -> ParsedComponentId -> [ ParsedComponentId ]
findProcessAlias aliasDb name = fromMaybe [] $ lookup name aliasDb


buildAliasDb :: ParsedModel -> AliasMap
buildAliasDb (ParsedModel _rDefs defs _system) =
    fixPoint processAllDefs empty
    where
    processAllDefs :: AliasMap -> AliasMap
    processAllDefs am = foldl processDef am defs

    addAlias :: ParsedComponentId -> ParsedComponentId
	     -> AliasMap -> AliasMap
    addAlias name1 name2 am = 
        insertList name1 newAliases am
        where
        name2Aliases = findProcessAlias am name2
        newAliases = if elem name2 name2Aliases
                     then name2Aliases
                     else name2 : name2Aliases

    processDef :: AliasMap -> ProcessSpec -> AliasMap
    processDef am (_    , StopProcess)            = am
    processDef am (_    , PrefixComponent _ _)    = am
    processDef am (name1, IdProcess name2)        =
        addAlias name1 name2 am
    processDef am (name1, ComponentSum cpt1 cpt2) =
        processDef am1 (name1, cpt2)
        where am1 = processDef am (name1, cpt1)
    processDef am (name1, CondBehaviour _ cpt1)   =
       processDef am (name1, cpt1)
    processDef am (_name1, Cooperation _ _ _)     = am
    processDef am (_name1, ProcessArray _ _ _)    = am
    processDef am (name1, Hiding p _)             =
        processDef am (name1, p)


{-
  An alias data base can be used to check for alias loops.
  This means where we have a set of processes that form a cycle
  and hence are undefined. The simplest example of this is.
  @ P = P @
  But we could also have
  @ P = S
    S = R
    R = P
  @
  Note that not all loops detected in this way will necessarily 
  be impossible to compile for example:
  @ P = S + R
    R = (a, r).R1
    S = P
  @
  however such a definition is effectively the same as simply:
  @ P = (a, r).R1
    S = P
    R = P
  @
  so we can reject it anyway since the model is defined in a strange
  way, and is quite possibly and error.
  
  todo: this only finds the first one and should in fact find all of them.
  The tricky part is that suppose we have
  @ P = S
    S = P
  @
  then the naive solution would report this loop twice, since it will be
  in the alias map twice, once under @P@ and once under @S@.
-}
checkForAliasLoops :: AliasMap -> Maybe AnalysisMessage
checkForAliasLoops aliasDb =
    do (ident, aliases) <- find isAliasLoop $ toList aliasDb
       return $ AnalysisError $ AliasLoop (ident : aliases)
    where
    isAliasLoop :: ( ParsedComponentId, [ ParsedComponentId ] ) -> Bool
    isAliasLoop (ident, aliases) = elem ident aliases


{-
  This section defines the functions required for detecting self-loops
  within the model.
  [@todo@] Write a bit more documentation, I've a couple of pages of
           notes in my notebook
  We maintain a dictionary mapping process identifiers to the actions
  on which they can do a self loop.

  Now we would like to detect a self loop by looking at a
  process definition and seeing:
  @
  P = (a, r) . P ;
  @
  Here we can see that the result of the prefix is the same as
  the name used in the definition.
  But this doesn't work because of aliases, for example:
  @
  P = (a, r) . S ;
  S = P ;
  @
  We would not detect that this is a simple loop because in
  the prefix the result is @S@ and not @P@. Even without
  a simple alias we can get the following:
  @
  P = (a, r) . S ;
  S = P + (a, r) . S1 ;
  @
  Here, the fact that @P@ is even a part of the definition of
  @S@ means that we end up with a self-loop. All of this
  goes to show that we must first build up a list of processes that
  each process may become without doing *any* action. So in the example
  above, @S@ can become @P@ without doing any actions.

  We have to store two kinds of actions, those on which the process
  may perform a self-loop, and those on which the process may perform
  a self-loop and which are hidden. The ones which are hidden cannot
  be cooperated over and hence will always be a self-loop.
-}
type ProcessSpec   = ( ParsedComponentId, ParsedComponent )
type LoopActions   = ( [ ParsedAction ], [ ParsedAction ] )
type ProcessLoopDb = FixMap ParsedComponentId LoopActions


{-
  To avoid testing for 'Nothing' we assume that if we do not
  find the component in the database it is merely because we have
  not yet added and we will do so in a future iteration.
-}
findProcessLoopActs :: ParsedComponentId -> ProcessLoopDb -> LoopActions
findProcessLoopActs name = (fromMaybe ([], [])) . (lookup name)


{-
    For now self-loops are not analysed. The change to allow parallel component
    definitions has invalidated the checking a little bit, however it should be
    fixable, it's just not the most important thing to do right now.

    Basically the problem boils down to:
    @
    P = (a, r).P; // self-loop
    Q = P \ { a } ; // still a self-loop
    R = < some process which does not self loop on 'a' >
    Q < a > R
    @
    I would say that this still does a self-loop but it will not be detected.
-}
selfLoopsModel :: AliasMap -> ParsedModel -> Maybe AnalysisMessage
selfLoopsModel aliasDb (ParsedModel _rDefs pDefs mainCoop) =
    case mainCoopLoopActions of
        ([], [])          -> Nothing
        (visible, hidden) -> 
            Just ( AnalysisError $ SystemSelfLoop mainCoop actions)
            where actions = union visible hidden
    where
    mainCoopLoopActions = analyseProcess defsLoopDb mainCoop
    defsLoopDb = fixPoint analyseAllDefs empty

    findAliases :: ParsedComponentId -> [ ParsedComponentId ]
    findAliases = findProcessAlias aliasDb

    -- Notice here we always apply it to the same definitions
    -- the definitions do *not* change, only the database
    analyseAllDefs :: ProcessLoopDb -> ProcessLoopDb
    analyseAllDefs db = foldl analyseDef db pDefs


    analyseProcess :: ProcessLoopDb -> ParsedComponent -> LoopActions
    analyseProcess _db (StopProcess)                        =
      ([], [])
    analyseProcess db (IdProcess ident)                     =
        findProcessLoopActs ident db
    analyseProcess db (CondBehaviour _ cpt)                 =
       analyseProcess db cpt
    analyseProcess db (PrefixComponent _ cpt)               =
        analyseProcess db cpt
    analyseProcess db (ComponentSum left right)             =
        ( union (fst leftActions) (fst rightActions)
        , union (snd leftActions) (snd rightActions)
        )
        where 
        leftActions  = analyseProcess db left
        rightActions = analyseProcess db right
    analyseProcess db (ProcessArray ident _ _)              =
        -- Note here it doesn't matter if we synchronise since all
        -- the processes are the same so if they synchronise then
        -- then it is still a self-loop.
        -- @todo@ however we should still move the synchronised
        -- actions over to the hidden side.
	findProcessLoopActs ident db
    analyseProcess db (Cooperation cpt1 coopActions cpt2)   =
        ( union exclusive mutual
        , union (snd loopActions1) (snd loopActions2)
        )
	where 
	loopActions1     = analyseProcess db cpt1
	loopActions2     = analyseProcess db cpt2
	actions1         = fst loopActions1
        actions2         = fst loopActions2
        {-
          The exclusive actions are those that either process can
          self loop on and which are *not* in the cooperation set.
          This kind of assumes that both processes do actually
          perform the actions involved in the cooperation.
        -}
        exclusive1   = (actions1 \\ coopActions)
        exclusive2   = (actions2 \\ coopActions)
        exclusive    = exclusive1 ++ exclusive2

        {-
            The mutual actions are those which both processes
            can self loop on, it is here where I think we may
            be *slightly* wrong, it could be that because of
            the actions within the cooperation set, that
            both cannot do that action as a self-loop because
            of where they otherwise synchronise.
            Of course "wrong" is not the correct word here, I mean
            conservative, here is where we may be a little conservative.
        -}
        mutual       = intersect coopActions $ intersect actions1 actions2
    analyseProcess db (Hiding innerCoop actions)            =
	(stillVisible, union nowHidden hActions)
	where
	(vActions, hActions)      = analyseProcess db innerCoop
        (stillVisible, nowHidden) = partition notHiddenHere vActions

        notHiddenHere :: ParsedAction -> Bool
        notHiddenHere act = elem act actions

    {-
      The function @analyseDef@ must check if a component can
      do an action and loop back to itself.
      A prefix action may loop back to itself, a sum can do
      this if one of the branches can loop back to the original
      definition.
      
      The main computation of this part is to decide whether or not
      the next component part of a prefix can become the original
      component without performing any actions, if so then
      the original component has a loop.
      
      Hiding complicates things slightly, because if we do a self-loop
      on an action which is hidden then there is no chance to prevent
      the self-loop with synchronisation. We get around this by
      the @analyse...@ functions returning a pair, those actions
      that we do a self-loop on and those actions that we do
      a self-loop on and are hidden. Now we cheat a little because
      in the process loop data base we only store *one* list of actions
      on which we may loop. 
    -}
    analyseDef :: ProcessLoopDb -> ProcessSpec -> ProcessLoopDb
    analyseDef db (name, PrefixComponent trans nextCpt) =
	insertWith loopActionsMerge name actions db
	where 
        action      = pepaTransAction trans
	nextActions = analyseProcess db nextCpt
	aliases     = name : (findAliases name)
	loopsBack   = canBecome nextCpt aliases
	newActions  = mapFst (action :) nextActions
	actions
	    | loopsBack &&
	        (not $ elem action (fst nextActions)) = newActions
	    | otherwise                               = nextActions

    analyseDef db (name, ComponentSum cpt1 cpt2)              =
	analyseDef db1 (name, cpt1)
	where db1 = analyseDef db (name, cpt2)
    analyseDef db (name, proc) =
	insertWith loopActionsMerge name actions db
	where actions = analyseProcess db proc

    -- A function suitable for use with 'insertWith', it must
    -- return the exact same loop actions, if the two lists are
    -- equal, in the sense that we are not adding to the current
    -- list of loop actions.
    loopActionsMerge :: LoopActions -> LoopActions -> LoopActions
    loopActionsMerge la1 la2
        | loopActionsChange la1 la2 =
            ( union (fst la1) (fst la2)
            , union (snd la1) (snd la2)
            )
        | otherwise                 = la2

    -- A function to say whether one loop action is
    -- subsumed by another.
    loopActionsChange :: LoopActions -> LoopActions -> Bool
    loopActionsChange la1 la2 =
        (any (\x -> not $ elem x (fst la2)) $ fst la1)
        ||
        (any (\x -> not $ elem x (snd la2)) $ snd la1)



    {-
      This is the function which tells us if we can return to
      the original component without doing any actions.
      Basically whenever we see a definition such as :
      @ P = (a, r). <component> @
      we want to check if @<component>@ may become a @P@ without
      doing any actions, for example if @<component>@ is @S@
      and we have
      @ S = P + (b, r).Q ; @
      then @S@ can 'become' (or behave as though it is) a @P@.
    -}
    canBecome :: ParsedComponent -> [ ParsedComponentId ] -> Bool
    canBecome (StopProcess)     _               = False
    canBecome (IdProcess ident) aliases2        =
      not $ null $ intersect aliases1 aliases2
      where aliases1 = ident : (findAliases ident)
    canBecome (ComponentSum cpt1 cpt2) aliases =
	canBecome cpt1 aliases || canBecome cpt2 aliases
    canBecome (CondBehaviour _ next)   aliases =
       canBecome next aliases
    -- A prefix can become nothing *without* doing an action
    canBecome (PrefixComponent _ _) _          = False
    canBecome (Cooperation _ _ _)   _          = False
    canBecome (ProcessArray _ _ _)  _          = False
    canBecome (Hiding p _) aliases             =
        -- Hmm, not sure about this actually.
        canBecome p aliases



{-
  Finding all the rate names defined within a model is very simple we simply
  walk through the definitions in the 'ParsedModel'.
  This returns the full spec, the names of those defined within the model
  can be gotten from simply mapping 'fst' over the list of rate specifications
  returned.
  Okay now that we have updated the 'ParsedModel' data type this function is
  even simpler, yay.
-}
ratesDefinedModel :: ParsedModel -> [ RateSpec ]
ratesDefinedModel (ParsedModel rDefs _pDefs _mainCompoent) = rDefs    


{-
    Finds all the rates used in a model. This is a little simplistic
    since it considers a use in a rate definition to still be a use.
    It shouldn't really.
    Consider
    @ r = 1.0 ;
      s = r + 1.0 ;
      < model that doesn't use s or r >
    @
    In this case @r@ will not be reported as not being used. But it
    doesn't actually get used in the model since it is only used
    in the definition of @s@ which itself isn't used. It would be
    simple to augment this. For now it isn't a big deal since the user
    will at least get one warning since @s@ isn't used.

    In fact it is worse than that, we should only actually consider
    the process definitions that are actually used.
  
    We also don't consider the main system, since we shouldn't really
    allow prefixes in that.
-}
ratesUsedModel :: ParsedModel -> [ RateIdentifier ]
ratesUsedModel (ParsedModel rDefs pDefs _mainCoop) =
   nub (ratesRates ++ ratesProcesses)
   where
   ratesRates     = concat $ map ratesOfRateSpec rDefs
   ratesProcesses = concat $ map ratesOfProcessDef pDefs

   ratesOfRateSpec :: RateSpec -> [ RateIdentifier ]
   ratesOfRateSpec = ratesOfRateExpr . snd

   ratesOfProcessDef :: ProcessDef -> [ RateIdentifier ]
   ratesOfProcessDef = ratesOfComponent . snd
   
   ratesOfComponent :: ParsedComponent -> [ RateIdentifier ]
   ratesOfComponent (IdProcess _)                  = []
   ratesOfComponent (StopProcess)                  = []
   ratesOfComponent (PrefixComponent trans nextP)  =
       (ratesOfRate $ pepaTransRate trans) ++ 
       (ratesOfComponent nextP)
   ratesOfComponent (ComponentSum left right)      =
       (ratesOfComponent left) ++ (ratesOfComponent right)
   ratesOfComponent (CondBehaviour cond next)      =
      (ratesOfCondition cond) ++ (ratesOfComponent next)
   ratesOfComponent (Cooperation left _ right)     =
       (ratesOfComponent left) ++ (ratesOfComponent right)
   ratesOfComponent (ProcessArray _ _ _)           = []
   ratesOfComponent (Hiding _ _)                   = []

   ratesOfCondition :: BeCondition -> [ RateIdentifier ]
   ratesOfCondition (BeCondition r _ rExpr) =
      r : (ratesOfRateExpr rExpr)

   ratesOfRate :: ParsedRate -> [ RateIdentifier ]
   ratesOfRate (RateTop)          = []
   ratesOfRate (RateImmediate _ ) = []
   ratesOfRate (RateTimed e)      = ratesOfRateExpr e

   ratesOfRateExpr :: ParsedRateExp -> [ RateIdentifier ]
   ratesOfRateExpr (RateId ident _)        = [ ident ]
   ratesOfRateExpr (RateNum _)             = []
   ratesOfRateExpr (RateBexp _ left right) =
       (ratesOfRateExpr left) ++ (ratesOfRateExpr right)
   ratesOfRateExpr (RateProc _)            = []
   ratesOfRateExpr (RateCond _ left right) = 
       (ratesOfRateExpr left) ++ (ratesOfRateExpr right)
   ratesOfRateExpr (RateDisabled)          = []

-------------------------------------------------------------------------
-- Individual analysis functions
-------------------------------------------------------------------------

-- The type of a transition from one state to another.
-- type Transition = (ParsedComponentId, ParsedComponentId)

{-|
   Returns the possible transitions from a given starting component in
   the given model. The model must have been reduced to the state where
   all prefix components have, as their right hand side a component
   identifier, and not an arbitrary form of sequential component.
-}
possibleSuccessors :: ParsedComponentId -> ParsedModel -> [ ParsedComponentId ]
possibleSuccessors startIdent model =
   possibleSuccessorsInModel [] startIdent
   where
   -- The first argument is the 'seen' list, ie the list of process identifiers
   -- which we have already processed and hence do not need to look up
   -- again. Note that this could quite easily be made tail recursive if
   -- we doubled up the successors with the tail-recursive-argument.
   possibleSuccessorsInModel :: [ ParsedComponentId ] -> ParsedComponentId
                             -> [ ParsedComponentId ]
   possibleSuccessorsInModel seen ident
      | elem ident seen = []
      | otherwise       =
         case findProcessDefinition ident model of
            Nothing   -> error notFound
            Just comp -> possibleComp comp
      where
      newSeen = ident : seen
      possibleComp :: ParsedComponent -> [ ParsedComponentId ]
      possibleComp (StopProcess)                         =
         []
      possibleComp (IdProcess ident2)                    =
         possibleSuccessorsInModel newSeen ident2
      possibleComp (ComponentSum left right)             =
         union (possibleComp left) (possibleComp right)
      possibleComp (PrefixComponent _ (IdProcess right)) =
         union [ right ] $ possibleSuccessorsInModel newSeen right
      possibleComp (CondBehaviour _ p)                   =
         possibleComp p
      
      {- The errors cases -}
      possibleComp (PrefixComponent _ _)                 =
         error notSimplified
      possibleComp (Cooperation _ _ _ )                  =
         error errorMessage
      possibleComp (ProcessArray _ _ _ )                 =
         error errorMessage
      possibleComp (Hiding _ _ )                         =
         error errorMessage
   
   errorMessage  = "Attempt to obtain the sequentional transitions " ++
                   "of a parallel component"
   notFound      = "component not found when attempting to gather next " ++
                   "transitions"
   notSimplified = "The model has not first been simplfied before calling " ++
                   "possibleTransitions"


{-|
   Returns the list of actions which it is possible for the given
   component to perform. This does NOT include those done by alias
   components. So for example if the component given is
   @ (a, \infty).P + R @
   the returned list is simply the singleton list containing @a@,
   regardless of what @R@ may be able to do. This is apparent
   from the type, since we do not take in a model or list of definitions
   there is no way for this function to know what @R@ may perform.

   Parallel components are ignored completely.
-}
possibleActionsNonAliased :: ParsedComponent -> [ ParsedAction ]
possibleActionsNonAliased (IdProcess _)              = []
possibleActionsNonAliased (StopProcess)              = []
possibleActionsNonAliased (PrefixComponent trans _)  = 
   [ pepaTransAction trans ]
possibleActionsNonAliased (ComponentSum left right)  =
   union (possibleActionsNonAliased left  )
         (possibleActionsNonAliased right )
possibleActionsNonAliased (CondBehaviour _ p)        =
   possibleActionsNonAliased p
possibleActionsNonAliased (Cooperation _ _ _ )       =
   error "possibleActionsNonAliased: parallel component found"
possibleActionsNonAliased (ProcessArray _ _ _ )      =
   error "possibleActionsNonAliased: parallel component found"
possibleActionsNonAliased (Hiding _ _ )              =
   error "possibleActionsNonAliased: parallel component found"
   

{-| The result of a call to 'getInitialConcentrations' -}
type ConcentrationMap = [ ( ParsedComponentId, Int) ]

{-|
   'getInitialConcentrations' maps a pepa model to an initial concentration
   list of states (here the states are not full states but just the states
   of individual components.

   The model must be in Normal form, that is NO parallel definitions,
   and prefix components must only have an 'IdProcess' as their next
   component. All components are used exactly once.
-}
getInitialConcentrations :: ParsedModel -> ConcentrationMap
getInitialConcentrations (ParsedModel _rDefs pDefs mainComp) =
   getParallel mainComp
   where
   getParallel :: ParsedComponent -> ConcentrationMap
   getParallel (IdProcess ident)                  = 
      (ident, 1) : (getSequentialId [ ident ] ident)
   getParallel (Cooperation left _ right)         =
      (getParallel left) ++ (getParallel right)
   getParallel (Hiding comp _)                    =
      getParallel comp
   getParallel (ProcessArray ident size _actions) =
      (ident, fromInteger size) : (getSequentialId [ ident ] ident)
   
   {- Sequential components in the main system equation-}
   getParallel (StopProcess)                      =
      error "Sequential component in the main system equation"
   getParallel (CondBehaviour _ _)                =
      error "Sequential component in the main system equation"
   getParallel (PrefixComponent _ _)              =
      error "Sequential component in the main system equation"
   getParallel (ComponentSum _ _)                 =
      error "Sequential component in the main system equation"

   getSequentialId :: [ ParsedComponentId ] -> ParsedComponentId
                   -> ConcentrationMap
   getSequentialId seen ident =
      -- Basically a 'lookup' but the Prelude 'lookup' is hidden
      case find ((ident ==) . fst) pDefs of 
         Nothing      -> error "Undefined process in getInitialConcentrations"
         Just (_, p)  -> getSequential seen p

   getSequential :: [ ParsedComponentId ] -> ParsedComponent -> ConcentrationMap
   getSequential _seen (StopProcess)            =
      []
   getSequential seen (IdProcess ident)
      | elem ident seen = []
      | otherwise       = (ident, 0) : (getSequentialId (ident : seen) ident)
   getSequential seen (PrefixComponent _ next)  =
      getSequential seen next
   getSequential seen (CondBehaviour _ next)    =
      getSequential seen next
   getSequential seen (ComponentSum left right) =
      union (getSequential seen left) (getSequential seen right)
   getSequential _seen (Cooperation _ _ _)      =
      error $ "Parallel component in sequential definition "
              ++ "detected in getInitialConcentrations"
   getSequential _seen (ProcessArray _ _ _)     =
      error $ "Parallel component in sequential definition "
              ++ "detected in getInitialConcentrations"
   getSequential _seen (Hiding _ _)             =
      error $ "Parallel component in sequential definition "
              ++ "detected in getInitialConcentrations"

   


{-



{-
  \subsection{Rate Expression Reduction}
  This section involves the reduction of rate expressions.

  The reduction of a parsed rate expression given a dictionary mapping
  rate identifiers to values. 

  @todo{We do not check for recursive definitions, obviously we should}
-}
type RateMapping = FixMap RateIdentifier ParsedRateExp

reduceRateSpecs :: [ RateSpec ] -> [ RateSpec ]
reduceRateSpecs specs = 
    toList $ fixPoint reduceAllSpecs empty
    where
    reduceAllSpecs :: RateMapping -> RateMapping
    reduceAllSpecs rm = foldl reduceRateSpec rm specs

    reduceRateSpec :: RateMapping -> RateSpec -> RateMapping
    reduceRateSpec mapping (rateName, rateExp) =
	insert rateName reducedExp mapping
	where reducedExp = reduceRateExpr mapping rateExp

    reduceRateExpr :: RateMapping -> ParsedRateExp -> ParsedRateExp
    reduceRateExpr _ RateTop                             = RateTop
    reduceRateExpr _ rate@(RateNum _)                    = rate
    reduceRateExpr _ rate@(RateProc _)                   = rate
    -- Note here that we do not reduce the original expression
    -- because of course the whole point of that is that it contains
    -- the original expression in order to show the user.
    reduceRateExpr mapping rate@(RateId ident _)         =
	case lookup ident mapping of
	Nothing   -> rate
	Just rExp -> reduceRateExpr mapping rExp
    reduceRateExpr mapping (RateBexp oper lexp rexp)     =
	case (rLexp, rRexp) of
	(RateNum v1, RateNum v2) -> applyOper v1 v2
	_                        -> RateBexp oper rLexp rRexp
	where
	rLexp = reduceRateExpr mapping lexp
	rRexp = reduceRateExpr mapping rexp

        applyOper :: ParsedValue -> ParsedValue -> ParsedRateExp
	applyOper leftValue rightValue =
	    case oper of
	    "+" -> RateNum $ leftValue + rightValue
	    "-" -> RateNum $ leftValue - rightValue
	    "*" -> RateNum $ leftValue * rightValue
	    "/" -> RateNum $ leftValue / rightValue
	    _   -> error $ "Undefined rate expression operator: " ++ oper
    -- For now we just assume that the condition can never be reduce
    reduceRateExpr mapping (RateCond cond lexp rexp)     =
	RateCond cond rLexp rRexp
	where
	rLexp = reduceRateExpr mapping lexp
	rRexp = reduceRateExpr mapping rexp
-}
