{-|
   A module definining 'GenericParallel' structures. The idea is that we can
   compile to several different output formats by going half of the way to
   a generic parallel structure. This can then be flattened out by each
   output format.
-}
module Language.Pepa.Compile.GenericParallel
   ( Successor
   , SuccessorMap

   , GenericParallel     ( .. )
   , GPelement           ( .. )

   , toGenericParallel
   )
where

{- Imported Standard Libraries -}
import Data.Maybe
   ( fromMaybe )
{- Imported Local Libraries -}
import Language.Pepa.Syntax
    ( ParsedModel          ( .. )
    , ParsedComponent      ( .. )
    , ParsedComponentId

    , ParsedAction
    , ParsedTrans
    )
import Language.Pepa.PepaUtils
   ( SuccessorTrans
   , allSuccessors
   )
import Language.Pepa.Analysis.Analysis
   ( possibleSuccessors )

{- End of Imports -}


{- The type of a successor map.
   It should be made by producing from a model a list of
   @(Ident, Trans, Maybe Ident)@ if any of the third components are 'Nothing'
   then the model wasn't put into the correct normal form first
   (or that failed) and we cannot proceed so we shouldn't get as far as
   calling 'fromPepaComponent'
-}
type Successor   = (ParsedComponentId, (ParsedTrans, ParsedComponentId) )
type SuccessorMap = [ Successor ]

data GPelement a  => GenericParallel a =
      PAR   (GenericParallel a) [ ParsedAction ] (GenericParallel a)
    | ARRAY (GenericParallel a) Integer [ ParsedAction ] 
    | SEQ a
--    | HID (GenericParallel a) [ ParsedAction ]



{-
   It's tempting for this to instead have 3 methods, one for each of
   the three possible sequential process that get translated by this.

   Okay but prefix is never reasonable here. Prefix would mean there
   is a prefix in the main system equation which probably shouldn't
   be allowed and even if so we should remove it with an alias before
   calling 'toGenericParallel'.

   Now that I think about this, that is also true of 'gpSumElement'
   The trick is that when we call for the successor to a given
   process identifier, if that process identifier is say @P@ and
   @P@ is defined as:
   @P = Q + R @
   then we should return all the transitions that @Q@ and @R@ can
   perform.
-}
{-| To be an element of a generic parallel structure ('GenericParallel')
    it must be possible to translate a given Pepa Component into that
    type.
-}
class GPelement a where
   gpIdElement     :: SuccessorMap -> ParsedComponentId -> a
   stopElement     :: a
   -- gpPrefixElement :: SuccessorMap -> ParsedTrans -> ParsedComponentId -> a
   -- gpSumElement    :: SuccessorMap -> ParsedComponentId
      --              -> ParsedComponentId -> a


{-| If you are reading this, ask adc to document this function.
   However in the mean-time a very important note: this function
   requires that the model is already transformed so that each
   occurrence of a component in a parallel structure has a unique
   identifier. This will be a transformation available in the
   'Language.Pepa.Transform' module (and is already actually
   written in the old ipc code).
   This has also removed parallel aliases, hence if we have an
   'IdProcess' component we may assume that it is sequential.
   This is done by calling 'Language.Pepa.Transform.removeParallelDefs'
-}
toGenericParallel :: GPelement a => ParsedModel -> GenericParallel a
toGenericParallel model@(ParsedModel _rDefs _pDefs mainComp) =
   mkGeneric mainComp
   where
   mkGeneric :: GPelement a => ParsedComponent -> GenericParallel a
   mkGeneric (StopProcess)                                     =
      SEQ stopElement
   mkGeneric (IdProcess ident)                                 = 
      SEQ $ gpIdElement appropriateSuccessors ident
      where
      appropriateSuccessors = filter isAppropriate successorMap
      succesorsOfIdent      = possibleSuccessors ident model
      -- Now we filter the successorMap so that we only give
      -- the actions which are performed by the component currently
      -- under consideration.
      isAppropriate :: Successor -> Bool
      isAppropriate (source,_) = 
         ( source == ident ) || ( elem source succesorsOfIdent )

   -- mkGeneric (PrefixComponent trans (IdProcess ident))         =
      -- SEQ $ gpPrefixElement successorMap trans ident
   mkGeneric (PrefixComponent _ _)                             =
      error "During mkGeneric, model not in normal form"
   mkGeneric (CondBehaviour _ _)                               =
      error $ "During mkGeneric, model not in normal form" ++
              " conditional behaviour should have been removed"
   -- mkGeneric (ComponentSum (IdProcess left) (IdProcess right)) =
      -- SEQ $ gpSumElement successorMap left right
   mkGeneric (ComponentSum _ _)                                =
      error "During mkGeneric, model not in normal form; SUM"

   -- Now the parallel components
   mkGeneric (Cooperation left actions right)                  = 
      PAR leftGP actions rightGP
      where
      leftGP  = mkGeneric left
      rightGP = mkGeneric right
   mkGeneric (ProcessArray ident size actions)                 =
      ARRAY element size $ fromMaybe [] actions
      where
      element = mkGeneric (IdProcess ident) 
   mkGeneric (Hiding _comp _actions)                           = 
      error "We should have removed hiding via renaming before getting here"
   --      HID (mkGeneric comp) actions
   
   {-
      'Language.Pepa.PepaUtils.allSuccessors' returns too generous
      a type since it is possible to use this function on a model
      which is not in normal form. Here we use that function and
      then check that none of the mappings it returns are 'Nothing'
      since that would indicate the model was not in normal form
      and we couldn't proceed.
   -}
   successorMap :: SuccessorMap
   successorMap = 
      map makeSuccessor $ allSuccessors model
      where
      makeSuccessor :: SuccessorTrans -> Successor
      makeSuccessor (Just ident, (trans, Just ident2)) =
         (ident, (trans, ident2))
      makeSuccessor (Nothing, _)                       =
         error $ concat [ "Language.Pepa.Compile.GenericParallel: "
                        , "Attempt to compile a model not in normal form. "
                        , "Caught from a nothing in the SOURCE position "
                        , "of a successor"
                        ]
      makeSuccessor _                                  =
         error $ concat [ "Language.Pepa.Compile.GenericParallel: "
                        , "Attempt to compile a model not in normal form. "
                        , "Caught from a nothing in the TARGET position "
                        , "of a successor"
                        ]
