{-|
    This module is a collection of utility functions performed over the
    syntax of Pepa models.
-}
module Language.Pepa.PepaUtils
   ( isPepaActionEnabled
   , Transition
   , possibleTransSuccessors
   , possibleTransitions
   , allTransitionsOfModel
   
   , SuccessorTrans
   , SuccessorTransMap
   , allSuccessors
   
   , findProcessDefinition
   
   , makePepaChoice
   , addProcessDefinitions
   
   , ParOrSeq                ( .. )
   , isParOrSeqComponent
   
   )
where

{- Standard Modules Imported -}

{- Local Modules Imported -}
import Language.Pepa.Syntax            
   ( ParsedModel          ( .. )
   , ProcessDef
   , ParsedComponent      ( .. )
   , ParsedComponentId
   , ParsedTrans          ( .. )
   , ParsedAction         ( .. )
   )
{- End of Imports -}

{-|
    Returns true if the pepa component given may perform the given action.
-}
isPepaActionEnabled :: ParsedAction -> ParsedComponentId -> ParsedModel -> Bool
isPepaActionEnabled  act ident model = 
    any ((act ==) . pepaTransAction) $ possibleTransitions model ident

-- | The type of transitions from components to components
type Transition = (ParsedTrans, Maybe ParsedComponentId)

-- | The type of /seen/ process identifiers
type Seen = [ ParsedComponentId ]

{-|
    Returns the list of possible transitions from a given starting point
    in a sequential component. 
-}
possibleTransSuccessors :: ParsedModel -> ParsedComponentId -> [ Transition ]
possibleTransSuccessors model ident =
    snd $ possibleTrans [] ident
    where   
    -- The first argument is a list of seen processes this stops us
    -- going into the inevitable infinite loop.
    possibleTrans :: Seen -> ParsedComponentId -> ( Seen , [ Transition ] )
    possibleTrans seen ident2
        | elem ident2 seen = (seen, [])
        | otherwise       = 
            case findProcessDefinition ident2 model of
               Nothing   -> error notFound
               Just comp -> possibleTransComp (ident2 : seen) comp

    -- The same thing here, the first argument is the list of seen
    -- processes.
    possibleTransComp :: Seen -> ParsedComponent -> (Seen, [ Transition ])
    possibleTransComp seen (StopProcess)              = (seen, [])
    possibleTransComp seen (IdProcess ident2)         =
        possibleTrans seen ident2
    possibleTransComp seen (ComponentSum left right)  =
        -- Note, that I think this actually messes up on: @P = Q + Q@
        -- that is it will produce less transitions that it should.
        (rightSeen, leftTrans ++ rightTrans)        
        where
        (leftSeen, leftTrans)   = possibleTransComp seen left
        (rightSeen, rightTrans) = possibleTransComp leftSeen right
    possibleTransComp seen (PrefixComponent trans s)  =
        case s of
            IdProcess ident2 -> (seen, [ (trans, Just ident2) ])
            _                -> (seen, [ (trans, Nothing ) ])
    possibleTransComp seen (CondBehaviour _ comp)     =
       possibleTransComp seen comp
    possibleTransComp _seen (Cooperation _ _ _ )      =
        error errorMessage
    possibleTransComp _seen (ProcessArray _ _ _ )     =
        error errorMessage
    possibleTransComp _seen (Hiding _ _ )             =
        error errorMessage

    errorMessage  = "Attempt to obtain the sequetional transitions " ++
                    "of a parallel component"
    notFound      = "component not found when attempting to gather next " ++
                    "transitions"


possibleTransitions :: ParsedModel -> ParsedComponentId -> [ ParsedTrans ]
possibleTransitions model = (map fst) . (possibleTransSuccessors model)

{-|
  Finds all the transitions of a model, note that it is possible that some
  of these can never be performed, for example if they are defined within
  an unused process.
-}
allTransitionsOfModel :: ParsedModel -> [ ParsedTrans ]
allTransitionsOfModel = 
   (map takeTrans) . allSuccessors 
   where
   takeTrans :: SuccessorTrans -> ParsedTrans
   takeTrans = fst . snd

{-|
   Yet another type of successor transition.
   This one includes the possibility of having the identifier
   of the component from which we are transisting.
   This is a good analysis to do when the model is in normal
   form such that every prefix component is a single prefix
   component within a process definition such as:
   @P = (a, r).Q@
   That is, there is no chaining of prefix components and they are not
   a part of other components such as prefixes.

   [@todo@] all the other ones should just call this and then
   map the result.

   I still think this is not quite right, if we have
   @P = S@
   then we should have all the successors of @S@ as being possible
   from @P@.

   So now this uses 'possibleTransSuccessors'

   This should be re-written to descend from the main composition.
   My worry here is that we might duplicate the transitions in the whole
   model due to aliases (of course we shouldn't call this if the model
   is not in normal form, but still). 
-}
type SuccessorTrans = ( Maybe ParsedComponentId
                      , (ParsedTrans, Maybe ParsedComponentId)
                      )
type SuccessorTransMap = [ SuccessorTrans ]

allSuccessors :: ParsedModel -> SuccessorTransMap
allSuccessors model@(ParsedModel _rDefs pDefs _mainCoop) =
   concat $ map succsOfDef pDefs
   where
   succsOfDef :: ProcessDef -> SuccessorTransMap
   succsOfDef (ident, _comp) =
      map addIdent transSuccs
      where
      transSuccs = possibleTransSuccessors model ident
      addIdent :: Transition -> SuccessorTrans
      addIdent t = (Just ident, t)
      
{-
   succsOfDef :: ProcessDef -> SuccessorTransMap
   succsOfDef (ident, PrefixComponent t (IdProcess ident2)) =
      [ (Just ident, (t, Just ident2)) ]
   succsOfDef (ident, PrefixComponent t next)               =
      ( Just ident, (t, Nothing) ) : (succsOfProcess next)
   succsOfDef (ident, ComponentSum left right)              =
      (succsOfDef (ident, left)) ++
      (succsOfDef (ident, right))
   succsOfDef (_, component)                                =
      succsOfProcess component
   
   succsOfProcess :: ParsedComponent -> SuccessorTransMap
   succsOfProcess (IdProcess _)                         =
      []
   -- So here the second maybe is a 'Just' because there is
   -- an ident, we assume that the definition is not directly
   -- above here since in that case it would have been caught
   -- by 'succsOfDef'
   succsOfProcess (PrefixComponent t (IdProcess ident)) =
      [ (Nothing, (t, Just ident)) ]
   -- Right, now we have no ident as the source or the target
   succsOfProcess (PrefixComponent t next)              =
      (Nothing, (t, Nothing)) : (succsOfProcess next)
   succsOfProcess (ComponentSum left right)             =
      (succsOfProcess left) ++ (succsOfProcess right)
   
   -- There shouldn't be any in the parallel ones, but oh well
   succsOfProcess (Cooperation left _ right)            =
      (succsOfProcess left) ++ (succsOfProcess right)
   succsOfProcess (ProcessArray _ _ _)                  =
      []
   succsOfProcess (Hiding left _)                       =
      succsOfProcess left
-}

-- | Finds a process definition in a pepa model
findProcessDefinition :: ParsedComponentId -> ParsedModel
                      -> Maybe ParsedComponent
findProcessDefinition ident (ParsedModel _rDefs pDefs _mainComp) =
    lookup ident pDefs


{-| makes a choice between a given number of components.
    Note that this currently just raises an error if passed the empty list,
    an alternative would be to go to the special @Stop@ process.
-}
makePepaChoice :: [ ParsedComponent ] -> ParsedComponent
makePepaChoice [] = error "makePepaChoice called with an empty list"
makePepaChoice ls =
    foldl1 ComponentSum ls

addProcessDefinitions :: [ ProcessDef ] -> ParsedModel -> ParsedModel
addProcessDefinitions defs (ParsedModel rDefs pDefs coop) =
   ParsedModel rDefs (pDefs ++ defs) coop

{-| The type returned by 'isParOrSeqComponent' -}
data ParOrSeq = IsPar | IsSeq | ParOrSeq
                deriving Eq

{-|
   Returns whether or not a component is a parallel component.
   It relies on a function to tell whether a component identifier
   is parallel or sequential. This will depend on the state of the
   model at the time. If for example we have applied a tranformation
   to remove all parallel defs then all component identifiers can
   be assumed to be sequential.
-}
isParOrSeqComponent :: ParOrSeq --  What to return for an identifier 
                    -> ParsedComponent -> ParOrSeq
isParOrSeqComponent idRes (IdProcess _)           = idRes
isParOrSeqComponent _     (StopProcess)           = IsSeq
isParOrSeqComponent _     (PrefixComponent _ _)   = IsSeq
isParOrSeqComponent _     (ComponentSum _ _)      = IsSeq
isParOrSeqComponent _     (CondBehaviour _ _)     = IsSeq -- dodgy?
isParOrSeqComponent _     (Cooperation _ _ _)     = IsPar
isParOrSeqComponent _     (ProcessArray _ _ _)    = IsPar
isParOrSeqComponent _     (Hiding _ _)            = IsPar

