{-|
  This module is intended to be a pretty printer for pepa models.
  It should be updated to use 'Text.PrettyPrint.HughesPJ'
-}
module Language.Pepa.Print
   ( HumanPrint                ( .. )
   , hPrintPepaModel
   , hprintProcessDef
   , hprintRateSpec
   , hprintRateIdentifier
   
   -- * Printing for the analysis messages
   , hprintAnReport
   , hprintMessages
   )
where


{- Imported Standard Libraries -}
import Data.Char
   ( toLower )
import Data.Map
   ( toList )
import Numeric
   ( showFFloat )
{- Imported Local Libraries -}
import Language.Pepa.Syntax
    ( QualifiedName        ( .. )
    
    , ParsedModel          ( .. )
    , ProcessDef
    , RateSpec
    , RateIdentifier
    , ParsedComponentId
    , ParsedComponent      ( .. )
    , BeCondition          ( .. )

    , ParsedTrans          ( .. )
    , ParsedAction         ( .. )
    , ParsedRate           ( .. )
    , ParsedRateExp        ( .. )
    , ParsedCond           ( .. )
    )

import Language.Pepa.Utils
    ( mkCSlist 
    , mkGenSeplist
    , dropFromEnd
    )
import Language.Pepa.Analysis.Analysis
    ( AnalysisMessage  ( .. )
    , AnalysisResult   ( .. )
    , AnalysisReport   ( .. )
    , messagesOfReport
    , splitMessages
    )
{- End of Imports -}



{-|
  A class for printing in a more human readable form than the 'Show' class.
-}
class HumanPrint a where
    hprint :: a -> String

{- A silly instance, however sometimes parameterised components
   during development it is convenient to parameterise them with
   the unit type, so allowing this just allows them to be printed.
-}
instance HumanPrint () where
   hprint () = "()"

{-| 
   Qualified names are human printable.
   Technically speaking namespaced names should be separated out using
   a double colon rather than a double underscore. In this way though
   I can use the same function to output a pepa model. I really should
   have a different function though.   
-}
instance HumanPrint QualifiedName where
    hprint (Qualified ident il)   = concat [ ident, "_", mkGenSeplist "_" il]
    hprint (Mangled prefix ident) = concat [ prefix, "_", ident ]
    hprint (Unqualified ident)    = ident
    hprint (NameSpaceId ident qn) = concat [ ident, "__", hprint qn ]


{-|
   The main exported function of this model. 
   Prints a pepa model in a human readable format.
-}
hPrintPepaModel :: ParsedModel -> String
hPrintPepaModel (ParsedModel rdefs pdefs parsedComp) =
    unlines $ (map hprintRateSpec rdefs) 
           ++ (map hprintProcessDef pdefs)
           ++ ("" : [ hprint parsedComp ] ) 


{-| Prints a rate specification in a human readable fashion -}
hprintRateSpec :: RateSpec -> String
hprintRateSpec (ident, rExpr) =
    unwords  [ hprintRateIdentifier ident
             , "="
             , hprint rExpr
             , ";"
             ]

{-| Prints in a human readable fashion, a process definition -}
hprintProcessDef :: ProcessDef -> String
hprintProcessDef (ident, comp) =
    unwords  [ hprint ident
             , "="
             , hprint comp
             , ";"
             ]


instance HumanPrint ParsedComponent where
    hprint = hPrintParsedComponent

{-
  [@todo@] What to do about parentheses here? sometimes they are required.
           I think we should, as stated above use 'Text.PrettyPrint' to write
           a proper unparser, and then just call that from here}
-}
hPrintParsedComponent :: ParsedComponent -> String
hPrintParsedComponent (IdProcess ident)                      = 
    hprint ident
hPrintParsedComponent (StopProcess)                          =
   "Stop"
hPrintParsedComponent (ComponentSum left right)              = 
    concat [ -- "(",
             unwords [ hprint left, "+"  ]
           , "\n"
           , unwords [ "    ", hprint right ]
           -- , ")"
           ]
hPrintParsedComponent (PrefixComponent trans next)           =
    unwords [ hprintTrans
            , "."
            , hprint next
            ]
    where
    hprintTrans :: String
    hprintTrans =
       case pepaTransRate trans of
          RateImmediate Nothing -> hprint $ pepaTransAction trans
          rate                  ->
            concat [ "( "
                   , hprint $ pepaTransAction trans
                   , ", "
                   , hprint rate
                   , ")"
                   ]

hPrintParsedComponent (CondBehaviour cond next)              =
   unwords [ "if"
           , hPrintCondBehaviour cond
           , "then"
           , hPrintParsedComponent next
           ]
hPrintParsedComponent (Cooperation left [] right)            = 
    unwords [ "(" 
            , hprint left
            , "<>"
            , hprint right
            , ")"
            ]
hPrintParsedComponent (Cooperation left actions right)       = 
    unwords [ "("
            , hprint left
            , "<"
            , mkCSlist $ map hprint actions
            , ">"
            , hprint right
            , ")"
            ]
hPrintParsedComponent (ProcessArray ident size Nothing)      = 
    concat [ hprint ident
           , "["
           , show size
           , "]"
           ]
hPrintParsedComponent (ProcessArray ident size (Just a))     = 
    concat [ hprint ident
           , "["
           , show size
           , "]["
           , mkCSlist $ map hprint a
           , "]"
           ]
hPrintParsedComponent (Hiding (IdProcess ident) actions)     = 
    concat [ hprint ident
           , "/{"
           , hprintActionList actions
           , "}"
           ]
hPrintParsedComponent (Hiding comp actions)              = 
    concat [ "("
           , hprint comp
           , ")/{"
           , mkCSlist $ map hprint actions
           , "}"
           ]

hprintActionList :: [ ParsedAction ] -> String
hprintActionList = mkCSlist . (map hprint)

hPrintCondBehaviour :: BeCondition -> String
hPrintCondBehaviour (BeCondition rateId op rateExp) =
   unwords [ hprintRateIdentifier rateId
           , op
           , hprint rateExp
           ]

instance HumanPrint ParsedRate where
   hprint = hprintParsedRate

hprintParsedRate :: ParsedRate -> String
hprintParsedRate (RateTop)                 = "infty"
hprintParsedRate (RateImmediate Nothing)   = "immediate"
hprintParsedRate (RateImmediate (Just x))  = (show x) ++ ":immediate"
hprintParsedRate (RateTimed e)             = hPrintParsedRateExp e

instance HumanPrint ParsedRateExp where
    hprint = hPrintParsedRateExp

hPrintParsedRateExp :: ParsedRateExp -> String
hPrintParsedRateExp (RateId ident Nothing)    = 
    hprintRateIdentifier ident
hPrintParsedRateExp (RateId ident (Just _))   = 
    hprintRateIdentifier ident
hPrintParsedRateExp (RateNum val)             = 
    hprint val
hPrintParsedRateExp (RateProc ident)          = 
    hprint ident
hPrintParsedRateExp (RateBexp oper lexp rexp) =
    concat [ hprint lexp 
           , " "
           , oper
           , " "
           , hprint rexp
           ] 
hPrintParsedRateExp (RateCond cond texp eexp) =
    unwords [ "( ("
            , hprint cond
            , ") ? ("
            , hprint texp
            , ") : ("
            , hprint eexp
            , ") )"
            ]
hPrintParsedRateExp (RateDisabled)            =
   "0.0"


{-
   Printing a rate identifier. Because we may have scoped to get this far
   the correct thing to do is to uncapitalise the first letter before printing
   out.
-}
hprintRateIdentifier :: RateIdentifier -> String
hprintRateIdentifier =
   uncapitaliseFirst . hprint    

uncapitaliseFirst :: String -> String
uncapitaliseFirst []         = []
uncapitaliseFirst (h : rest) = (toLower h) : rest

instance HumanPrint ParsedCond where
    hprint = hPrintParsedCond

hPrintParsedCond :: ParsedCond -> String
hPrintParsedCond (PCondId ident)            = 
    hprint ident
hPrintParsedCond (PCondSum left right)      =
    unwords [ "("
            , hprint left
            , "+"
            , hprint right
            , ")"
            ]
hPrintParsedCond (PCondRate ident oper rExp) =
   unwords [ "("
           , hprintRateIdentifier ident
           , oper
           , hPrintParsedRateExp rExp
           , ")"
           ]

instance HumanPrint ParsedAction where
    hprint = hPrintParsedAction

hPrintParsedAction :: ParsedAction -> String
hPrintParsedAction (Action ident)    = hprint ident
hPrintParsedAction (Tau    ident)    = "`" ++ (hprint ident)
hPrintParsedAction (ComAction ident) = hprint ident -- hmmm not sure about this.

instance HumanPrint Double where
    hprint = myShowFloat

-- My version of show float basically attempts to print out the value
-- using 'showFFloat' but since this will print out @0.7@ as
-- @0.700000000@ (if you have the digits value set to 9) I chop off
-- the spurious zeros.
myShowFloat :: Double -> String
myShowFloat d = 
   dropSpuriousZeros s
   where
   s = showFFloat (Just 9) d "" 

   dropSpuriousZeros :: String -> String
   dropSpuriousZeros []            = []
   dropSpuriousZeros ('.' : rest)
      | all (== '0') rest          = ".0"
      | otherwise                  = '.' : (dropFromEnd (== '0') rest)
   dropSpuriousZeros (h : rest)    = h : (dropSpuriousZeros rest)



{-------------------------------------------------------
 The Analysis Print functions
--------------------------------------------------------}
instance HumanPrint AnalysisMessage where
    hprint = hprintAnMessage

instance HumanPrint AnalysisResult where
    hprint = hprintAnResult

instance HumanPrint AnalysisReport where
    hprint = hprintAnReport

hprintAnMessage :: AnalysisMessage -> String
hprintAnMessage (AnalysisError message)   = "Error: " ++ hprint message
hprintAnMessage (AnalysisWarning message) = "Warning: " ++ hprint message

hprintAnResult :: AnalysisResult -> String
hprintAnResult ( ExcessActions par acts1 acts2 )       =
    unwords [ "The process:"
            , hprint par
            , "cooperates over the actions"
            , hprintActionCooperation acts1
            , "but neither side performs the actions:"
            , hprintActionCooperation acts2
            ]    
hprintAnResult ( ExcessActionsLeft par acts1 acts2 )   =
    unwords [ "The process:"
            , hprint par
            , "cooperates over the actions"
            , hprintActionCooperation acts1
            , "but the left hand side does not perform the actions:"
            , hprintActionCooperation acts2
            ]
hprintAnResult ( ExcessActionsRight par acts1 acts2 )  =
    unwords [ "The process:"
            , hprint par
            , "cooperates over the actions"
            , hprintActionCooperation acts1
            , "but the right hand side does not perform the actions:"
            , hprintActionCooperation acts2
            ]
hprintAnResult ( NeedlessActionsHide par acts1 acts2 ) =
    unwords [ "The process:"
            , hprint par
            , "hides the actions"
            , hprintActionCooperation acts1
            , "but does not ever perform the actions:"
            , hprintActionCooperation acts2
            ]
hprintAnResult ( UsedButUndefinedProcesses pids )      =
    unwords [ "The following process identifers are used but not defined"
            , hprintProcessIdList pids
            ]
hprintAnResult ( DefinedButNotUsedProcesses pids )     =
    unwords [ "The following process identifers are defined but not used"
            , hprintProcessIdList pids
            ]
hprintAnResult ( DefinedMultiplyProcesses pids )       =
    unwords [ "The following process identifers are defined more than once"
            , hprintProcessIdList pids
            ]
hprintAnResult ( UsedButUndefinedRates rids )      =
    unwords [ "The following rate identifers are used but not defined"
            , hprintProcessIdList rids
            ]
hprintAnResult ( DefinedButNotUsedRates rids )     =
    unwords [ "The following rate identifers are defined but not used"
            , hprintProcessIdList rids
            ]
hprintAnResult ( DefinedMultiplyRates rids )       =
    unwords [ "The following rate identifers are defined more than once"
            , hprintProcessIdList rids
            ]

hprintAnResult ( AliasLoop pids )                      =
    unwords [ "The definitions of the following process identifiers"
            , "form an alias loop"
            , hprintProcessIdList pids
            ]
hprintAnResult ( SystemSelfLoop par acts )             =
    unwords [ "The main system equation :"
            , hprint par
            , "may perform a self loop over the actions"
            , hprintActionCooperation acts
            ]
hprintAnResult (TypingError s)                         =
    s

hprintActionCooperation :: [ ParsedAction ] -> String
hprintActionCooperation acts = unwords [ "<", hprintActionList acts, ">" ]

hprintProcessIdList :: [ ParsedComponentId ] -> String
hprintProcessIdList = mkCSlist . (map hprint)


{-|
    Print an entire analysis report.
    [@todo@] consider moving this into the 'Language.Pepa.Analysis.Analysis' 
    module since I think there it would be easier to maintain.
    To do this though I think
    we would need to restructure things a little, in particular the
    /Analysis/ module would depend on this module rather than vice versa
    and so the 'hprintMessages' function below would need to be moved
    there as well.
-}
hprintAnReport :: AnalysisReport -> String
hprintAnReport report = 
    unlines [ if processArraysPresent report
              then "The model contains process arrays"
              else "The model does not contain any process arrays"
            , pActionsMap
            , "The following actions are performed (and not hidden)"
              ++ " by the main composition " ++
              (hprintActionCooperation $ compositionActions report)
            , hprintMessages $ messagesOfReport report
            ]
    where
    pActionsMap = unlines $ map hprintPidActPair 
                    (toList $ processActionsMap report)

    hprintPidActPair :: (ParsedComponentId, [ ParsedAction ] ) -> String
    hprintPidActPair (pid, actions) =
        unwords [ hprint pid
                , ":"
                , hprintActionCooperation actions
                ]


{-|
    Human print just the messages, that is the warnings and errors
    of a full analysis report.
-}
hprintMessages :: [ AnalysisMessage ] -> String
hprintMessages messages =
    unlines [ "There were " ++ show noErrors ++ " errors."
            , "There were " ++ show noWarnings  ++ " warnings."
            , "The warnings were:"
            , warningLines
            , "The errors were :"
            , errorLines
            ]
    where
    errorLines   = unlines $ map hprint errors
    errors       = fst warnsErrs
    noErrors     = length errors
    warnings     = snd warnsErrs
    noWarnings   = length warnings
    warningLines = unlines $ map hprint warnings
    warnsErrs    = splitMessages messages 
