{-| 
    This module defines the stochastic probes which are used to define
    performance measures over pepa models.
-}
module Language.Pepa.Probes.StochasticProbes
   ( ProbeR        ( .. )
   , ProbeDef
   , probeAlphabet
   , firstNames

   -- * Probe analysis functions
   , containsNeedlessRep
   , isReplicatorProbe
   , endsInReplicatorProbe
   )
where

{- Imported Standard Libraries -}
import Data.Set
   ( Set
   , union
   , singleton
   , insert
   )
{- Imported External Libraries -}
{- Imported Local Libraries -}

import Language.Pepa.Syntax
    ( ActionIdentifier
    , ParsedComponentId
    )
{- End of Imports -}


{-| 
   A probe definition is simply a top level (of the grammar)
   stochastic probe.
-}
type ProbeDef = (Maybe ParsedComponentId, ProbeR)

{-|
  A stochastic probe has the grammar:
  @
  R ::= a | T,T | (S)
  S ::= T | T|S
  T ::= R | R{n} | R{m,n} | R+ | R* | R? | R\act
  a ::= act | act:stop | act:start
  @
  
  The following data type definitions follow this specification
  pretty closely, although we greatly simplify this because
  we don't need the different levels of probes they can all be
  defined as one (a T can be an R anyway so we lose nothing in
  the way of type checking here, well that's not quite true the
  parser may benefit from defining ProbeR and ProbeS and ProbeT
  types, but not *that* much).
-}
data ProbeR = 
     SPact    ActionIdentifier        -- ^ Simple action\/activity observation
   | SPlabel  ProbeR ActionIdentifier -- ^  R \: label 
   | SPseq    ProbeR ProbeR           -- ^  R1, R2 
   | SPchoice ProbeR ProbeR           -- ^  (R1 | R2) 
   | SPiter   ProbeR Int              -- ^  R{n} 
   | SPbit    ProbeR (Int, Int)       -- ^  R{m,n} 
   | SPoom    ProbeR                  -- ^  R+ 
   | SPzom    ProbeR                  -- ^  R* 
   | SPzoo    ProbeR                  -- ^  R? 
   | SPwo     ProbeR ActionIdentifier -- ^  R \/ act
   deriving (Show, Eq)  -- shouldn't really derive


{-|
  The alphabet of a probe is the actions over which it acts.
-}
probeAlphabet :: ProbeR -> Set ActionIdentifier
probeAlphabet (SPact act)      = singleton act
probeAlphabet (SPlabel r _)    = probeAlphabet r
probeAlphabet (SPchoice r1 r2) = union (probeAlphabet r1) (probeAlphabet r2)
probeAlphabet (SPseq    r1 r2) = union (probeAlphabet r1) (probeAlphabet r2)
probeAlphabet (SPiter r1 _)    = probeAlphabet r1
probeAlphabet (SPbit r1 _)     = probeAlphabet r1
probeAlphabet (SPoom r1)       = probeAlphabet r1
probeAlphabet (SPzom r1)       = probeAlphabet r1
probeAlphabet (SPzoo r1)       = probeAlphabet r1
probeAlphabet (SPwo  r1 act)   = insert act $ probeAlphabet r1

{-|
    Computes the set of actions names which are explicitly 
    used at the beginning of a stochastic probe.
-}
firstNames :: ProbeR -> Set ActionIdentifier 
firstNames (SPact act)                  = singleton act
firstNames (SPlabel r _)                = firstNames r
firstNames (SPseq r1 r2)
   | canBeEmptySeq r1 = union (firstNames r1) (firstNames r2)
   | otherwise        = firstNames r1
firstNames (SPchoice r1 r2)             = union (firstNames r1) (firstNames r2)
firstNames (SPiter r1 _)                = firstNames r1
firstNames (SPbit r1 _)                 = firstNames r1
firstNames (SPzoo r1)                   = firstNames r1
firstNames (SPoom r1)                   = firstNames r1
firstNames (SPzom r1)                   = firstNames r1
firstNames (SPwo r1 _)                  = firstNames r1

{-|
   Returns true if the given probe may be the empty sequence
-}
canBeEmptySeq :: ProbeR -> Bool
canBeEmptySeq (SPzom _)        = True
canBeEmptySeq (SPzoo _)        = True
canBeEmptySeq (SPiter _ 0)     = True
canBeEmptySeq (SPbit _ (0, _)) = True
canBeEmptySeq (SPbit _ (_, 0)) = True
canBeEmptySeq (SPchoice r1 r2) = (canBeEmptySeq r1) || (canBeEmptySeq r2)
canBeEmptySeq (SPseq r1 r2)    = (canBeEmptySeq r1) && (canBeEmptySeq r2)
canBeEmptySeq (SPwo r1 _)      = canBeEmptySeq r1
canBeEmptySeq _                = False


{-|
   Returns true if the probe is determined to contain a sub-probe
   which may perform an immediate communication directly after
   a replicator probe (such that the choice makes no sense).
-}
containsNeedlessRep :: ProbeR -> Bool
containsNeedlessRep (SPlabel r _)  = 
   endsInReplicatorProbe r || containsNeedlessRep r
containsNeedlessRep (SPact _)      = False
containsNeedlessRep (SPseq l r)    =
   containsNeedlessRep l || containsNeedlessRep r
containsNeedlessRep (SPchoice l r) =
   containsNeedlessRep l || containsNeedlessRep r
containsNeedlessRep (SPiter r _)   =
   containsNeedlessRep r
containsNeedlessRep (SPbit r _)    =
   containsNeedlessRep r
containsNeedlessRep (SPoom r)      =
   containsNeedlessRep r
containsNeedlessRep (SPzom r)      =
   containsNeedlessRep r
containsNeedlessRep (SPzoo r)      =
   containsNeedlessRep r
containsNeedlessRep (SPwo r _)     =
   containsNeedlessRep r

{-| This returns true if the probe is a replicator probe.
   You almost never wish to use this on its own though.
   For example @(a* | b)@ returns @False@ even though the
   left hand is a replicator.
   So you would normally want a function such as @maybeReplicator@
   or 'endsInReplicatorProbe'
-}
isReplicatorProbe :: ProbeR -> Bool
isReplicatorProbe (SPact _)      = False
isReplicatorProbe (SPlabel  _ _) = False
isReplicatorProbe (SPseq _ _)    = False
isReplicatorProbe (SPchoice _ _) = False
isReplicatorProbe (SPiter _ _)   = True
isReplicatorProbe (SPbit  _ _)   = True
isReplicatorProbe (SPoom  _ )    = True
isReplicatorProbe (SPzom  _ )    = True
isReplicatorProbe (SPzoo  _ )    = True
isReplicatorProbe (SPwo r _ )    = isReplicatorProbe r

{-|
   Returns true if this probe may end in a replicator probe.
-}
endsInReplicatorProbe :: ProbeR -> Bool
endsInReplicatorProbe (SPseq _ r)    = endsInReplicatorProbe r
endsInReplicatorProbe (SPchoice l r) =
   endsInReplicatorProbe l || endsInReplicatorProbe r
endsInReplicatorProbe (SPact _)      = False
endsInReplicatorProbe (SPlabel  _ _) = False
endsInReplicatorProbe (SPiter _ _)   = True
endsInReplicatorProbe (SPbit  _ _)   = True
endsInReplicatorProbe (SPoom  _ )    = True
endsInReplicatorProbe (SPzom  _ )    = True
endsInReplicatorProbe (SPzoo  _ )    = True
endsInReplicatorProbe (SPwo r _ )    = endsInReplicatorProbe r
