{-| 
    This module defines the translation from probe specifications to
    pepa component definitions
-}
module Language.Pepa.Probes.Translate
   ( translateProbeR
   , probeToDotFile
   , ProbeTranslateFlags     ( .. )
   , comActionPriority
   )
where

{- Imported Standard Libraries -}
import Prelude hiding
   ( any )
import Data.Foldable
   ( any )
import Data.List
   ( find )
import Data.Set
   ( fromList
   , toList
   , union
   , singleton
   )
{- Imported External Libraries -}
{- Imported Local Libraries -}

import Language.Pepa.Syntax
   ( QualifiedName      ( .. )
   , ActionIdentifier
   
   , ProcessDef
   , ParsedTrans        ( .. )
   , ParsedPriority
   , defaultPepaPriority
   , ParsedComponent    ( .. )
   , ParsedComponentId
   , ParsedRate         ( .. )
   , ParsedAction       ( .. )
   
   , involvesAction
   )
import Language.Pepa.Print
   ( HumanPrint    ( .. ) )
import Language.Pepa.Utils
   ( mkCSlist
   , fullGroupBy
   )
import Language.Pepa.Probes.StochasticProbes
   ( ProbeR      ( .. )
   , probeAlphabet
   )
import Language.Pepa.Probes.StateMachines.NfaTypes
   ( Move        ( .. )
   , Nfa         ( .. )
   , NfaLabel    ( .. )

   , isMoveFrom
   )
import Language.Pepa.Probes.StateMachines.NfaCombinators
   ( nfaOr
   , nfaThen
   , nfaStar
   , nfaQuestion
   , nfaReset
   )
import Language.Pepa.Probes.StateMachines.NfaLib
   ( movesFromState )
import Language.Pepa.Probes.StateMachines.NfaToDfa
   ( makeDeterministic )
import Language.Pepa.Probes.StateMachines.MinimiseDfa
   ( minimiseDfa )
{- End of Imports -}

{-|
   The type of flags which modify how the translation is done.
-}
data ProbeTranslateFlags = 
   -- | Do not minimise the dfa
   NoMinimise
   -- | Add the self-loops before minimisation
 | MinSelfLoops
   -- | Minimise before adding the self-loops
 | NoMinSelfLoops
   -- | Do not add any self-loops
 | NoSelfLoops
   -- | Leave the state machine as non-deterministic 
 | NonDeterm  
   deriving Eq


{-| The main exported function translates a probe into a set
    of pepa definitions.
-}
translateProbeR ::
      [ ProbeTranslateFlags ] -- ^ The flags which control the translation
   -> String                  -- ^ The name of the probe
   -> [ ParsedTrans ]         -- ^ The transitions used to get the right
                              --   action kinds (immediate or top)
   -> ProbeR                  -- ^ The probe to be translated
   -> [ ProcessDef ]          -- ^ The returned pepa process definitions
translateProbeR trFlags probeName transitions probe =
   dfaToPepa transitions probeName $ 
   probeToMinimisedDfa trFlags probe

{-|
   This translates a given probe into a dot file format.
-}
probeToDotFile ::  [ ProbeTranslateFlags ] -> String -> ProbeR -> String
probeToDotFile trFlags probeName probe =
   unlines $ [ "digraph " ++ probeName ++" {" ]
             ++ stateLines ++ moveLines ++ [ "}" ]
   where
   NFA states moves _start _accepts = probeToMinimisedDfa trFlags probe
   stateLines = map toStateLine $ toList states
   moveLines  = map toArrowLine $ map showTransGroup moveGroups

   -- Groups the transitions so that those with the same source
   -- source and target states are depicted on the same line.
   moveGroups :: [ [ ProbeMove ] ]
   moveGroups = fullGroupBy sameSourceAndTarget $ toList moves

   toStateLine :: ProbeState -> String
   toStateLine i = concat [ "\""
                          , show i
                          , "\""
                          , " [ shape = circle , color = red ] ;"
                          ]

   toArrowLine :: (String, String, String) -> String
   toArrowLine (s1, l, s2) = concat [ "\""
                                    , s1
                                    , "\" -> \""
                                    , s2
                                    , "\" [ label = \""
                                    , l
                                    , "\" ] ; "
                                    ]

   sameSourceAndTarget :: ProbeMove -> ProbeMove -> Bool
   sameSourceAndTarget s1 s2 = (sourceOf s1 == sourceOf s2)
                               && (targetOf s1 == targetOf s2)
   sourceOf :: ProbeMove -> ProbeState
   sourceOf (Move s _ _) = s
   sourceOf (Emove s _ ) = s

   targetOf :: ProbeMove -> ProbeState
   targetOf (Move _ _ s) = s
   targetOf (Emove _  s) = s

   labelOf :: ProbeMove -> String
   labelOf (Move _ l _) = hprint l
   labelOf (Emove _ _ ) = "Epsilon"

   -- Turns a group of transitions into a single line for the dot
   -- file so that we have a single arrow with a comma separated list
   -- of labels rather than multiple arrows with the same source
   -- and target states.
   showTransGroup :: [ ProbeMove ] -> (String, String, String)
   showTransGroup []          = ("", "error: empty group", "")
   showTransGroup (s1 : rest) =
      ( show $ sourceOf s1
      , mkCSlist $ map labelOf (s1 : rest)
      , show $ targetOf s1
      ) 
       

{-
   This function does the common work of 'translateProbeR'
   and 'probeToDotFile'. This does mean that if a program wishes
   to do *both* to a dot file and adding the probes to a model then
   it will re-do all of this work. To avoid this would require a little
   re-factoring, but not *that* much. I just don't see it as all that
   likely that someone is going to want a dot-file of a massively complicated
   probe hence even in the case that we do the work twice it won't matter
   that much.
-}
probeToMinimisedDfa :: [ ProbeTranslateFlags ] -> ProbeR -> ProbeNfa
probeToMinimisedDfa trFlags probe =
   localAddSelfMinimise $ localMakeDeterministic cyclicNfa
   where
   -- If we specify no self loops then adding them is the id
   localAddSelfLoops
      | elem NoSelfLoops trFlags = id
      | otherwise                = addSelfLoops actionList
   -- if we specify no minimise then the minimise function is the id
   localMinimise
      | elem NoMinimise trFlags  = id
      | otherwise                = minimiseDfa
   -- The order of minimise and add-self-loops depends on whether we
   -- should minimise the self-loops or not.
   localAddSelfMinimise
      | elem MinSelfLoops trFlags   = localMinimise . localAddSelfLoops
      | elem NoMinSelfLoops trFlags = localAddSelfLoops . localMinimise
      | otherwise                   = localAddSelfLoops . localMinimise

   localMakeDeterministic
      | elem NonDeterm trFlags    = id
      | otherwise                 = makeDeterministic
   cyclicNfa   =  makeCyclic $ buildNfa probe
   actionList  = toList $ probeAlphabet probe

{- Define NFAs -}

{- Our NFAs are NFAs where the states are integers and the 
   labels on the transitions are 'ActionIdentifier's
   
   Okay states *should* be this, but for now we just let them
   be integers and solve the problem of adding self loops 
-}
-- data ProbeState = Communication Int
--                 | Ordinary Int
type ProbeState     = Int
type ProbeMoveLabel = ActionIdentifier
type ProbeMove      = Move ProbeState ProbeMoveLabel
type ProbeNfa       = Nfa  ProbeState ProbeMoveLabel
   
instance NfaLabel QualifiedName where
   labelOfString = Unqualified
{- Define DFAs -}

{-
data ProbeR = SPact    ActionIdentifier -- SPaction
            | SPlabel  ProbeR ActionIdentifier -- | R : label
            | SPseq    ProbeR ProbeR           -- | R1, R2
            | SPchoice ProbeR ProbeR           -- | (R1 | R2)
            | SPiter   ProbeR Int              -- | R{n}
            | SPbit    ProbeR (Int, Int)       -- | R{m,n}
            | SPoom    ProbeR                  -- | R+
            | SPzom    ProbeR                  -- | R*
            | SPzoo    ProbeR                  -- | R?
            | SPwo     ProbeR ActionIdentifier 
              deriving (Show, Eq)  -- don't derive just temporary for debugging
-}

{- Translate Probes to NFAs -}

{- Because we want our pepa process to be cyclic, we must add an episilon
   transition from all the terminating (or more rather accepting) states
   to the start state.

   Here I also make the accepting state the start state, since this means
   that it will  'accept' the empty string, this avoids two similar states
   in the end dfa produced.
-}
makeCyclic :: ProbeNfa -> ProbeNfa
makeCyclic (NFA states moves startState accepting) =
   (NFA states newMoves startState $ singleton startState)
   where
   newMoves     = union moves (fromList epsilonMoves)
   
   epsilonMoves = map makeEpsilonMove $ 
                  filter requiresEpsilon (toList accepting)

   makeEpsilonMove :: Int -> ProbeMove
   makeEpsilonMove i = Emove i startState

   requiresEpsilon :: Int -> Bool
   requiresEpsilon i = not $ any isCurrentEpsilon moves
      where
      isCurrentEpsilon :: ProbeMove -> Bool
      isCurrentEpsilon (Move _ _ _) = False
      isCurrentEpsilon (Emove l m)  =
         l == i && m == startState

buildNfa :: ProbeR -> ProbeNfa
buildNfa (SPact a)        = 
   NFA ( fromList [ state0, state1 ] )
       ( singleton $ Move state0 a state1 )
       state0
       ( singleton state1 )
   where
   state0 = 0 -- Ordinary 0
   state1 = 1 -- Ordinary 1
buildNfa (SPlabel r a)    = 
   nfaThen (buildNfa r) (NFA ( fromList [ state0, state1 ] )
                            ( singleton $ Move state0 a state1 )
                            state0
                            ( singleton state1 )
                       )
   where
   state0 = 0 -- Communication 0
   state1 = 1 -- Ordinary 1
buildNfa (SPseq r1 r2)    = nfaThen (buildNfa r1) (buildNfa r2)
buildNfa (SPchoice r1 r2) = nfaOr (buildNfa r1) (buildNfa r2)
buildNfa (SPzom r)        = nfaStar $ buildNfa r
buildNfa (SPoom r)        = buildNfa (SPseq r $ SPzom r)
buildNfa (SPiter r n)     = iterNdfa n $ buildNfa r
buildNfa (SPzoo r)        = nfaQuestion $ buildNfa r
buildNfa (SPwo r a )      = nfaReset a $ buildNfa r

buildNfa (SPbit r (m, n))
   | m > n     = buildNfa $ SPbit r (n, m)
   | m == n    = buildNfa $ SPiter r n
   | otherwise = nfaThen (iterNdfa m rNdfa) $ buildNfa rQnmTimes
   where
   rNdfa     = buildNfa r
   rQnmTimes = SPiter rQ (n - m) -- iterNdfa (m - n) $ nfaQuestion rNdfa
   rQ        = SPzoo r

iterNdfa :: Int -> ProbeNfa -> ProbeNfa
iterNdfa n r
   | n < 1     = error "an iter probe with n < 1"
   | n == 1    = r
   | otherwise = nfaThen r $ iterNdfa (n - 1) r

{- Add self-loops to DFAs -}


addSelfLoops :: [ ActionIdentifier ] -> ProbeNfa -> ProbeNfa
addSelfLoops alphabet machine@(NFA states moves initial accepts) =
   (NFA states newMoveSet initial accepts)
   where
   newMoveSet = union moves $ fromList newMoves
   newMoves   = concat $ map selfLoopsOneState (toList states)

   -- Recall that we should only add the self-loops if the state
   -- does not represent one in which immediate communication can
   -- take place from. Since the immediate communication must take
   -- precedence over observing the model.
   selfLoopsOneState :: Int -> [ ProbeMove ]
   selfLoopsOneState i
      | shouldAddSelfLoops = []
      | otherwise          =
         map makeSelfLoop $ filter (not . fromIwithA) alphabet
      where
      makeSelfLoop :: ActionIdentifier -> ProbeMove
      makeSelfLoop a = Move i a i
 
   
      -- Determines if a move goes from the current state via
      -- the given action
      fromIwithA :: ActionIdentifier -> Bool
      fromIwithA a = any moveExists $ movesFromState machine i
         where
         moveExists :: ProbeMove -> Bool
         moveExists (Move _ b _) = b == a
         moveExists (Emove _ _)  = False

      -- Determines if (all/any which should it be??) 
      -- the moves from one state are an
      -- immediate communication, in which case we shouldn't add
      -- any self-loops
      shouldAddSelfLoops :: Bool
      shouldAddSelfLoops = isImmediateState alphabet machine i

         
{-
   This determines whether or not the state is one in which immediate
   communication can/should take place. This is useful for two reasons,
   first of all we wish not to add self-loops in such a state and second
   of all we wish to make such a state an accepting state for the purposes
   of minimisation.
-}
isImmediateState :: [ ActionIdentifier ] -> ProbeNfa -> ProbeState -> Bool
isImmediateState alphabet machine state =
   any isImmediateCommunication movesFromCurrentState
   where
   movesFromCurrentState = movesFromState machine state

   -- Determines if a move is an immediate communication, this
   -- is not well done I must admit, we just use the fact that
   -- the immediate communication labels should *not* be in
   -- the alphabet of the probe.
   isImmediateCommunication :: ProbeMove -> Bool
   isImmediateCommunication (Move _ name _) =
      not $ elem name alphabet
   isImmediateCommunication (Emove _ _    ) = False

   
{- Translate DFAs to Pepa defs -}
{- Note here the list of parsed transitions, this should include
   all the transitions in the model plus all the labels of the
   probe mapped to immediate communication actions. This way we know
   how to add the action names of the Dfa to the probe definitions
   as transitions.
-}
dfaToPepa :: [ ParsedTrans ] -- The transitions in the model
          -> String          -- The name of the probe
          -> ProbeNfa        -- The Dfa representation of the probe
          -> [ ProcessDef ]
dfaToPepa transitions probeName (NFA states moves startState _) =
   map defOfState $ toList states
   where
   moveList = toList moves
   defOfState :: Int -> ProcessDef
   defOfState i = ( makeProcessName i
                  , foldr1 ComponentSum prefixes
                  )
                  where
                  prefixes = map prefixOfMove iMoves
                  iMoves   = filter (isMoveFrom i) moveList
   
   prefixOfMove :: ProbeMove -> ParsedComponent
   prefixOfMove (Move _ a j) =
      PrefixComponent (transOfAction a)
                      (IdProcess $ makeProcessName j)
   prefixOfMove (Emove _ _)  =
      error "Dfa contains an epsilon move"

   -- If it is the start state then it should have the same name
   -- as the probe.
   makeProcessName :: Int -> ParsedComponentId
   makeProcessName i
      | i == startState = Unqualified probeName
      | otherwise       = Unqualified $ probeName ++ (show i)
   {- This looks up how to add the transition as an immediate
      action or as a timed (with infty) activity.
   -}
   transOfAction :: ActionIdentifier -> ParsedTrans
   transOfAction actionId =
      case find (involvesAction actionId) transitions of
         Just t  -> t
         -- If the action isn't in the list of transitions for
         -- the model then it must be a communication activity
         Nothing -> ParsedTrans { pepaTransAction   = ComAction actionId
                                , pepaTransPriority = comActionPriority
                                , pepaTransRate     = RateImmediate Nothing
                                }

comActionPriority :: ParsedPriority
comActionPriority = defaultPepaPriority