{-|
   A suite of data types describing the abstract syntax of pepa models.
-}
module Language.Pepa.Syntax
   ( ShowOrig             ( .. )
   -- * Identifier Types
   , Identifier -- A generic identifier
   , RateIdentifier
   , ActionIdentifier
   , ProcessIdentifier
   , QualifiedName        ( .. )
   , getOrigName
   , sameOrigName
   , qualifyName
   , qualifyQName
   , suffixQName
   , prefixQName
   , concatQNames
   
   -- * Parsed Syntax
   , ParsedModel          ( .. )
   , RateSpec
   , ProcessDef
   , ParsedDefinition     ( .. )
   , splitDefinitions
   
   , ParsedComposition
   , ParsedComponent      ( .. )
   , BeCondition          ( .. )
   , ProcessConcentration
   , ParsedTrans          ( .. )
   , ParsedPriority
   , defaultPepaPriority
   , defaultIncreasedPriority

   , actionNameOfTrans
   , involvesAction
   , ParsedValue
   , ParsedComponentId
   , ConcentrationSpec
   , RateWeight
   , ParsedRate           ( .. )
   , ParsedRateExp        ( .. )
   , isTimedRate
   , isImmediateRate
   , isPassiveRate
   , ParsedCond           ( .. )
   , ParsedAction         ( .. )
   , makeTauAction
   , nameOfAction
   ) 
   where

{- Imported Standard Libraries -}
{- Imported Local Libraries -}
{- End of Imports -}

{-|
  The 'ShowOrig' type class is similar to the @Show@ class.
  However it is used where we want to give a message to the user.
  Often a parsed piece of pepa syntax will be manipulated in someway
  to allow it to be more managable in the rest of the compiler.
  If we want to print some information about it to the user (say an
  error message involving it) then showing the mangled version will
  not be very helpful. So we want to show the entity as it was originally
  seen by the user. Of course for a type to implement this class
  properly then it must be able to 'get back' the original form, this
  will most often be done simply by carrying the original form
  around with it.
-}
class ShowOrig a where
    showOrig :: a -> String


{-|
    A generic Identifier type, this is used for all names in a pepa
    model which are either not rates, actions or processes or not yet
    known to be one of those.
-}
type Identifier        = String 

-- | The type associated with rate names
type RateIdentifier    = QualifiedName

-- | The type associated with activity names
type ActionIdentifier  = QualifiedName

-- | The type associated with process identifiers
type ProcessIdentifier = QualifiedName

{-|
  A qualified name is an identifier together with a distinguishing
  integer. We also allow there to be no qualifying number.
  It used to be that we could just represent these with the simple
  identifier, to qualify the identifier @ident@ we just did
  @ ident ++ (show i) @, where @i@ was the qualifying integer.
  However we now wish to be able to check qualified names against
  their original name, and also I think this way is better since we
  know whether or not we have already qualified the name.

  We also allow then mangling of names, this generally involves
  the addition of a prefix or suffix.
  Currently we cannot mangle and qualify a qualified name.
-}
data QualifiedName = Qualified   Identifier [ Int ]
		   | Mangled     String Identifier
		   | Unqualified Identifier
		   | NameSpaceId Identifier QualifiedName
		     deriving (Eq, Ord, Show)


{-|
  Show the name as it was originally parsed, this is most useful for
  giving output to the user.
-}
instance ShowOrig QualifiedName where
    showOrig (Qualified ident _)           = ident
    showOrig (Mangled _ ident)             = ident
    showOrig (Unqualified ident)           = ident
    showOrig (NameSpaceId namespace qname) = concat [ namespace
                                                    , "::"
                                                    , showOrig qname
                                                    ]

getOrigName :: QualifiedName -> Identifier
getOrigName = showOrig

sameOrigName :: QualifiedName -> QualifiedName -> Bool
sameOrigName qn1 qn2 = (showOrig qn1) == (showOrig qn2)


{-|
  Takes a simple string identifier and qualifies it with the given integer.
-}
qualifyName :: Identifier -> Int -> QualifiedName
qualifyName c i 
    | i < 0     = error "qualifyName: negative argument"
    | otherwise = Qualified c [i]


{-| Further qualify an already qualified name.
    Note that the qualified name may not actually be qualifed,
    that is it may have been created with 'Unqualified'
-}
qualifyQName :: QualifiedName -> Int -> QualifiedName 
qualifyQName (Unqualified ident ) n   = Qualified ident [n]
qualifyQName (Qualified ident il) n   = Qualified ident $ il ++ [n]
qualifyQName (NameSpaceId ident qn) n = NameSpaceId ident $ qualifyQName qn n
qualifyQName name@(Mangled _ _)     _ =
    error $ "qualifyModel: Attempt to qualify a mangled " ++
		  "name: " ++ (show name)


{-|
  Add a suffix to a qualified name, useful for obscuring the name in the
  output so that it won't clash with other names, for example appending
  "PEPA" to names.
-}
suffixQName :: QualifiedName -> String -> QualifiedName
suffixQName (Qualified ident i)    suffix = Qualified (ident ++ suffix) i
suffixQName (Mangled prefix ident) suffix = Mangled prefix (ident ++ suffix)
suffixQName (Unqualified ident)    suffix = Unqualified (ident ++ suffix)
suffixQName (NameSpaceId ident qn) suffix = 
   NameSpaceId ident $ suffixQName qn suffix

{-|
  The same as 'suffixQName' adds a prefix instead.
-}
prefixQName :: QualifiedName -> String -> QualifiedName
prefixQName (Qualified ident i)    prefix = Qualified   (prefix ++ ident) i
prefixQName (Mangled p ident)      prefix = Mangled p   (prefix ++ ident)
prefixQName (Unqualified ident)    prefix = Unqualified (prefix ++ ident)
prefixQName (NameSpaceId ident qn) prefix = 
   NameSpaceId ident $ prefixQName qn prefix


{-|
   Concatenates two qualified names putting the separator in
   between. 
-}
concatQNames :: QualifiedName -> String -> QualifiedName -> QualifiedName
concatQNames (Qualified s1 i1) sep (Qualified s2 i2) =
   Qualified (s1 ++ sep ++ s2) (i1 ++ i2)
concatQNames (Unqualified s1) sep (Unqualified s2)   =
   Unqualified $ s1 ++ sep ++ s2
concatQNames (Mangled s1 m1) sep (Mangled s2 m2)
   | m1 == m2  = Mangled (s1 ++ sep ++ s2) m1
   | otherwise = Mangled (s1 ++ sep ++ s2) (m1 ++ m2)
concatQNames (NameSpaceId _ _ ) _ _                  =
   error "attempt to concatenate namespaced names"
concatQNames _ _ (NameSpaceId _ _ )                  =
   error "attempt to concatenate namespaced names"
-- If we are concatenating two different 'kinds' of qualified name
-- then the result is unlikely to dependent on what kind we make
-- so we just combine the original names and make in unqualified
-- this is probably not correct behaviour, but ah well it will do
-- for now.
concatQNames s1 sep s2
   | qualifiers == [] = Unqualified combined
   | otherwise        = Qualified combined qualifiers
   where
   combined = (showOrig s1) ++ sep ++ (showOrig s2)
   qualifiers = (qualifiersOfQName s1) ++ (qualifiersOfQName s2)

qualifiersOfQName :: QualifiedName -> [ Int ]
qualifiersOfQName (Qualified _ is)   = is
qualifiersOfQName (Mangled _ _)      = []
qualifiersOfQName (Unqualified _)    = []
qualifiersOfQName (NameSpaceId _ qn) = qualifiersOfQName qn

{-|
  To date we do not parse in process concentrations but it is on the
  to do list.
-}
type ProcessConcentration = Int

{-| Concentration specification is basically the type which stores
    a pair of the component state, and its initial concentration.
-}
type ConcentrationSpec = (ParsedComponentId, ProcessConcentration)

{-|
  Rate definitions are given on the command line as well as in
  the pepa source file. The rate identifiers are initially mapped
  to rate expressions but hopefully we can reduce them to actual
  values. Of course this is not always possible with the introduction
  of functional rates.
-}
type RateSpec = (RateIdentifier, ParsedRateExp)

{-
  A 'RateDictionary' is used to map rate identifier to
  their values. This is used in the reduction of rate expressions.
  The rate dictionaries themselves must be built up not just from
  the pepa model in the input file but also from the command line
  arguments.
-}
--type RateDictionary = [(RateIdentifier, Double)]

{-|
   A process definition is a process identifier and a parsed component.
-}
type ProcessDef = (ParsedComponentId, ParsedComponent)

{-
  All of these derive the 'Show' class, this is required for printing
  out during testing and also (perhaps) printing the parsed model to
  the log file, however I should actually instantiate the show functions
  using some pretty printing. Well perhaps not, perhaps for debugging the
  derived show is enough, and for pretty printing to the log file a
  separate pretty printing function should be written.
-}

{-|
  The type of a whole parsed pepa model.
-}
data ParsedModel = ParsedModel [ RateSpec ] [ ProcessDef ] ParsedComposition

{-|
  A definition in a pepa model can define one of two things:
  1. A process definition
  2. A rate definition
  We would like to restrict the type a bit more. A process definition
  should consist only of sequential componenents. However we also want
  to allow such definitions as:
  @P = P1 <> P1@
  to aid in the writing of the system component.
  We remove this from the 'ParsedModel' which is returned from the parser
  because we separate them out there. So this is a temporary holder data
  structure.
   
  Update: We now parameterise this type and allow a generic definition
  which could be used to allow many different extensions to Pepa.
-}
data ParsedDefinition a   = ProcessDef ProcessDef
			  | RateDef RateSpec
                          | OtherDef a



{-|
   Splits a list of definitions into three separate lists
   corresponding to the three separate kinds of definitions.
   This is useful during parsing where the definitions may
   come in any order. For a pure pepa model the third list
   should be empty.
-}
splitDefinitions :: [ ParsedDefinition a ] -> ( [ RateSpec ]
                                              , [ ProcessDef ]
                                              , [ a ]
                                              )
splitDefinitions []                       = ([], [], [])
splitDefinitions ((RateDef rd) : rest)    = 
   (\(r,p,s) -> (rd : r, p, s)) $ splitDefinitions rest
splitDefinitions ((ProcessDef pd) : rest) =
   (\(r,p,s) -> (r, pd : p, s)) $ splitDefinitions rest
splitDefinitions ((OtherDef a) : rest)   =
   (\(r,p,s) -> (r, p, a : s)) $ splitDefinitions rest

{-|
    The main pepa composition is simply a component.
-}
type ParsedComposition   = ParsedComponent

{-|
  A parsed parallel component is cooperation between two
  components over a set of actions as in @(P1 <a b> P2)@.
  We now refuse to have sequential components within parallel components.
  It would be nice if we could have a sub-type such that the operands to
  component sums and prefix components could only be sequential components.
  However we require a typing phase anyway since we may have something:
  @
  P  = (a, r).P1 ;
  P1 = Q <> Q
  @
  Hence although syntactically the prefix looks okay, it is not since @P1@
  is actually a parallel component.
  The parser should still disallow obvious nonsense such as
  @ P = (a, r).(Q <> Q) ; @
-}
data ParsedComponent    = 
    IdProcess ParsedComponentId
  | PrefixComponent ParsedTrans ParsedComponent
  | ComponentSum ParsedComponent ParsedComponent
  | CondBehaviour BeCondition ParsedComponent
  | StopProcess
  | Cooperation ParsedComponent [ ParsedAction ] ParsedComponent
  | ProcessArray ParsedComponentId Integer ( Maybe [ ParsedAction ] )
  | Hiding ParsedComponent [ ParsedAction ]
  deriving Show

data BeCondition = 
   BeCondition RateIdentifier String ParsedRateExp
   deriving Show

{-|
  A parsed transition from one component to another, this forms
  part of a prefix component.
-}
data ParsedTrans =
   ParsedTrans { pepaTransAction   :: ParsedAction
               , pepaTransPriority :: ParsedPriority
               , pepaTransRate     :: ParsedRate
               }
               deriving Show


{-| The type of a priority is simply and integer -}
type ParsedPriority = Int

{-| The default pepa priority -}
defaultPepaPriority :: ParsedPriority
defaultPepaPriority = 1

{-| The default increased priority -}
defaultIncreasedPriority :: ParsedPriority
defaultIncreasedPriority = 1000


{-| Returns the name of the action in a 'ParsedTrans' -}
actionNameOfTrans :: ParsedTrans -> ActionIdentifier
actionNameOfTrans = nameOfAction . pepaTransAction

{-| Returns whether or not a given transition involves the
    given action identifier.
-}
involvesAction :: ActionIdentifier -> ParsedTrans -> Bool
involvesAction ident = (ident ==) . actionNameOfTrans

{-
  Finally the base definitions.
-}
type ParsedComponentId = ProcessIdentifier

{-|
  'ParsedAction' is a bit of a misnomer now because they contain actions
  which cannot be parsed such as 'Tau' which may result from hiding and
  'ComAction' which are inserted by the translation of a performance
  specification probe.
-}
data ParsedAction      = Tau       ActionIdentifier
		       | Action    ActionIdentifier
		       | ComAction ActionIdentifier
		        deriving Show
       
instance Eq ParsedAction where
    (==) = equalParsedActions

equalParsedActions :: ParsedAction -> ParsedAction -> Bool
equalParsedActions (Action id1)    (Action id2)    = id1 == id2
equalParsedActions (ComAction id1) (ComAction id2) = id1 == id2
equalParsedActions _               _               = False

makeTauAction :: ParsedAction -> ParsedAction
makeTauAction (Action ident)    = Tau ident
makeTauAction (ComAction ident) = Tau ident
makeTauAction act@(Tau _)       = act

nameOfAction :: ParsedAction -> ActionIdentifier
nameOfAction (Tau       ident) = ident
nameOfAction (Action    ident) = ident
nameOfAction (ComAction ident) = ident


{-|
   A rate can be a rate expression, or an immediate weight, or the
   special rate, infty\/top.
-}
data ParsedRate = 
   RateTop                            -- ^ The special passive rate
 | RateImmediate (Maybe RateWeight)   -- ^ An immediate weight
 | RateTimed ParsedRateExp            -- ^ A timed rate expression
 deriving Show

{-|
  [@todo@]make the operator an actual tagged type rather than a string
  [@todo@]of course this should be defined in @PParser.hs@
  
  Notice here that the 'RateId' constructor takes a second parameter,
  this is the possible rate expression which the identifier represents.
  The reason for this is that we make a transformation turning all
  of the rate expressions within process definitions into a new identifier
  and then put a rate definition defining the new identifier to be equal
  to the original expression. This hampers good error messaging so instead
  we allow us to /remember/ the original expression such that it maybe
  printed out.

  I think this should be separated out as @ParsedRateProcExpr@ and
  @ParsedRateExpr@ so that actually the argument to 'RateProc'
  is a @ParsedRateProcExpr@.

  Currently we do not allow the parsing of weighted (or disabled)
  rate expressions, however they are included as they assist with the
  implementation of calulating apparent rates.

  [@todo@] Consider getting rid of the maybe rate expression tagged
  onto a rate identifier. Do we really need it now?
-}
data ParsedRateExp     = RateId   RateIdentifier (Maybe ParsedRateExp)
		       | RateNum  ParsedValue
		       | RateBexp String ParsedRateExp ParsedRateExp
                       | RateProc ParsedComponentId
		       | RateCond ParsedCond ParsedRateExp ParsedRateExp
                         -- @ RateWeighted ParsedRateExp @
                         -- These are implemented as @0:immediate@ check that this
                         -- is reasonable?
                       | RateDisabled --ParsedRateExp
                       deriving Show
type RateWeight        = Double
type ParsedValue       = Double
data ParsedCond        = PCondId   ParsedComponentId
                       | PCondSum  ParsedCond ParsedCond
                       | PCondRate RateIdentifier String ParsedRateExp
                         deriving Show -- (Eq, Show)

{-|
   This is a very simple function, however should note that if a conditional
   can have 'RateTop' (or an immediate weight) as either branch then this will
   have to check the rate expression for that.
-}
isTimedRate :: ParsedRate -> Bool
isTimedRate (RateTimed _)     = True
isTimedRate (RateTop)         = True
isTimedRate (RateImmediate _) = False

{-| Relates whether the rate is immediate or not -}
isImmediateRate :: ParsedRate -> Bool
isImmediateRate (RateImmediate _) = True
isImmediateRate _                 = False

{-| Relates whether the given rate is passive or not -}
isPassiveRate :: ParsedRate -> Bool
isPassiveRate RateTop = True
isPassiveRate _       = False