{-| 
   This module implements a transformation in which all needlessly
   exported actions are hidden. So if a component within a model
   performs an action which it never cooperates over then it is
   hidden.
-}
module Language.Pepa.Transform.HideAll
   ( hideAllActions )
where

{- Standard Libraries modules imported -}
import Prelude hiding ( lookup )
import Data.List
   ( union )
import Data.Map
   ( lookup )
{- External Library modules imported -}
{- Local modules imported imported -}
import Language.Pepa.Syntax
   ( ParsedModel       ( .. )
   , ParsedComponent   ( .. )
   , ActionIdentifier
   , ParsedAction
   , nameOfAction
   )
import Language.Pepa.Analysis.Analysis
   ( AnalysisReport    ( .. ) )
{- End of imports -}


type CooperationSet = [ ActionIdentifier ]

{-|
   The main function exported by the model.
   It expects there to be no parallel definitions as the definitions
   are left unchanged. This is a simple transformation over the main
   composition which will simply turn process identifiers into
   a hiding component. Such as @P => P\/{a,b}@.
-}
hideAllActions :: AnalysisReport -> ParsedModel -> ParsedModel
hideAllActions report (ParsedModel rDefs pDefs mainComp) =
   ParsedModel rDefs pDefs $ hideComponent [] mainComp
   where
   hideComponent :: CooperationSet -> ParsedComponent -> ParsedComponent
   hideComponent coopSet component@(IdProcess ident)        =
      case lookup ident actionMap of
         Nothing   -> error "process not in the analysis's action map"
         Just aSet -> 
            case createHideSet coopSet aSet of
               [] -> component
               l  -> Hiding component l
   hideComponent coopSet (Cooperation left actions right)   =
      Cooperation (hideComponent newCoopSet left)
                  actions
                  (hideComponent newCoopSet right)
      where
      newCoopSet = union coopSet $ map nameOfAction actions
   hideComponent coopSet (Hiding comp actions)              =
      Hiding (hideComponent newCoopSet comp) actions
      where
      newCoopSet = union coopSet $ map nameOfAction actions
   -- Process arrays are actually just the same as identifiers
   hideComponent coopSet component@(ProcessArray ident _ _) =
      case lookup ident actionMap of
         Nothing   -> error "process not in the analysis's action map"
         Just aSet -> 
            case createHideSet coopSet aSet of
               [] -> component
               l  -> Hiding component l
   -- The sequential components which should not be in the main
   -- composition
   hideComponent _coopSet (PrefixComponent _ _)             =
      error "hideAllActions: sequential component found in main composition"
   hideComponent _coopSet (StopProcess)                     =
      error "hideAllActions: sequential component found in main composition"
   hideComponent _coopSet (CondBehaviour _ _)               =
      error "hideAllActions: sequential component found in main composition"
   hideComponent _coopSet (ComponentSum _ _)                =
      error "hideAllActions: sequential component found in main composition"

   -- Takes in the cooperation set and the set of actions performed by
   -- a component and returns the set of actions which should be hidden.
   createHideSet :: CooperationSet -> [ ParsedAction ] -> [ ParsedAction ]
   createHideSet coopSet actions =
      filter ( not . ((flip elem coopSet) . nameOfAction) ) actions

   actionMap = processActionsMap report