{-| 
   This module implements transformations which take the form of
   replacing parts of a model with new parts.
   Generally names with either parts or other names.
-}
module Language.Pepa.Transform.Replace
   ( -- * Renaming processes
     ProcessRename
   , ProcessRenameMap
   , replaceComponentNames
   , renameCompsInRateDef
   , renameCompsProcessDef

   , ComponentReplace
   , ComponentMap
   , replaceCompNamesComponent

     -- * Renaming rates
   , RateRename
   , RateRenameMap
   , replaceRateNames
   , renameRatesInRateDef

     -- * Overriding\/adding rate definition
   , overrideRateDefinition
   , overrideRateDefinitionList
   )
where

{- Standard Libraries modules imported -}
import Data.Maybe
   ( mapMaybe )
{- External Library modules imported -}
{- Local modules imported imported -}
import Language.Pepa.Syntax
   ( ParsedModel       ( .. )
   , RateSpec
   , ProcessDef
   , ParsedComponentId
   , ParsedComponent   ( .. )
   , BeCondition       ( .. )
   , ParsedTrans       ( .. )
   , RateIdentifier
   , ParsedRate        ( .. )
   , ParsedRateExp     ( .. )
   , ParsedCond        ( .. )
   )   
import Language.Pepa.Utils
   ( mapSnd )
{- End of imports -}

{-| The type of a single process renaming -}
type ProcessRename = ( ParsedComponentId, ParsedComponentId )

{-| The type of a whole mapping from process names to process names -}
type ProcessRenameMap = [ ProcessRename ]

{-|
   'replaceComponentNames' takes in a mapping from component names to 
   components and a model and replaces each occurrence of a name 
   within the mapping with it's associated component.
   [@todo@] should be replaced within a rate expression as well.
-}
replaceComponentNames :: ProcessRenameMap -> ParsedModel -> ParsedModel
replaceComponentNames mapping (ParsedModel rDefs pDefs mainCoop) =
   ParsedModel (map (renameCompsInRateDef mapping) rDefs) 
               (map (renameCompsProcessDef mapping) pDefs)
               (renameCompsInComponent mapping mainCoop)


{-| Renaming of component namesapplied to rate specifications -}
renameCompsInRateDef :: ProcessRenameMap -> RateSpec -> RateSpec
renameCompsInRateDef mapping = mapSnd $ replaceCompNamesInRateExp mapping

renameCompsProcessDef :: ProcessRenameMap -> ProcessDef -> ProcessDef
renameCompsProcessDef mapping (name, comp) =
   case lookup name mapping of
      Nothing -> (name, renameCompsInComponent mapping comp)
      Just p  -> (p   , renameCompsInComponent mapping comp)
  
{-| Rename all the component names within a given component -}
renameCompsInComponent :: ProcessRenameMap -> ParsedComponent -> ParsedComponent
renameCompsInComponent mapping = 
   replaceCompNamesComponent componentMap
   where   
   componentMap = map (mapSnd IdProcess) mapping


-- | The type of a single component replacement
type ComponentReplace = ( ParsedComponentId, ParsedComponent ) 

-- | The type of a mapping from process names to components
type ComponentMap = [ ComponentReplace ]


{-|
   'replaceCompNamesComponent' takes in a mapping from component names to 
   components and a component and replaces each occurrence of a name 
   within the mapping with it's associated component.
   [@todo@] should be replaced within a rate expression as well.
-}
replaceCompNamesComponent :: ComponentMap -> ParsedComponent -> ParsedComponent
replaceCompNamesComponent _db proc@(StopProcess)    = proc
replaceCompNamesComponent db proc@(IdProcess ident) =
   case lookup ident db of
      Nothing -> proc
      Just p  -> p
replaceCompNamesComponent db (PrefixComponent trans nextP)     =
   PrefixComponent newTrans $ replaceCompNamesComponent db nextP
   where
   newTrans = trans { pepaTransRate = replaceCompNamesInRate renameDb rate }
   rate     = pepaTransRate trans
   renameDb = convertComponentMap db
replaceCompNamesComponent db (CondBehaviour cond nextP)        =
   CondBehaviour newCond $ replaceCompNamesComponent db nextP
   where
   newCond = replaceCompNamesInCondition cond
   renameDb = convertComponentMap db

   replaceCompNamesInCondition :: BeCondition -> BeCondition
   replaceCompNamesInCondition (BeCondition rident operator rExpr) =
      BeCondition rident operator $ replaceCompNamesInRateExp renameDb rExpr
replaceCompNamesComponent db (ComponentSum left right)         =
   ComponentSum (replaceCompNamesComponent db left) 
                (replaceCompNamesComponent db right)
replaceCompNamesComponent db (Cooperation left actions right ) =
   Cooperation (replaceCompNamesComponent db left) actions
               (replaceCompNamesComponent db right)
replaceCompNamesComponent db  proc@(ProcessArray ident s a)    =
   case lookup ident db of
      Nothing -> proc
      Just (IdProcess ident2) -> (ProcessArray ident2 s a)
      Just _                  ->
         error "Process array names cannot be changed with components"
replaceCompNamesComponent db (Hiding proc actions)             =
   Hiding (replaceCompNamesComponent db proc) actions


{- Converts a 'ComponentMap' into a 'ProcessRenameMap' -}
convertComponentMap :: ComponentMap -> ProcessRenameMap
convertComponentMap = 
   mapMaybe getRenameMapping
   where   
   getRenameMapping :: ComponentReplace -> Maybe ProcessRename
   getRenameMapping (pident1, IdProcess pident2) = Just (pident1, pident2)
   getRenameMapping _                            = Nothing


replaceCompNamesInRate :: ProcessRenameMap -> ParsedRate -> ParsedRate
replaceCompNamesInRate _db rate@(RateTop)             = rate
replaceCompNamesInRate _db rate@(RateImmediate _)     = rate
replaceCompNamesInRate db  (RateTimed e)              =
   RateTimed $ replaceCompNamesInRateExp db e

{- 
   Replace component names within a rate expression, note that they may only be
   replaced with a component name, not with an arbitrary component.
-}
replaceCompNamesInRateExp :: ProcessRenameMap -> ParsedRateExp -> ParsedRateExp
replaceCompNamesInRateExp _db rate@(RateId _ _)          = rate
replaceCompNamesInRateExp _db rate@(RateNum _)           = rate
replaceCompNamesInRateExp db  (RateBexp oper left right) =
   RateBexp oper (replaceCompNamesInRateExp db left)
                 (replaceCompNamesInRateExp db right)
replaceCompNamesInRateExp db  rate@(RateProc pident)     =
   case lookup pident db of
      Nothing -> rate
      Just p  -> RateProc p
replaceCompNamesInRateExp db (RateCond pCond left right) =
   RateCond (replaceInPCond pCond)
            (replaceCompNamesInRateExp db left)
            (replaceCompNamesInRateExp db right)
   where
   replaceInPCond :: ParsedCond -> ParsedCond
   replaceInPCond cond@(PCondId pident) =
      case lookup pident db of
         Nothing -> cond
         Just p  -> PCondId p
   replaceInPCond (PCondSum leftp rightp)  =
      PCondSum (replaceInPCond leftp) (replaceInPCond rightp)
   replaceInPCond (PCondRate rIdent operator rExpr) =
      PCondRate rIdent operator $ replaceCompNamesInRateExp db rExpr
replaceCompNamesInRateExp _db rate@(RateDisabled)        = rate


{-| The type of a single rate replacement -}
type RateRename    = ( RateIdentifier, RateIdentifier )
{-| The type of a rate renaming mapping -}
type RateRenameMap = [ RateRename ]

{-| Replace the rate names in an entire model -}
{- Note that we do not replace the rates in the system component, since
   this should normally not contain any rates
-}
replaceRateNames :: RateRenameMap -> ParsedModel -> ParsedModel
replaceRateNames db (ParsedModel rDefs pDefs mainCoop) =
   ParsedModel newRDefs newPDefs mainCoop
   where
   newRDefs = map (renameRatesInRateDef db) rDefs
   newPDefs = map renamePDef pDefs
   
   renamePDef :: ProcessDef -> ProcessDef
   renamePDef = mapSnd $ replaceRateInComponent db


renameRatesInRateDef :: RateRenameMap -> RateSpec -> RateSpec
renameRatesInRateDef db (name, rexp) =
   case lookup name db of
      Just r  -> (r,    replaceRateRexp db rexp)
      Nothing -> (name, replaceRateRexp db rexp)

{- Replace the rate names in a rate. Of course if the rate is not a
   rate expression then it just remains the same (if it is immediate or infty)
   if however it's a rate expression then we must descent into the rate 
   expression (via 'replaceRateRexp')
-}
replaceRateInRate :: RateRenameMap -> ParsedRate -> ParsedRate
replaceRateInRate _db r@(RateTop)                = r
replaceRateInRate _db r@(RateImmediate _)        = r
replaceRateInRate db (RateTimed e)               =
   RateTimed $ replaceRateRexp db e

replaceRateRexp :: RateRenameMap -> ParsedRateExp -> ParsedRateExp
replaceRateRexp db (RateId rateId mRexp)       =
      case lookup rateId db of
         Just r  -> RateId r newMrexp
         Nothing -> RateId rateId newMrexp
      where
      newMrexp = do rexp <- mRexp
                    return $ replaceRateRexp db rexp
replaceRateRexp _db r@(RateNum _)              = r
replaceRateRexp db (RateBexp oper left right)  = 
      RateBexp oper (replaceRateRexp db left)
                    (replaceRateRexp db right)
replaceRateRexp _db r@(RateProc _)             = r
replaceRateRexp db (RateCond p left right)     =
   RateCond (replaceInPCond p)
            (replaceRateRexp db left)
            (replaceRateRexp db right)
   where
   replaceInPCond :: ParsedCond -> ParsedCond
   replaceInPCond cond@(PCondId _pident)            = cond
   replaceInPCond (PCondSum leftp rightp)           =
      PCondSum (replaceInPCond leftp) (replaceInPCond rightp)
   replaceInPCond (PCondRate rIdent operator rExpr) =
      case lookup rIdent db of
         Just r  -> PCondRate r operator (replaceRateRexp db rExpr)
                              
         Nothing -> PCondRate rIdent operator (replaceRateRexp db rExpr)
   -- Rate weights, they cannot currently contain
   -- rate identifiers, but maybe even if we just allow weights to
   -- themselves be identifiers we would want to replace them here.
replaceRateRexp _db r@(RateDisabled)           = r

replaceRateInComponent :: RateRenameMap -> ParsedComponent -> ParsedComponent
replaceRateInComponent _db p@(StopProcess)              = p
replaceRateInComponent _db p@(IdProcess _)              = p
replaceRateInComponent db (PrefixComponent trans next)  =
      PrefixComponent (trans { pepaTransRate = replaceRateInRate db rate } )
                      (replaceRateInComponent db next)
      where
      rate = pepaTransRate trans
replaceRateInComponent db (CondBehaviour cond next)     =
   CondBehaviour (replaceRateInCondition cond)
                 (replaceRateInComponent db next)
   where
   replaceRateInCondition :: BeCondition -> BeCondition
   replaceRateInCondition (BeCondition rIdent operator rExpr) =
      case lookup rIdent db of
         Just r  -> BeCondition r operator (replaceRateRexp db rExpr)
                              
         Nothing -> BeCondition rIdent operator (replaceRateRexp db rExpr)
replaceRateInComponent db (ComponentSum left right)     =
      ComponentSum (replaceRateInComponent db left)
                   (replaceRateInComponent db right)
replaceRateInComponent db (Cooperation left acts right) =
      Cooperation (replaceRateInComponent db left) acts
                  (replaceRateInComponent db right)
replaceRateInComponent _db p@(ProcessArray _ _ _)       = p
replaceRateInComponent db (Hiding p acts)               =
      Hiding (replaceRateInComponent db p) acts

{-|
   Override a rate definition. Generally this is to allow a @--rate 1.0@
   style of command-line option. Note that if the rate is not originally
   defined then this has the effect of simply adding the given rate
   definition.
-}
overrideRateDefinition :: RateSpec -> ParsedModel -> ParsedModel
overrideRateDefinition rDef@(rateName, _rExp) 
                       (ParsedModel rDefs pDefs mainSys) =
   ParsedModel newRdefs pDefs mainSys
   where
   newRdefs = rDef : filtered
   filtered = filter notGivenRateName rDefs

   notGivenRateName :: RateSpec -> Bool
   notGivenRateName = not . (== rateName) . fst

{-| Applies 'overrideRateDefinition' for all the given rate definitions -}
overrideRateDefinitionList :: ParsedModel -> [ RateSpec ] -> ParsedModel
overrideRateDefinitionList = foldr overrideRateDefinition