{-|
   The main module of the smc compiler, which compiles srmc models to
   batches of pepa models.
-}
module Main
    ( main )
where

{- Standard library modules imported -}
import System.Directory
   ( doesFileExist )
{- External Library modules imported -}
{- Local modules imported -}
import Ipc.Cli
   ( getCliArgs
   , Cli            ( .. )
   , CliOpt         ( .. )
   , toCli
   , getNonStandard
   )
import Smc.Cli
   ( smcVersion
    , SmcOptions    ( .. )
    , smcOptions
    )
import Language.Pepa.Syntax
   ( RateSpec
   , ParsedRateExp
   )
import Language.Pepa.Srmc.Syntax
   ( SrmcModel )
import Language.Pepa.Srmc.Parser
   ( parseSrmcFile )
import Language.Pepa.Print
   ( HumanPrint     ( .. )
   , hPrintPepaModel
   , hprintRateIdentifier
   )
import Language.Pepa.Utils
   ( switchSuffix
   , switchDirectory
   , dirName
   , mkCSlist
   , permuteDefinitionArrays
   )
-- import Language.Pepa.Srmc.Print
--    ( hPrintSrmcModel )
import Language.Pepa.Srmc.SrmcPepa
   ( TranslatedModel   ( .. )
   , RateArray
   , translateSrmc
   )

{- End of module imports -}


main :: IO ()
main = do args         <- getCliArgs
          processArgs args
   
-- Very simplistic options parsing there is no need to go to the
-- length of using 'System.GetOpt' since so far we only have one option.
-- In the future probably the whole library will have a common set of
-- command-line options and we can use that function to parse them.       
processArgs :: [ String ] -> IO () 
processArgs cliArgs =
   case toCli True smcVersion "smc" smcOptions cliArgs of
     CliValid options inputFiles ->
        mapM_ (convertFile options) inputFiles
     CliInter _ _                ->   
         putStrLn "smc cannot be run in interactive mode"
     CliInfo  _ _ infoString     -> 
         putStrLn infoString
     CliError  _ _ errorString   -> 
         putStrLn errorString


-- We convert a file to possibly many pepa models.
-- The number of models depends on the number of process sets defined
-- So we number the models and output them according to their number.
convertFile :: [ CliOpt SmcOptions ] -> FilePath -> IO ()
convertFile options fileName =
   do fileExists <- doesFileExist fileName
      if fileExists
         then do contents   <- readFile fileName
                 case parseSrmcFile fileName contents of
                    Left err      -> putStrLn $ show err
                    Right sModel  -> processSrmcModel options fileName sModel
         else putStrLn $ "Input file: " ++ fileName ++ " does not exist"



-- Does the processing of the srmc model once we have gotten past all
-- the file IO and parsing stuff.
processSrmcModel :: [ CliOpt SmcOptions ] -> FilePath -> SrmcModel -> IO ()
processSrmcModel options fileName sModel =
   mapM_ (printModelToFile options outputName) numberedModels
   where
   numberedModels = zip [0..] pModels
   pModels        = translateSrmc options sModel
   -- If the user has set the output directory in which to place all the 
   -- output pepa models then we update the template name to reflect this.
   outputName = case [ directory |
                      (CliNonStandard (PepaCondorDir directory)) <- options
                     ]
                of
                   []      -> fileName
                   (h : _) -> switchDirectory fileName h


-- Outputs the model according to its number and the template file name
-- which may have been updated to reflect a new directory.
printModelToFile :: [ CliOpt SmcOptions ]
                 -> FilePath
                 -> (Int, TranslatedModel)
                 -> IO ()
printModelToFile options templateName 
                 (number, tModel@(TModel model rateArraysUnPermuted)) =
   do writeFile fileName $ hPrintPepaModel model
      printCondorToFile options fileName tModel
      printDataBaseFile fileName rateArrays
      printBatchCommandFile options fileName number rateArrays  
   where
   fileName     = switchSuffix ".srmc" suffix templateName
   suffix       = (show number) ++ ".pepa"
   rateArrays   = permuteDefinitionArrays rateArraysUnPermuted

-- Outputs a condor file based on the rate arrays of the translated model
printCondorToFile :: [ CliOpt SmcOptions ] 
                  -> FilePath
                  -> TranslatedModel
                  -> IO ()
printCondorToFile options pepaFileName (TModel _ rateArrays) =
   writeFile fileName contents
   where
   fileName = switchSuffix ".pepa" ".condor" pepaFileName
   contents =   
      unlines [ "################################################"
              , "#"
              , commentLine
              , "#"
              , "# This file has been automatically generated by"
              , "# you should edit initial directory option"
              , "#" 
              , "# Execute the hypergraph response-time analyser "
              , "# (hydra) on output from the ipc many times"
              , "#"
              , "################################################"
              , ""
              , "universe      = vanilla"
              , ""
              , " # May require edit here"
              , "initialdir    = " ++ initialdir
              , "executable    = $(initialdir)/hydrus.sh"
              , "arguments     = " ++ arguments
              , "environment   = HYDRA_S_DIR=$(initialdir)/"
              , ""
              , "GetEnv        = True"
              , ""
              , "should_transfer_files = YES"
              , "when_to_transfer_output = ON_EXIT"
              , ""
              , "output        = "
                ++ "$(initialdir)/output/hydrus.output.$(Process).$(cluster)"
              , "error         = "
                ++ "$(initialdir)/output/hydrus.error.$(Process).$(cluster)"
              , "log           = /tmp/output.$(Process).$(cluster)"
              , "#log           = "
              ++ "$(initialdir)/output/hydrus_condor.log.$(Process).$(cluster)"
              , ""
              , "#Recieve an e-mail notification only if there is an error"
              , "notification  = ERROR"
              , ""
              , "# kill off any instance of setiathome which is running"
              , "nice_user     = false"
              , ""
              , "# run copies of Hydrus simultaneously"
              , queueLine
              ]
   commentLine  = "# Filename: " ++ fileName
   queueLine    = "queue " ++ show numProcesses
   arguments    = unwords [ "$(initialdir) $(Process) $(cluster)"
                          , pepaFileName
                          , ipcArgs
                          , rateArgs
                          ]
   ipcArgs      = produceCondorArgs options
   
   rateArgs     = unwords $ map makeRateArrayOpt rateArrays

   initialdir   = dirName pepaFileName
   
   -- This is the number of combinations of the rate arrays
   numProcesses = foldr (*) 1 $ map (length . snd) rateArrays

   makeRateArrayOpt :: RateArray -> String
   makeRateArrayOpt (rateId, rateExps) =
      concat  [ "--rates "
              , hprintRateIdentifier rateId
              , "="
              , mkCSlist $ map hprint rateExps
              ]
{-
   Produces a string representing the arguments which are suitable for
   batch processing.
   Note the source and target options are temporary until we actually look
   at the options given into srmc, we might also wish to do different ones
   once we have generated the pepa models. That is we might not wish to
   set it here.

   I probably want to check the cli options for a couple of more things that
   should be passed on.
-}
produceBatchArgs :: [ CliOpt SmcOptions ] -> String
produceBatchArgs args = 
   unwords $ allIpcArgs ++ alwaysArgs
   where
   allIpcArgs = [ s | IpcArgs s <- getNonStandard args ]
   -- We currently have /--no-static-analysis/ and probably always will
   -- do because any static analysis should be done at the srmc level
   -- and may not even make sense at the pepa level, eg cooperating over
   -- an action which does not get performed by one side of the cooperation
   -- simply because of the current state of the system.
   alwaysArgs = [ "--no-static-analysis" ]
-- "--staunch --permute --source start --target finish --timestep 1.0"

produceCondorArgs :: [ CliOpt SmcOptions ] -> String
produceCondorArgs = produceBatchArgs


{-
  In a similar way to producing the condor file we can also produce
  a process identifier data base. This maps process identifiers to
  rate constants so that once all the experiments have been done the
  results can be looked up.
  A quick example of what will be produced if the translated model
  used rates @r@ and @s@ and they were defined in the model as:
  @
   r = {0.1, 0.2 } ;
   s = {0.4, 0.5 } ;
  }
  Then for that model the produced file would be:
  @
  #Pid    #r    #s
  0       0.1   0.4
  1       0.1   0.5
  2       0.2   0.4
  3       0.2   0.5
  @

   Note that the rate arrays which are passed in here should already
   have been permuted.
-}
printDataBaseFile :: FilePath -> [ RateArray ] -> IO ()
printDataBaseFile pepaFileName rateArrays =
   writeFile fileName contents
   where
   fileName = switchSuffix ".pepa" ".db" pepaFileName
   contents = produceProcessIdFile rateArrays

{-
   The rate arrays given in here should already be permuted.
-}
produceProcessIdFile :: [ RateArray ] -> String
produceProcessIdFile rateArrays =
   unlines (headLine : fileLines)
   where
   headLine   = unwords $ "#Pid" : (map makeColumnName rateArrays)
   processIds = [ 0..(numProcesses - 1) ]
   fileLines  = map stringLine rateLines
   rateLines  = map makeRateLine processIds
   
   makeColumnName :: RateArray -> String
   makeColumnName (idNo, _) = "#" ++ (hprintRateIdentifier idNo)
   
   takeRate :: Int -> RateArray -> ParsedRateExp
   takeRate i (_, rates) = rates !! i
   
   makeRateLine :: Int -> (Int, [ ParsedRateExp ])
   makeRateLine i = (i, map (takeRate i) rateArrays)
   
   stringLine :: (Int, [ ParsedRateExp ]) -> String
   stringLine (i, rates) = unwords $ (show i) : (map hprint rates)
   
   numProcesses = case rateArrays of
                     []      -> 0
                     (h : _) -> length $ snd h


{-
   Again the rate arrays which are passed in here are already permuted.
-}
printBatchCommandFile :: [ CliOpt SmcOptions ] 
                       -> FilePath
                       -> Int
                       -> [ RateArray ]
                       -> IO ()
printBatchCommandFile options pepaFileName batchNumber rateArrays =
   writeFile fileName contents
   where
   fileName = switchSuffix ".pepa" ".batch.sh" pepaFileName
   contents = produceBatchCommandFile options pepaFileName 
                                      batchNumber rateArrays

{-
   Rather than running over condor we may wish to run on a single machine.
   This is essentially the same as the process id file together with the
   condor file. 
   The rate arrays passed in should be already permuted.
-}
produceBatchCommandFile :: [ CliOpt SmcOptions ] 
                        -> FilePath
                        -> Int
                        -> [ RateArray ]
                        -> String
produceBatchCommandFile options pepaFileName _batchNumber rateArrays =
   unlines commandLines
   where
   commandLines = map makeCommandLine processIds
   processIds   = [ 0..(numProcesses - 1) ]
   numProcesses = case rateArrays of
                     []      -> 0
                     (h : _) -> length $ snd h

   ipcCommand   = case [ file | IpcLocation file <- getNonStandard options ] of
                     []      -> "ipc"
                     [ ipc ] -> ipc
                     _       -> error "multiple --ipc options specified"

   makeCommandLine :: Int -> String
   makeCommandLine i = 
      unwords $ [ ipcCommand
                , pepaFileName
                , "--mod-file " ++ modFileName
                ]
                ++ [ ipcArgs ]
                ++ rateArgs
{- The old hydrus way
      unwords $ [ "./hydrus.sh"
                , initialdir
                , show i
                , show batchNumber
                , pepaFileName
                ]
                ++ [ ipcArgs ]
                ++ rateArgs
-}
      where
      modFileName = switchSuffix ".pepa" modSuffix pepaFileName
      modSuffix   = "." ++ (show i) ++ ".mod"
      -- initialdir  = dirName pepaFileName
      ipcArgs     = produceBatchArgs options

      rateArgs    = map makeRateArg rateSpecs
      rateSpecs   = map (takeRate i) rateArrays
      
      takeRate :: Int -> RateArray -> RateSpec
      takeRate n (rId, rates) = (rId, rates !! n)

      makeRateArg :: RateSpec -> String
      makeRateArg (ident, expr) = 
         concat [ "--rate "
                , hprintRateIdentifier ident
                , "="
                , hprint expr
                ]
