{-|
    This is a simple example use of the Pepa library.
    Given a pepa model file we produce an equivalent
    timed system view of the model.
-}
module Main
    ( main )
where

{- Standard library modules imported -}
import System.Console.GetOpt
   ( OptDescr  ( .. )
   , ArgDescr  ( .. )
   )
import System.Directory
   ( doesFileExist )
import Text.ParserCombinators.Parsec 
    ( ParseError )
{- External Library modules imported -}
{- Local modules imported -}
import Ipc.Cli
   ( getCliArgs
   , Cli            ( .. )
   , CliOpt         ( .. )
   , toCli
   , baseCliOptions
   , getProbeSpecs
   , getMasterBool
   , getStartActions
   , getStopActions

   , getNonStandard
   )
import Language.Pepa.Syntax
    ( ParsedModel )
import Language.Pepa.Parser
    ( parsePepaFile )
import Language.Pepa.Print
    ( hPrintPepaModel )

import Language.Pepa.Probes.StochasticProbes
   (  ProbeDef )
import Language.Pepa.Probes.AddProbes
    ( addMasterProbe
    , addProbeDefs
    , ProbeTranslateFlags    ( .. )
    )
import Language.Pepa.Probes.Translate
   (  probeToDotFile )
import Language.Pepa.Probes.ProbeParser
    ( parseProbe )
{- End of module imports -}


main :: IO ()
main = do args         <- getCliArgs
          processArgs args
   
-- Processing of the command-line arguments.
processArgs :: [ String ] -> IO () 
processArgs cliArgs =
   case toCli True version "pepaprobe" pepaProbeOptions cliArgs of
     CliValid options inputFiles ->
        case parseProbeStrings probeStrings of
           Right probes
               | elem OutputDot probeOpts ->
                  mapM_ (convertToDot trFlags) probes
               | otherwise                ->
                  mapM_ (convertFile options addMaster probes) inputFiles
           Left err     -> putStrLn $ show err
        where
        probeStrings = getProbeSpecs  options
        addMaster    = getMasterBool  options
        probeOpts    = getNonStandard options
        trFlags      = getTrFlags     options
     CliInter _ _                ->   
         putStrLn "pepaprobe cannot be run in interactive mode"
     CliInfo  _ _ infoString     -> 
         putStrLn infoString
     CliError  _ _ errorString   -> 
         putStrLn errorString


version :: String
version = "This is pepaprobe version 0.01"

data PepaProbeOpts = OutputDot
                   | TranslateFlag ProbeTranslateFlags
                   deriving Eq

pepaProbeOptions :: [ OptDescr ( CliOpt PepaProbeOpts ) ]
pepaProbeOptions = 
   baseCliOptions ++
   [ Option ""     [ "output-dot" ] 
     (NoArg (CliNonStandard OutputDot) )
     "Instead of adding to the model output the probe as a .dot file"

   , Option ""     [ "no-minimise" ] 
     (NoArg (CliNonStandard $ TranslateFlag NoMinimise) )
     "Do not minimise the deterministic finite automata"

   , Option ""     [ "min-self-loops" ]
     (NoArg (CliNonStandard $ TranslateFlag MinSelfLoops) )
     "add the self-loops before performing minimisation (not really tested)"

   , Option ""     [ "no-min-self-loops" ]
     (NoArg (CliNonStandard $ TranslateFlag NoMinSelfLoops) )
     ( "perform minimisation before adding the self-loops " ++
       "(such that they are not minimised)" )

   , Option ""     [ "no-self-loops" ]
     (NoArg (CliNonStandard $ TranslateFlag NoSelfLoops) )
     "do not add the self-loops (useful mostly for debugging)"

   , Option ""     [ "non-deterministic" ]
     (NoArg (CliNonStandard $ TranslateFlag NonDeterm) )
     ( "leave the finite automata non-deterministic" ++
       "(again useful mostly for debugging)" )
   ]

getTrFlags :: [ CliOpt PepaProbeOpts ] -> [ ProbeTranslateFlags ]
getTrFlags options = [ f | TranslateFlag f <- getNonStandard options ]

                   

parseProbeStrings :: [ String ] -> Either ParseError [ ProbeDef ]
parseProbeStrings []         = Right []
parseProbeStrings (h : rest) =
   joinParseResults (:) (parseProbe h) $ parseProbeStrings rest


{-
   The 'addMaster' argument here specifies whether or not to add
   a master probe (essentially whether or not --no-master was in
   in the list of options).
-}
convertFile :: [ CliOpt PepaProbeOpts ] -> Bool -> [ ProbeDef ] 
            -> FilePath -> IO ()
convertFile options addMaster probeDefs file = 
    do fileExists <- doesFileExist file
       if fileExists 
          then do contents <- readFile file
                  let pModel     = showEitherParseError $ parsePepaFile file contents
                      newModel   = bindEithers addProbes pModel
                      output     = formatResult newModel
                  putStrLn output
          else putStrLn ("file: " ++ file ++ " does not exist")
   where
   addProbes :: ParsedModel -> Either String ParsedModel
   addProbes = addProbesToModel options addMaster probeDefs

addProbesToModel :: [ CliOpt PepaProbeOpts ] -> Bool -> [ ProbeDef ] 
                 -> ParsedModel -> Either String ParsedModel
addProbesToModel options False probeDefs model = 
   addProbeDefs (getTrFlags options) probeDefs model
addProbesToModel options True  probeDefs model =
   bindEithers ( Right . (addMasterProbe startActs stopActs) )$ 
      addProbeDefs (getTrFlags options) probeDefs model
   where
   startActs    = getStartActions options
   stopActs     = getStopActions  options
   

{-
   Convert a file into dot format
-}
convertToDot :: [ ProbeTranslateFlags ] -> ProbeDef -> IO ()
convertToDot trFlags probedef =
   putStrLn $ probeToDotFile trFlags "ProbeName" $ snd probedef


formatResult :: Either String ParsedModel -> String
formatResult (Left err)    =
   "Failure: " ++ err
formatResult (Right model) = hPrintPepaModel model

{-
formatParseResult :: Either ParseError a -> (a -> String) -> String
formatParseResult (Left err) _ = show err
formatParseResult (Right a)  f = f a
-}

showEitherParseError :: Either ParseError a -> Either String a
showEitherParseError (Left err) = Left $ show err
showEitherParseError (Right a)  = Right a

bindEithers :: (a -> Either err c) -> Either err a  -> Either err c
bindEithers _ (Left err) = Left err
bindEithers f (Right b)  = f b

{-
interpretParseResult :: Either ParseError a -> (a -> b) 
                     -> Either ParseError b
interpretParseResult (Left err) _ = Left err
interpretParseResult (Right a)  f = Right $ f a
-}

joinParseResults :: (a -> b -> c)
                 -> Either ParseError a
                 -> Either ParseError b
                 -> Either ParseError c
joinParseResults _ (Left err) _          = Left err
joinParseResults _ _          (Left err) = Left err
joinParseResults f (Right a)  (Right b)  = Right $ f a b
